#include "disp.hpp"

extern ObsConf obs;
extern off_t size;

/* calculate phase due to dispersion */ 
double get_disp_phase (double f) {
 double z, index, freq, fedge;
 freq = obs.get_ref_freq ();
 fedge = (obs.is_band_direct() ? obs.get_center_freq() - obs.get_bandwidth()/2. : obs.get_center_freq() + obs.get_bandwidth()/2.);
 index = 1.e6;

 if (freq >= obs.get_center_freq() - obs.get_bandwidth()/2. && freq <= obs.get_center_freq() + obs.get_bandwidth()/2.) {
   z = ( ( ALPHA1 * obs.get_dispersion_measure() ) / f / ( fedge * fedge * index ) ) * ( f - fedge ) * ( f - fedge );
  } else { 
   z = ( ( ALPHA1 * obs.get_dispersion_measure() ) / f / ( fedge * fedge * index ) ) * ( f - fedge ) * ( f - fedge ) - ( ( ALPHA1 * obs.get_dispersion_measure() ) / index ) * f * ( 1./( fedge * fedge ) - 1./( freq * freq ) );
  }
 // z = ( ALPHA1 * obs.get_dispersion_measure() ) / f / index;
 return z;
}

/* compute dedispersion matrix */
fftwf_complex* get_matrix (fftwf_complex *matrix, off_t size) {
 double z = 0.;
 double T, freq, f0;
 float sign = (obs.is_band_direct() ? -1. : 1. ); // direct band = -1, inverse band = 1
 T = (obs.get_time_resolution() * ((double)size))/((double)(1.e3));
 f0 = (obs.is_band_direct() ? obs.get_center_freq() - obs.get_bandwidth()/2. : obs.get_center_freq() + obs.get_bandwidth()/2.);

 memset (matrix, 0, sizeof(fftwf_complex)*(size/2+1));
 for (off_t i=0; i < size/2 + 1; ++i) {
   freq = f0 + (-1. * sign)*(((double)i)/T);
   z = get_disp_phase (freq);
   matrix[i][0] = (float)cos(z);
   matrix[i][1] = sign * (float)sin(z);
  }
 return matrix;
}

/* ship FFT data out in two complex arrays */
/* part1 originally has packed completely, part2 is empty */
void ship_out2complex (fftwf_complex *part1, fftwf_complex *part2, off_t size) {
 float rep, rem, aip, aim;

 part2[0][0] = part1[0][1];
 part1[0][1] = part2[0][1] = 0.;

 for (off_t j=1; j<size/2+1; ++j) {
  rep = 0.5 * (part1[j][0] + part1[size-j][0]);
  rem = 0.5 * (part1[j][0] - part1[size-j][0]);
  aip = 0.5 * (part1[j][1] + part1[size-j][1]);
  aim = 0.5 * (part1[j][1] - part1[size-j][1]);
  part1[j][0] = rep;
  part1[j][1] = aim;
  part1[size-j][0] = rep;
  part1[size-j][1] = -aim;
  part2[j][0] = aip;
  part2[j][1] = -rem;
  part2[size-j][0] = aip;
  part2[size-j][1] = rem;
 }
}

/* ship FFT data from the GBT to arrange in total band */
/* dest is 4*n size, part1 & part is n size */
fftwf_complex* ship_out2gbt_total (fftwf_complex *dest, fftwf_complex *part1, fftwf_complex *part2, off_t size) {
 fftwf_complex *temp;
 temp = (fftwf_complex *)fftwf_malloc (sizeof(fftwf_complex) * size); 
 if (!temp) { perror ("fftwf_malloc"); exit (1); }
 memset (temp, 0, sizeof(fftwf_complex) * size);

 /* part2 has Y-channel in Re, and Z-channel in Im */
 ship_out2complex (part2, temp, size);
 /* now part2 has Y-channel and temp has Z-channel */
 for (off_t i=size/2+1, j=0; i<size; ++i, ++j) { dest[j][0] = temp[i][0]; dest[j][1] = temp[i][1]; } 
 for (off_t i=0, j=size/2-1; i<size/2+1; ++i, ++j) { dest[j][0] = part2[i][0]; dest[j][1] = part2[i][1]; } 

 memset (temp, 0, sizeof(fftwf_complex) * size);
 /* part1 has W-channel in Re, and X-channel in Im */
 ship_out2complex (part1, temp, size);
 /* now part1 has W-channel and temp has X-channel */
 for (off_t i=size/2+1, j=size; i<size; ++i, ++j) { dest[j][0] = temp[i][0]; dest[j][1] = temp[i][1]; } 
 for (off_t i=0, j=size+size/2-1; i<size/2+1; ++i, ++j) { dest[j][0] = part1[i][0]; dest[j][1] = part1[i][1]; } 
 
 fftwf_free (temp);
 return dest;
}

/* making the dedispersion */
fftwf_complex* dedispersion (fftwf_complex *ampl_sp, fftwf_complex *matrix, float *band, fftwf_plan plan, off_t size, bool do_fraction, bool do_zero) {
 float re;
 if (!do_fraction) {
  for (off_t i=0; i<size/2 + 1; ++i) {
   re = ampl_sp[i][0];
   ampl_sp[i][0] = 2. * (re * matrix[i][0] - matrix[i][1] * ampl_sp[i][1]) * band[i];
   ampl_sp[i][1] = 2. * (re * matrix[i][1] + matrix[i][0] * ampl_sp[i][1]) * band[i];
  }
 } else { // do fraction
  float im;
  double z;
  for (off_t i=0; i<size/2 + 1; ++i) {
   z = ((-0.50 * M_PI) * ((double)i))/((double)size);
   re = ampl_sp[i][0] * cos(z) - ampl_sp[i][1] * sin(z);
   im = ampl_sp[i][0] * sin(z) + ampl_sp[i][1] * cos(z);
   ampl_sp[i][0] = 2. * (re * matrix[i][0] - matrix[i][1] * im) * band[i];
   ampl_sp[i][1] = 2. * (re * matrix[i][1] + matrix[i][0] * im) * band[i];
  }
 }
 if (do_zero) for (off_t i=size/2 + 1; i<size; ++i) { ampl_sp[i][0] = 0.; ampl_sp[i][1] = 0.; }
 fftwf_execute (plan);
 return ampl_sp;
}

/* get ACF0 value */
double get_acf0 (fftwf_complex *x, off_t size) {
 double acf = 0.;
 for (long i=0; i<size; ++i)
  acf += (x[i][0] * x[i][0] + x[i][1] * x[i][1]) * cos(((double)(M_PI * (double)size * i))/((double)size));
 acf *= (double)(2./(double)size);
 return fabs(acf);
}

/*get ACF0 value calculating the ACF itself */
double get_acf (fftwf_complex *x, off_t size) {
 fftwf_complex *temp;
 temp = (fftwf_complex *)fftwf_malloc (sizeof(fftwf_complex) * size);
 if (!temp) { perror ("fftwf_malloc"); exit (1); }
 memset (temp, 0, sizeof(fftwf_complex) * size);
   
 for (off_t i=0; i<size/4+1; ++i) {
  temp[i][0] = x[i][0] * x[i][0] + x[i][1] * x[i][1];
  temp[i][1] = x[i][1] * x[i][0] - x[i][0] * x[i][1];
 }
 for (off_t i=size/4+1; i<size; ++i) { temp[i][0] = 0.; temp[i][0] = 0.; }
 fftwf_plan back;
 back = fftwf_plan_dft_1d (size, temp, temp, FFTW_BACKWARD, FFTW_EXHAUSTIVE);
 fftwf_execute (back);
 // double acf0 = ((temp[0][0]/(double)size) * (temp[0][0]/(double)size) + (temp[0][1]/(double)size) * (temp[0][1]/(double)size));
 double acf0 = (double)temp[0][0] * temp[0][0] + (double)temp[0][1] * temp[0][1];
 fftwf_free (temp);
 fftwf_destroy_plan (back);
 return acf0;
}
