#include "disp.hpp"

extern ObsConf obs;
extern off_t size;

/* calculate phase due to dispersion */ 
double get_disp_phase (double f) {
 double z, index, freq, fedge;
 freq = obs.get_ref_freq ();
 fedge = (obs.is_band_direct() ? obs.get_center_freq() - obs.get_bandwidth()/2. : obs.get_center_freq() + obs.get_bandwidth()/2.);
 index = 1.e6;

 if (freq >= obs.get_center_freq() - obs.get_bandwidth()/2. && freq <= obs.get_center_freq() + obs.get_bandwidth()/2.) {
   z = ( ( ALPHA1 * obs.get_dispersion_measure() ) / f / ( fedge * fedge * index ) ) * ( f - fedge ) * ( f - fedge );
  } else { 
   z = ( ( ALPHA1 * obs.get_dispersion_measure() ) / f / ( fedge * fedge * index ) ) * ( f - fedge ) * ( f - fedge ) - ( ( ALPHA1 * obs.get_dispersion_measure() ) / index ) * f * ( 1./( fedge * fedge ) - 1./( freq * freq ) );
  }
 // z = ( ALPHA1 * obs.get_dispersion_measure() ) / f / index;
 return z;
}

/* calculate phase due to dispersion for separate channel */ 
double get_channel_disp_phase (double f, double frefi) {
 double z, index, freq, fedge;
 freq = frefi;
 index = 1.e6;
 z = ( ( ALPHA1 * obs.get_dispersion_measure() ) / f / ( freq * freq * index ) ) * ( f - freq ) * ( f - freq );
 return z;
}

/* give delay between fcentr and separate channle */
/* if delay > 0 so we have to shift data in direction to the end */
/* if delay < 0 so we have to move data to the beginning */
signed long get_channel_delay (double frefi) {
 double tbad, tdelay;
 double f1 = obs.get_center_freq();
 tbad = obs.get_dispersion_measure() * ALPHA2 * ( 1./(f1 * f1) - 1./(frefi * frefi) );
 tdelay = (tbad * 0.001) / (obs.get_time_resolution() * obs.get_channel_number());

 return (signed long)rint(tdelay);
}

/* compute dedispersion matrix */
fftwf_complex* get_matrix (fftwf_complex *matrix, off_t size) {
 double z = 0.;
 double T, freq, f0;
 float sign = (obs.is_band_direct() ? -1. : 1. ); // direct band = -1, inverse band = 1
 T = (obs.get_time_resolution() * ((double)size))/((double)(1.e3));
 f0 = (obs.is_band_direct() ? obs.get_center_freq() - obs.get_bandwidth()/2. : obs.get_center_freq() + obs.get_bandwidth()/2.);

 memset (matrix, 0, sizeof(fftwf_complex)*(size/2+1));
 for (off_t i=0; i < size/2 + 1; ++i) {
   freq = f0 + (-1. * sign)*(((double)i)/T);
   z = get_disp_phase (freq);
   matrix[i][0] = (float)cos(z);
   matrix[i][1] = sign * (float)sin(z);
  }
 return matrix;
}

/* compute dedispersion matrix for individual channel */
fftwf_complex* get_channel_matrix (fftwf_complex *matrix, off_t size, double fcentri, double frefi) {
 double z = 0.;
 double T, freq, f0;
 float sign = (obs.is_band_direct() ? -1. : 1. ); // direct band = -1, inverse band = 1
 double tresi = obs.get_time_resolution() * obs.get_channel_number();
 double bwi = obs.get_bandwidth() / obs.get_channel_number();
 T = (tresi * ((double)size))/((double)(1.e3));
 f0 = (obs.is_band_direct() ? fcentri - bwi/2. : fcentri + bwi/2.);

 memset (matrix, 0, sizeof(fftwf_complex)*(size/2+1));
 for (off_t i=0; i < size/2 + 1; ++i) {
   freq = f0 + (-1. * sign)*(((double)i)/T);
   z = get_channel_disp_phase (freq, frefi);
   matrix[i][0] = (float)cos(z);
   matrix[i][1] = sign * (float)sin(z);
  }
 return matrix;
}

/* ship FFT data out in two complex arrays */
/* part1 originally has packed completely, part2 is empty */
void ship_out2complex (fftwf_complex *part1, fftwf_complex *part2, off_t size) {
 float rep, rem, aip, aim;

 part2[0][0] = part1[0][1];
 part1[0][1] = part2[0][1] = 0.;

 for (off_t j=1; j<size/2+1; ++j) {
  rep = 0.5 * (part1[j][0] + part1[size-j][0]);
  rem = 0.5 * (part1[j][0] - part1[size-j][0]);
  aip = 0.5 * (part1[j][1] + part1[size-j][1]);
  aim = 0.5 * (part1[j][1] - part1[size-j][1]);
  part1[j][0] = rep;
  part1[j][1] = aim;
  part1[size-j][0] = rep;
  part1[size-j][1] = -aim;
  part2[j][0] = aip;
  part2[j][1] = -rem;
  part2[size-j][0] = aip;
  part2[size-j][1] = rem;
 }
}

/* ship FFT data from the GBT to arrange in total band */
/* dest is 4*n size, part1 & part is n size */
fftwf_complex* ship_out2gbt_total (fftwf_complex *dest, fftwf_complex *part1, fftwf_complex *part2, off_t size) {
 fftwf_complex *temp;
 temp = (fftwf_complex *)fftwf_malloc (sizeof(fftwf_complex) * size); 
 if (!temp) { perror ("fftwf_malloc"); exit (1); }
 memset (temp, 0, sizeof(fftwf_complex) * size);

 /* part2 has Y-channel in Re, and Z-channel in Im */
 ship_out2complex (part2, temp, size);
 /* now part2 has Y-channel and temp has Z-channel */
 for (off_t i=size/2+1, j=0; i<size; ++i, ++j) { dest[j][0] = temp[i][0]; dest[j][1] = temp[i][1]; } 
 for (off_t i=0, j=size/2-1; i<size/2+1; ++i, ++j) { dest[j][0] = part2[i][0]; dest[j][1] = part2[i][1]; } 

 memset (temp, 0, sizeof(fftwf_complex) * size);
 /* part1 has W-channel in Re, and X-channel in Im */
 ship_out2complex (part1, temp, size);
 /* now part1 has W-channel and temp has X-channel */
 for (off_t i=size/2+1, j=size; i<size; ++i, ++j) { dest[j][0] = temp[i][0]; dest[j][1] = temp[i][1]; } 
 for (off_t i=0, j=size+size/2-1; i<size/2+1; ++i, ++j) { dest[j][0] = part1[i][0]; dest[j][1] = part1[i][1]; } 
 
 fftwf_free (temp);
 return dest;
}

/* making the dedispersion */
fftwf_complex* dedispersion (fftwf_complex *ampl_sp, fftwf_complex *matrix, float *band, fftwf_plan plan, off_t size, bool do_fraction, bool do_zero) {
 float re;
 if (!do_fraction) {
  for (off_t i=0; i<size/2 + 1; ++i) {
   re = ampl_sp[i][0];
   ampl_sp[i][0] = 2. * (re * matrix[i][0] - matrix[i][1] * ampl_sp[i][1]) * band[i];
   ampl_sp[i][1] = 2. * (re * matrix[i][1] + matrix[i][0] * ampl_sp[i][1]) * band[i];
  }
 } else { // do fraction
  float im;
  double z;
  for (off_t i=0; i<size/2 + 1; ++i) {
   z = ((-0.50 * M_PI) * ((double)i))/((double)size);
   re = ampl_sp[i][0] * cos(z) - ampl_sp[i][1] * sin(z);
   im = ampl_sp[i][0] * sin(z) + ampl_sp[i][1] * cos(z);
   ampl_sp[i][0] = 2. * (re * matrix[i][0] - matrix[i][1] * im) * band[i];
   ampl_sp[i][1] = 2. * (re * matrix[i][1] + matrix[i][0] * im) * band[i];
  }
 }
 if (do_zero) for (off_t i=size/2 + 1; i<size; ++i) { ampl_sp[i][0] = 0.; ampl_sp[i][1] = 0.; }
 fftwf_execute (plan);
 return ampl_sp;
}

/* get ACF0 value */
double get_acf0 (fftwf_complex *x, off_t size) {
 double acf = 0.;
 for (long i=0; i<size; ++i)
  acf += (x[i][0] * x[i][0] + x[i][1] * x[i][1]) * cos(((double)(M_PI * (double)size * i))/((double)size));
 acf *= (double)(2./(double)size);
 return fabs(acf);
}

/*get ACF0 value calculating the ACF itself */
double get_acf (fftwf_complex *x, off_t size) {
 fftwf_complex *temp;
 temp = (fftwf_complex *)fftwf_malloc (sizeof(fftwf_complex) * size);
 if (!temp) { perror ("fftwf_malloc"); exit (1); }
 memset (temp, 0, sizeof(fftwf_complex) * size);
   
 for (off_t i=0; i<size/4+1; ++i) {
  temp[i][0] = x[i][0] * x[i][0] + x[i][1] * x[i][1];
  temp[i][1] = x[i][1] * x[i][0] - x[i][0] * x[i][1];
 }
 for (off_t i=size/4+1; i<size; ++i) { temp[i][0] = 0.; temp[i][0] = 0.; }
 fftwf_plan back;
 back = fftwf_plan_dft_1d (size, temp, temp, FFTW_BACKWARD, FFTW_EXHAUSTIVE);
 fftwf_execute (back);
 // double acf0 = ((temp[0][0]/(double)size) * (temp[0][0]/(double)size) + (temp[0][1]/(double)size) * (temp[0][1]/(double)size));
 double acf0 = (double)temp[0][0] * temp[0][0] + (double)temp[0][1] * temp[0][1];
 fftwf_free (temp);
 fftwf_destroy_plan (back);
 return acf0;
}

/* to dedisperse the data in separate channels and create filterbank in the output */
/* to present as 2D-plot in pgplot */
float* filterbank (float *bank, float *aligned, fftwf_complex *ampl_sp, float *band, off_t size, off_t ts, bool do_fraction, bool do_zero) {

 double bwi; // bandwidth of separate channel in MHz
 double tresi; // time resolution in separate channel in MHz
 double fcentri; // center frequency of separate channel in MHz
 double frefi; // reference frequency of separate channel (choose it as lower edge)
 off_t channelsize; // size of the separate channel
 int Numch = obs.get_channel_number(); // number of frequency channels
 float sign = (obs.is_band_direct() ? 1. : -1. ); // direct band = 1, inverse band = -1
 unsigned long spoiled_counts = obs.get_channel_spoiled_points(); // spoiled counts
 signed long delay; // delay of the frequency channel (with sign)
 unsigned long absdelay; // absolute value of delay

 // bandwidth of separate channel in MHz
 bwi = obs.get_bandwidth() / Numch;
 // time resolution in separate channel in ns
 tresi = obs.get_time_resolution() * Numch;
 // the size of the channel
 channelsize = size / Numch;
	
 fftwf_complex *z; 
 z = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * channelsize);
 if (!z) { perror("fftwf_malloc"); exit (1); }
 fftwf_plan backplan;
 backplan = fftwf_plan_dft_1d (channelsize, z, z, FFTW_BACKWARD, FFTW_EXHAUSTIVE);

 fftwf_complex *matrix;
 matrix = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * (channelsize/2 + 1));
 if (!matrix) { perror("fftwf_malloc"); exit (1); }
 
 /* loop on channels */
 for (int ch=0; ch<Numch; ++ch) {

 fcentri = obs.get_center_freq() + (-1.*sign)*(obs.get_bandwidth()/2.) + sign*(bwi/2.) + sign * ch * bwi;	 
 frefi = obs.get_center_freq() + (-1.*sign)*(obs.get_bandwidth()/2.) + sign * ch * bwi;
 matrix = get_channel_matrix (matrix, channelsize, fcentri, frefi);
 memset (z, 0, sizeof(fftwf_complex) * channelsize);
	
 if (!do_fraction) {
  for (off_t i=ch*(channelsize/2), j=1; i<(ch+1)*(channelsize/2); ++i, ++j) {
   z[j][0] = 2. * (ampl_sp[i+1][0] * matrix[j][0] - matrix[j][1] * ampl_sp[i+1][1]) * band[i+1];
   z[j][1] = 2. * (ampl_sp[i+1][0] * matrix[j][1] + matrix[j][0] * ampl_sp[i+1][1]) * band[i+1];
  }
 } else { // do fraction
  float re, im;
  double ph;
  for (off_t i=ch*(channelsize/2), j=1; i<(ch+1)*(channelsize/2); ++i, ++j) {
   ph = ((-0.50 * M_PI) * ((double)(i+1)))/((double)size);
   re = ampl_sp[i+1][0] * cos(ph) - ampl_sp[i+1][1] * sin(ph);
   im = ampl_sp[i+1][0] * sin(ph) + ampl_sp[i+1][1] * cos(ph);
   z[j][0] = 2. * (re * matrix[j][0] - matrix[j][1] * im) * band[i+1];
   z[j][1] = 2. * (re * matrix[j][1] + matrix[j][0] * im) * band[i+1];
  }
 }
 if (do_zero) for (off_t i=channelsize/2 + 1; i<channelsize; ++i) { z[i][0] = 0.; z[i][1] = 0.; }
 fftwf_execute (backplan);
 // band is direct
 if (obs.is_band_direct()) {
  for (off_t i=0, j=spoiled_counts, k=ch*ts; i<ts; ++i, ++j, ++k)
    bank[k] = (z[j][0] / channelsize) * (z[j][0] / channelsize) + (z[j][1] / channelsize) * (z[j][1] / channelsize);
 } else { // band is inverse
  for (off_t j=0, k=ch*ts; j<ts; ++j, ++k)
    bank[k] = (z[j][0] / channelsize) * (z[j][0] / channelsize) + (z[j][1] / channelsize) * (z[j][1] / channelsize);
 }

 delay = get_channel_delay (frefi);
 absdelay = labs (delay);
 if (delay == 0) for (off_t j=0, k=ch*ts; j<ts; ++j, ++k) aligned[k] = bank[k];
 if (delay > 0) {
  for (off_t i=absdelay, k=ch*ts+absdelay, j=ch*ts; i<ts; ++i, ++k, ++j) aligned[k] = bank[j];
  for (off_t k=ch*ts+ts-absdelay, i=0, j=ch*ts; i<absdelay; ++i, ++j, ++k) aligned[j] = bank[k];
 }
 if (delay < 0) {
  for (off_t j=0, i=ch*ts+absdelay, k=ch*ts; j<ts-absdelay; ++i, ++j, ++k) aligned[k] = bank[i];
  for (off_t j=ts-absdelay, k=ch*ts+ts-absdelay, i=ch*ts; j<ts; ++i, ++j, ++k) aligned[k] = bank[i];
 }

 } // loop

 fftwf_destroy_plan (backplan); 
 fftwf_free (z);
 fftwf_free (matrix);
 
 return bank;
}
