#include "processdata.hpp"
#include "bitstat.hpp"
#include "op.hpp"
#include "constant.hpp"
#include "mjd.hpp"

// constructor
ProcessData::ProcessData (size_t ss) {
  bitstat_points = 100000;
  fftsize = 2097152;
  fftsize2 = 2 * fftsize;
  datasize = fftsize + bitstat_points;
  current_size = 0;
  if (ss == 0) samplesize = fftsize + bitstat_points; else samplesize = ss;

  center_freq[0] = 1450.;
  center_freq[1] = 1700.;
  ref_freq[0] = 1450.;
  ref_freq[1] = 1700.;
  bandwidth = 32.;
  band_direct = true; 
  timeres = 15.625;
  dm = 71.0255;

  sefd = 10000.; // sefd = 10 Jy at L-band - in mJy!!!!

  //period = 1557785.9; // period of 1937+21 in ns at UTC 2007/02/02  20:30:00
  period = 3115571.8; // doubled pulsar period in ns
  cperiod = (unsigned long)rint(period/timeres);

  threshold = 21.; // threshold in sigma

  av = 250;
  if (av >= cperiod) av = 1;

  Niter = 0;
  number_of_periods = 0;
  curNperiods = 0;
  is_extra_sample = false;

  for (int i=0; i<8; ++i) {
   dataunit[i] = new signed char[datasize];
   if (!dataunit[i]) { perror("new dataunit[i]"); exit (1); }
   memset (dataunit[i], 0, datasize);
  }

  for (int i=0; i<2; ++i) {
   databitcor[i] = new float[datasize];
   if (!databitcor[i]) { perror("new databitcor[i]"); exit (1); }
  }

  part1 = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * fftsize);
  dest = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * fftsize2);
  if (!part1 || !dest) { perror("fftwf_malloc"); exit(1); }

  if (!is_wisdom) {
   plan1 = fftwf_plan_dft_1d (fftsize, part1, part1, FFTW_FORWARD, FFTW_ESTIMATE);
   backplan = fftwf_plan_dft_1d (fftsize2, dest, dest, FFTW_BACKWARD, FFTW_ESTIMATE);
  } else {
   plan1 = fftwf_plan_dft_1d (fftsize, part1, part1, FFTW_FORWARD, FFTW_EXHAUSTIVE);
   backplan = fftwf_plan_dft_1d (fftsize2, dest, dest, FFTW_BACKWARD, FFTW_EXHAUSTIVE);
  }

  // get matrix
  matrix1 = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * (fftsize + 1));
  matrix2 = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * (fftsize + 1));
  if (!matrix1 || !matrix2) { perror("fftwf_malloc"); exit (1); }
  memset (matrix1, 0, (fftsize + 1)*sizeof(fftwf_complex));
  memset (matrix2, 0, (fftsize + 1)*sizeof(fftwf_complex));
  matrix1 = get_matrix (matrix1, fftsize2, center_freq[0], ref_freq[0], bandwidth, band_direct, timeres, dm);
  matrix2 = get_matrix (matrix2, fftsize2, center_freq[1], ref_freq[1], bandwidth, band_direct, timeres, dm);

  // calc spoiled points
  begpoints = get_start_spoiled_points();
  endpoints = get_end_spoiled_points();

//  shift = begpoints;
  shift = 0;

  det = new float[fftsize2 - begpoints - endpoints];
  if (!det) { perror("new det"); exit (1); }

  profile = new float*[4];
  if (!profile) { perror("new profile"); exit (1); }
  for (int i=0; i<4; ++i) {
   profile[i] = new float[cperiod];
   if (!profile[i]) { perror("new profile[i]"); exit (1); }
   memset (profile[i], 0, cperiod * sizeof (float));
  }
  y = new float*[4];
  if (!y) { perror("new y"); exit (1); }
  for (int i=0; i<4; ++i) {
   y[i] = new float[cperiod/av + 1];
   if (!y[i]) { perror("new y[i]"); exit(1); }
   memset (y[i], 0, (cperiod/av + 1) * sizeof(float));
  }

  mjd = new char[255];
  if (!mjd) { perror("new mjd"); exit (1); }
  memset (mjd, 0, 255);

  mjdGP = new char[255];
  if (!mjdGP) { perror("new mjd"); exit (1); }
  memset (mjdGP, 0, 255);

  date = new char[255];
  if (!date) { perror("new date"); exit (1); }
  memset (date, 0, 255);

  time = new char[255];
  if (!time) { perror("new time"); exit (1); }
  memset (time, 0, 255);

  profnames = new char*[4];
  if (!profnames) { perror("new profnames"); exit (1); }
  for (int i=0; i<4; ++i) {
   profnames[i] = new char[255];
   if (!profnames[i]) { perror("new profnames[i]"); exit (1); }
   memset (profnames[i], 0, 255);
  }
  basename = new char[255];
  if (!basename) { perror("new basename"); exit (1); }
  memset (basename, 0, 255);

  x = new float[cperiod/av];
  if (!x) { perror("new x"); exit (1); }
  memset (x, 0, (cperiod/av) * sizeof(float));
}

// destructor
ProcessData::~ProcessData() {
 for (int i=0; i<8; ++i) delete (dataunit[i]);
 for (int i=0; i<2; ++i) delete (databitcor[i]);
 delete (det);
 for (int i=0; i<4; ++i) delete (profile[i]);
 delete (profile);
 for (int i=0; i<4; ++i) delete (y[i]);
 delete (y);
 delete (mjd);
 delete (mjdGP);
 delete (date);
 delete (time);
 for (int i=0; i<4; ++i) delete (profnames[i]);
 delete (profnames);
 delete (basename);
 delete (x);

 fftwf_destroy_plan (plan1);
 fftwf_destroy_plan (backplan);
 fftwf_free (matrix1);
 fftwf_free (matrix2);
 fftwf_free (part1);
 fftwf_free (dest);
}

// update sample data
void ProcessData::update (signed char **array, size_t current_frame, char *startmjd) {

  if (number_of_periods == 0 && current_size == 0) {
   mjd = getMJD (startmjd, begpoints, timeres, 5, mjd);
   date = getDate (atof(mjd), date);
   time = getTime (getSeconds(mjd), time);

   sprintf (basename, "%s/%s.%s.ps/cps", outputpath, outname, "prof");
   for (int u=0; u<4; ++u) 
    sprintf (profnames[u], "%s/%s.%s.%s", outputpath, outname, "prof", u==0 ? "AE" : (u==1 ? "CG" : (u==2 ? "BF" : "DH")));

   FILE *gg;
   gg = fopen ("gplog", "wt");
   fprintf (gg, "#ch     lev        mean      sigma           Nperiod        left pos          globpos               mjd             offset (sec)   Nframe\n");
   fclose (gg);
  }

  off_t processed_size, left_to_process;
  if (current_size + samplesize > datasize) left_to_process = current_size + samplesize - datasize; else left_to_process = 0;
  processed_size = samplesize - left_to_process;
  for (int i=0; i<8; ++i) {
   for (off_t j=0, k=current_size; j<processed_size; ++j, ++k) dataunit[i][k] = array[i][j];
  } // i
  current_size += processed_size;

  if (current_size == datasize) {
   // processing is starting
   // this code will be optimized for our current configuration in the test obs.
   // but I have to make it flexible later
   // I need to join channels A(0) and E(4) , C(2) and G(6) , B(1) and F(5) , D(3) and H(7)

   curNperiods = (unsigned long)floor((timeres*(fftsize2 - begpoints - (is_extra_sample ? 1 : 0) - endpoints)) / period);

   /*----   A and E ------*/ get_profile (dataunit[4], dataunit[0], matrix2, profile[0], profnames[0], number_of_periods == 0 ? true : false);
   /*----   C and G ------*/ get_profile (dataunit[6], dataunit[2], matrix2, profile[1], profnames[1], number_of_periods == 0 ? true : false);
   /*----   B and F ------*/ get_profile (dataunit[5], dataunit[1], matrix1, profile[2], profnames[2], number_of_periods == 0 ? true : false);
   /*----   D and H ------*/ get_profile (dataunit[7], dataunit[3], matrix1, profile[3], profnames[3], number_of_periods == 0 ? true : false);

   number_of_periods += curNperiods;
   /* average */ for (int i=0; i<4; ++i) avg (profile[i], y[i], cperiod, av);
   plot_profile(x, y, av, basename, mjd, date, time, period, cperiod, number_of_periods, timeres, current_frame+1, sefd);

   off_t rightpos = (off_t)rint( (number_of_periods * period)/timeres) - shift + (is_extra_sample ? 1 : 0);
   off_t rightpos2 = rightpos/2;

   shift += (rightpos - (is_extra_sample ? 1 : 0)); // pos for the next iter cycle
   Niter++;
   if (rightpos%2 != 0) is_extra_sample = true; else is_extra_sample = false;

   for (int i=0; i<8; ++i) {
    for (off_t j=rightpos2, k=0; j<datasize; ++j, ++k) dataunit[i][k] = dataunit[i][j];
    for (off_t j=processed_size, k=datasize-rightpos2; j<samplesize; ++j, ++k) dataunit[i][k] = array[i][j];
   } // i
   current_size = (datasize - rightpos2 + left_to_process);
  } // if (current_size == datasize)
}

// get power of 2 size
size_t ProcessData::resize (size_t sz) { 
 long n = 0;
 sz += 1; 
 while ( (1<<(++n)) < sz) ;
 return (1<<(--n));
}

/* ship FFT data out in two complex arrays */
/* part1 originally has packed completely, part2 is empty */
void ProcessData::ship_out2complex (fftwf_complex *part1, fftwf_complex *part2, off_t size) {
 float rep, rem, aip, aim;

 part2[0][0] = part1[0][1];
 part1[0][1] = part2[0][1] = 0.;

 for (off_t j=1; j<size/2+1; ++j) {
  rep = 0.5 * (part1[j][0] + part1[size-j][0]);
  rem = 0.5 * (part1[j][0] - part1[size-j][0]);
  aip = 0.5 * (part1[j][1] + part1[size-j][1]);
  aim = 0.5 * (part1[j][1] - part1[size-j][1]);
  part1[j][0] = rep;
  part1[j][1] = aim;
  part1[size-j][0] = rep;
  part1[size-j][1] = -aim;
  part2[j][0] = aip;
  part2[j][1] = -rem;
  part2[size-j][0] = aip;
  part2[size-j][1] = rem;
 }
}

/* ship FFT data from the GBT to arrange in total band */
/* dest is 2*n size, part1 is n size */
fftwf_complex* ProcessData::ship_out2gbt_total (fftwf_complex *dest, fftwf_complex *part1, off_t size) {
 fftwf_complex *temp;
 temp = (fftwf_complex *)fftwf_malloc (sizeof(fftwf_complex) * size);
 if (!temp) { perror ("fftwf_malloc"); exit (1); }
 memset (temp, 0, sizeof(fftwf_complex) * size);

 /* part1 has A-channel in Re, and E-channel in Im */
 ship_out2complex (part1, temp, size);
 /* now part1 has A-channel and temp has E-channel */
 // -- THIS WAS CORRECT ------
 for (off_t i=size/2+1, j=0; i<size; ++i, ++j) { dest[j][0] = part1[i][0]; dest[j][1] = part1[i][1]; }
 for (off_t i=0, j=size/2-1; i<size/2+1; ++i, ++j) { dest[j][0] = temp[i][0]; dest[j][1] = temp[i][1]; }
 
 /* this is for checking, consider A - direct and E - inverse */
// for (off_t i=0, j=0; i<size/2+1; ++i, ++j) { dest[j][0] = part1[i][0]; dest[j][1] = part1[i][1]; }
// for (off_t i=size/2+1, j=size/2+1; i<size; ++i, ++j) { dest[j][0] = temp[i][0]; dest[j][1] = temp[i][1]; }

 fftwf_free (temp);
 return dest;
}

/* calculate phase due to dispersion */
double ProcessData::get_disp_phase (double f, double f0, double cf, double rf, double bw, double DM) {
 double z, index = 1.e6;

 if (rf >= cf - bw/2. && rf <= cf + bw/2.) {
   z = ( ( ALPHA1 * DM ) / f / ( f0 * f0 * index ) ) * ( f - f0 ) * ( f - f0 );
  } else {
   z = ( ( ALPHA1 * DM ) / f / ( f0 * f0 * index ) ) * ( f - f0 ) * ( f - f0 ) - ( ( ALPHA1 * DM ) / index ) * f * ( 1./( f0 * f0 ) - 1./( rf * rf ) );
  }
 return z;
}

/* compute dedispersion matrix */
fftwf_complex* ProcessData::get_matrix (fftwf_complex *matrix, off_t size, double cf, double rf, double bw, bool is_dir, double tres, double DM) {
 double z = 0.;
 double T, freq, f0;
 float sign = (is_dir ? -1. : 1. ); // direct band = -1, inverse band = 1
 T = (tres * ((double)size))/((double)(1.e3));
 f0 = (is_dir ? cf - bw/2. : cf + bw/2.);

 for (off_t i=0; i < size/2 + 1; ++i) {
   freq = f0 + (-1. * sign)*(((double)i)/T);
   z = get_disp_phase (freq, f0, cf, rf, bw, DM);
   matrix[i][0] = (float)cos(z);
   matrix[i][1] = sign * (float)sin(z);
  }
 return matrix;
}

/* get total number of spoiled points */
/* f1 - lower frequency, f2 - upper frequency */
/* to get spoiled points at the beginning: f1 = ref freq, f2 = highest freq of the band */
/* to get spoiled points at the end: f1 = lowest freq of the band, f2 = ref freq */
unsigned long ProcessData::get_spoiled_points (double f1, double f2) {
 double tbad, tdelay;
 tbad = dm * ALPHA2 * ( 1./(f1 * f1) - 1./(f2 * f2) );
 tdelay = (tbad * 0.001)/timeres;
 return (unsigned long)ceil(fabs(tdelay));
}

/* get number of spoiled points in the beginning */
unsigned long ProcessData::get_start_spoiled_points () {
 double tbad, tdelay;
 double f1 = ref_freq[0]; // we take 1450. because the number of spoiled points is larger at lower freqs 
                          // and we want to have both profiles at 1450 and 1700 we the same ref start time 
 double f2 = center_freq[0] + bandwidth/2.;
 tbad = dm * ALPHA2 * ( 1./(f1 * f1) - 1./(f2 * f2) );
 tdelay = (tbad * 0.001)/timeres;
 return (unsigned long)ceil(fabs(tdelay));
}

/* get number of spoiled points in the end */
unsigned long ProcessData::get_end_spoiled_points () {
 double tbad, tdelay;
 double f1 = center_freq[0] - bandwidth/2.;
 double f2 = ref_freq[0];
 tbad = dm * ALPHA2 * ( 1./(f1 * f1) - 1./(f2 * f2) );
 tdelay = (tbad * 0.001)/timeres;
 return (unsigned long)ceil(fabs(tdelay));
}

/* making the dedispersion */
fftwf_complex* ProcessData::dedispersion (fftwf_complex *ampl_sp, fftwf_complex *matrix, fftwf_plan plan, off_t size, bool do_fraction, bool do_zero) {
 float re;
 if (!do_fraction) {
  for (off_t i=0; i<size/2 + 1; ++i) {
   re = ampl_sp[i][0];
   ampl_sp[i][0] = 2. * (re * matrix[i][0] - matrix[i][1] * ampl_sp[i][1]);
   ampl_sp[i][1] = 2. * (re * matrix[i][1] + matrix[i][0] * ampl_sp[i][1]);
  }
 } else { // do fraction
  float im;
  double z;
  for (off_t i=0; i<size/2 + 1; ++i) {
   z = ((-0.50 * M_PI) * ((double)i))/((double)size);
   re = ampl_sp[i][0] * cos(z) - ampl_sp[i][1] * sin(z);
   im = ampl_sp[i][0] * sin(z) + ampl_sp[i][1] * cos(z);
   ampl_sp[i][0] = 2. * (re * matrix[i][0] - matrix[i][1] * im);
   ampl_sp[i][1] = 2. * (re * matrix[i][1] + matrix[i][0] * im);
  }
 }
 if (do_zero) for (off_t i=size/2 + 1; i<size; ++i) { ampl_sp[i][0] = 0.; ampl_sp[i][1] = 0.; }
 fftwf_execute (plan);
 return ampl_sp;
}

// run a number of functions to get a profile
void ProcessData::get_profile (signed char *first, signed char *second, fftwf_complex *matrix, float *prof, char *name, bool flag) {

 memset (part1, 0, fftsize * sizeof(fftwf_complex));
 memset (dest, 0, fftsize2 * sizeof(fftwf_complex));

 memset (databitcor[0], 0, datasize * sizeof(float));
 bits_correction_end (first, databitcor[0], datasize, bitstat_points);
 remove_level (databitcor[0], datasize, get_level(databitcor[0], datasize));
 for (off_t i=0, j=datasize-fftsize; i<fftsize; ++i, ++j) part1[i][0] = databitcor[0][j];

 memset (databitcor[1], 0, datasize * sizeof(float));
 bits_correction_end (second, databitcor[1], datasize, bitstat_points);
 remove_level (databitcor[1], datasize, get_level(databitcor[1], datasize));
 for (off_t i=0, j=datasize-fftsize; i<fftsize; ++i, ++j) part1[i][1] = databitcor[1][j];

 fftwf_execute (plan1);
 ship_out2gbt_total (dest, part1, fftsize);
 dest = dedispersion (dest, matrix, backplan, fftsize2, false, false);

 memset (det, 0, (fftsize2 - begpoints - endpoints) * sizeof (float));
 for (off_t i=begpoints + (is_extra_sample ? 1 : 0), j=0; i<fftsize2 - endpoints; ++i, ++j) 
  det[j] = (dest[i][0] / fftsize2) * (dest[i][0] / fftsize2) + (dest[i][1] / fftsize2) * (dest[i][1] / fftsize2);
 
/*
 if (flag) {
  int tf;
  if ((tf = open ("test.det", O_CREAT | O_APPEND | O_WRONLY | O_LARGEFILE, 0644)) == -1) { perror("open"); exit(1); }
  write (tf, det, (unsigned long)(((off_t)rint( ((number_of_periods + curNperiods) * period)/timeres) - shift)) * sizeof (float));
  close (tf);
 }
 */

 /* to search for GPs */
 searchGP (det, fftsize2 - begpoints - endpoints - (is_extra_sample ? 1 : 0), name);
 /* --- */
 for (off_t j=0; j<cperiod; ++j) prof[j] *= number_of_periods;
 off_t leftpos;
 for (off_t i=0; i<curNperiods; ++i) {
  leftpos = (off_t)rint( ((number_of_periods + i) * period)/timeres) - shift;
  for (off_t j=0, k=leftpos; j<cperiod; ++j, ++k) prof[j] += det[k];
 }
 unsigned long np = number_of_periods + curNperiods;
 for (off_t j=0; j<cperiod; ++j) prof[j] /= np;

 int out;
#ifdef apple
 if ((out = open (name, O_CREAT | O_TRUNC | O_WRONLY, 0644)) == -1) { perror("open"); exit(1); }
#else
 if ((out = open (name, O_CREAT | O_TRUNC | O_WRONLY | O_LARGEFILE, 0644)) == -1) { perror("open"); exit(1); }
#endif
 write (out, mjd, mjd_header);
 write (out, &np, sizeof(unsigned long));
 write (out, prof, cperiod * sizeof(float));
 close (out);
}

// average the array
void ProcessData::avg (float *profile, float *y, unsigned long size, long aver) {
  float sum;
  for (int i=0,j,k=0; i<size; i+=aver, ++k) {
   sum = 0.;
   for (j=i; j<i+aver; ++j) sum += profile[j];
   y[k] = sum / ((float)(j-i));
  }
}

/* to search for GPs */
void ProcessData::searchGP (float *x, unsigned long size, char *name) {
 double mean, sig, ss = 0., ss2 = 0., lev;
 unsigned long lp, lp1, Np, Np1;
 off_t globpos;
 double timediff;
 unsigned long Nfr;
 FILE *gplog;
 gplog = fopen ("gplog", "at");
 for (unsigned long i=0; i<size; ++i) {
  ss += x[i];
  ss2 += x[i] * x[i];
 } //i
 mean = ss / size;
 sig = sqrt ( ss2 / size - mean * mean );
 for (unsigned long i=0; i<size; ++i) {
  lev = (x[i] - mean) / sig;
  if (lev >= threshold) {
   // calc necessary info and print out to the file
   Np = (unsigned long)floor((double)i/cperiod);
   lp = (unsigned long)rint( (Np * period)/timeres); 
   lp = i - lp; // pos in the period
   Np += number_of_periods; // number of the period
   globpos = shift + i;
   Np1 = (unsigned long)floor((((double)(globpos))*timeres)/period);
   lp1 = (unsigned long)rint((double)(globpos) - ((double)(Np1 * period)/timeres));
   memset (mjdGP, 0, 255);
   mjdGP = getMJD (mjd, globpos, timeres, 5, mjdGP);
   timediff = getTimeDiff (mjdGP, mjd);
   Nfr = (unsigned long)floor(timediff/sec_step);
   fprintf (gplog, "%c%c   %lf   %lf   %lf   %6ld (%6ld)   %6ld (%6ld)   %10lld   %-27s   %lf   %6ld\n", name[(int)strlen(name)-2], name[(int)strlen(name)-1], lev, mean, sig, Np, Np1, lp, lp1, (unsigned long long)globpos, mjdGP, timediff, Nfr);
  }
 } //i
 fclose (gplog);
}

