#include "hdsd.hpp"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

extern bool is_check;   // true when check mode is set
Error er;  // error global object
static ErrorInfo errors;          // errors of all the frames

// initialize freq_id_table
// this table keeps correspondence between freq_id & freq
const float FrameInfo::freq_id_table[16] = { 0.04, 0.1, 0.2, 0.5, 1., 2., 4., 8., 16., 32., 64., 128., 256., 512., 1024., 2048. };


// sign & magn bit correspondence
const signed char SampleData::vlba[4] = { -3, -1, 1, 3 };
// Codings:
// --------------------------------------------
// VLBA |  sign | magn       AT |  sign | magn
// --------------------------------------------
// +3       1      1         +3     0      1
// +1       1      0         +1     0      0
// -1       0      1         -1     1      0
// -3       0      0         -3     1      1


// constructor
FrameData::FrameData(size_t fs, bool swap) {
  size_of_long = sizeof (long);
  size_of_long_long = sizeof (long long);
  frame_number = 0;
  framesize = fs;
  streamsize = framesize - 8; // 8 - header size in bytes
  to_reverse = swap;

  raw = new unsigned long long[framesize/size_of_long_long];
  if (!raw) { perror("new raw"); exit(1); }
  memset (raw, 0, framesize);
}

// destructor
FrameData::~FrameData() {
 delete (raw);
}

// update frame with new data
void FrameData::update (unsigned long long * src, size_t current_frame) {
  frame_number = current_frame;
  
  memset (raw, 0, framesize);
  raw = (unsigned long long *)memcpy (raw, src, framesize); 
  // reverse bits
  if (to_reverse) for (int i=0; i<framesize/size_of_long_long; ++i) raw[i] = swapBytes (raw[i]);
  
   try {
    info.setup(raw[0], getFrameNumber());
   } // try
   catch (Error er) {
    if (!is_check) {
     printf ("%s\nCheck your data with --check option!\n", er.getMsg());
     exit (1);
    }
    else {
     switch (er.getType()) {
      case Error::SYNC: errors.addSyncMismatch(); break;
      case Error::SYNC2: errors.addSync2Mismatch(); break;
      case Error::CONTINOUS: errors.addTimeInconsistence(); break;
      case Error::FIELDS: errors.addFieldsNotTheSame(); break;
      default: break;
     } // switch
    }
   }
   catch (...) {
    printf ("Unknown exception!\n");
    exit (1);
   }
}

// return seconds
unsigned long FrameData::getSeconds () {
 return info.getSeconds();
}

// print header info
void FrameData::getHeader () {
 info.getHeader();
}

// this function swap bytes in 64 bits word
unsigned long long FrameData::swapBytes (unsigned long long word) {
 int i;
 unsigned long long nVal = 0;
 for (i=7; i>=0; i--) { nVal |= (word & 255) << 8*i; word >>= 8; }
 return nVal;
}

// print all check info
void FrameData::getCheckInfo () {
 printf ("\n");
 printf ("Syncword mismatches: %ld\n", errors.getSyncMismatch());
 printf ("Sync2 pattern mismatches: %ld\n", errors.getSync2Mismatch());
 printf ("Time continuity mismatches: %ld\n", errors.getTimeInconsistence());
 printf ("Fields differencies events: %ld\n", errors.getFieldsNotTheSame());
}

// get the frame number
size_t FrameData::getFrameNumber () {
 return frame_number;
}



// constructor
SaveInfo::SaveInfo () {
}

// destructor
SaveInfo::~SaveInfo () {
}

// return seconds record
unsigned long SaveInfo::getSeconds () {
 return seconds;
}

// return Nch_id record
unsigned char SaveInfo::getNch () {
 return Nch_id;
}

// return freq_id record
unsigned char SaveInfo::getFreq () {
 return freq_id;
}

// return Nbits_id record
unsigned char SaveInfo::getNbits () {
 return Nbits_id;
}

// return frame number
size_t SaveInfo::getFrameNumber () {
 return frame_number;
}

// set Nch_id
void SaveInfo::setNch (unsigned char src) {
 Nch_id = src;
}

// set freq_id
void SaveInfo::setFreq (unsigned char src) {
 freq_id = src;
}

// set Nbits_id
void SaveInfo::setNbits (unsigned char src) {
 Nbits_id = src;
}

// set seconds
void SaveInfo::setSeconds (unsigned long src) {
 seconds = src;
}

// set frame number
void SaveInfo::setFrameNumber (size_t frame) {
 frame_number = frame;
}






// constructor
FrameInfo::FrameInfo () {
 initialized = false;
}

// destructor
FrameInfo::~FrameInfo () {
}

// initialize class data from track
void FrameInfo::setup (unsigned long long fr, size_t frame) {
  frame_number = frame;
  if (frame_number == 0) initialized = false;
  syncword = (unsigned long)(fr & 0x00000000ffffffffl);
  if (syncword == sync_mask) is_syncword_ok = true; else is_syncword_ok = false;
  hd2 = (unsigned long)(fr >> 32);
  seconds = hd2 & 0x0001ffffl;
  Nch_id = (hd2 >> 17) & 0x0001;
  if (Nch_id) Nch = 4; else Nch = 1;
  freq_id = (hd2 >> 18) & 0x000f;
  freq = freq_id_table[freq_id];
  Nbits_id = (hd2 >> 22) & 0x0003;
  Nbits = 1 << Nbits_id;
  sync2 = (hd2 >> 24) & 0x00ff;
  if (sync2 == sync2_mask) is_sync2_ok = true; else is_sync2_ok = false;
  
  if (!initialized) { save.setNch (Nch_id); save.setFreq (freq_id); save.setNbits (Nbits_id); save.setSeconds (seconds); initialized = true; save.setFrameNumber(frame_number); }
   else {
    are_fields_same = (save.getNch() == Nch_id ? true : false);
    are_fields_same = (save.getFreq() == freq_id ? true : false);
    are_fields_same = (save.getNbits() == Nbits_id ? true : false);
    // check continuous of time
    if ( seconds - save.getSeconds() != frame_number-save.getFrameNumber() ) is_time_continous = false;
     else { is_time_continous = true; save.setFrameNumber(frame_number); }
    save.setNch (Nch_id); save.setFreq (freq_id); save.setNbits (Nbits_id); save.setSeconds (seconds);

    // check the info is ok, if not exit
    if (!is_syncword_ok && is_check) { er.setType(Error::SYNC); er.setMsg("\nSyncword mismatch!"); throw er; }
    if (!is_sync2_ok && is_check) { er.setType(Error::SYNC2); er.setMsg("\nSync2 pattern %x mismatch [should be %x]!", sync2, sync2_mask); throw er; }
    if (!is_time_continous) { er.setType(Error::CONTINOUS); er.setMsg("\nTime is not continuous!"); throw er; }
    if (!are_fields_same && is_check) { er.setType(Error::FIELDS); er.setMsg("\nSome header fields are different for different frames (not time :))!"); throw er; }
   }
}

// print header info
void FrameInfo::getHeader () {
  printf (" secs = %ld  Nch: %d  freq: %g MHz  Bits: %d  Sync2 = %x [%s]\n", seconds, Nch, freq, Nbits, sync2, is_sync2_ok ? "OK" : "Mismatch");
}

// return seconds record
unsigned long FrameInfo::getSeconds () {
 return seconds; 
}





// constructor
SampleData::SampleData(size_t fs, bool swap) : FrameData (fs, swap) {
  samplesize = 4*streamsize;
  for (int j=0; j<1; ++j) data[j] = new signed char[samplesize];
}

// destructor
SampleData::~SampleData() {
  for (int j=0; j<1; ++j) delete (data[j]); 
}

// update sample data
void SampleData::update (unsigned long long *src, size_t current_frame) {

  FrameData::update (src, current_frame);
  int samplecounter;

//   for (int b=0; b<1; ++b) { // cycle on channels
    samplecounter = 0;
    for (long i=0; i<streamsize/8; ++i) { // cycle on long-long words
     for (long j=0; j<32; ++j) { // cycle on different time counts within long-long word
      data[/*b*/0][samplecounter++] = vlba[(raw[i] >> (/*2*b+*/j*2)) & 3];
     } // j
    } // i
//   } // b
}

// return the size of the sample
size_t SampleData::getSampleSize () {
  return samplesize;
}

// return pointer to the W channel
signed char* SampleData::getW () {
  return data[0];
}

// return pointer to the channel by its number from chan0 (W) till chan3 (Z)
signed char* SampleData::getChannel (int ch) {
  if (ch < 0 || ch > 0) {
   printf ("Bad channel number!\n");
   exit (1);
  }
  return data[ch];	
}







// constructor
Error::Error () {
 msg = new char[255];
 if (!msg) { printf("Not enough memory!\n"); exit (1); }
 memset (msg, 0, 255);
};

// another constructor
Error::Error (char * src) {
 msg = new char[255];
 if (!msg) { printf("Not enough memory!\n"); exit (1); }
 memset (msg, 0, 255);
 strcpy (msg, src);
};

// constructor with template and 2 arguments
Error::Error (char * mask, unsigned char arg1, unsigned char arg2) {
 msg = new char[255];
 if (!msg) { printf("Not enough memory!\n"); exit (1); }
 memset (msg, 0, 255);
 sprintf (msg, mask, arg1, arg2);
}

// destructor
Error::~Error () {
 msg = 0;
}

// get message
char * Error::getMsg () {
 return msg;
}

// set message
void Error::setMsg (char * src) {
 memset (msg, 0, 255);
 strcpy (msg, src);
}

// set message by template and two arguments
void Error::setMsg (char * mask, unsigned char arg1, unsigned char arg2) {
 memset (msg, 0, 255);
 sprintf (msg, mask, arg1, arg2);
}

// get type of the error
Error::ErrorType Error::getType () {
 return type;
}

// set type of the error
void Error::setType (Error::ErrorType src) {
 type = src;
}

