#include "hdsd.hpp"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

extern bool is_check;   // true when check mode is set
extern bool is_ignore; // true if one ignores the time-continuity error
extern bool is_to_search_sync; // true if we search for the sync word after every read frame
Error er;  // error global object
static ErrorInfo errors;          // errors of all the frames

// initialize freq_id_table
// this table keeps correspondence between freq_id & freq
const float FrameInfo::freq_id_table[16] = { 0.04, 0.1, 0.2, 0.5, 1., 2., 4., 8., 16., 32., 64., 128., 256., 512., 1024., 2048. };


// sign & magn bit correspondence
////const signed char SampleData::vlba[4] = { -3, -1, 1, 3 };
// Codings:
// --------------------------------------------
// VLBA |  sign | magn       AT |  sign | magn
// --------------------------------------------
// +3       1      1         +3     0      1
// +1       1      0         +1     0      0
// -1       0      1         -1     1      0
// -3       0      0         -3     1      1
const unsigned char SampleData::coding[4] = { 1, 3, 15, 255 };


// constructor
FrameData::FrameData(size_t fs, bool swap) {
  size_of_long = sizeof (long);
  size_of_long_long = sizeof (long long);
  frame_number = 0;
  framesize = fs;
  streamsize = framesize - 8; // 8 - header size in bytes
  to_reverse = swap;

  raw = new unsigned long long[framesize/size_of_long_long];
  if (!raw) { perror("new raw"); exit(1); }
  memset (raw, 0, framesize);
}

// destructor
FrameData::~FrameData() {
 delete (raw);
}

// update frame with new data
void FrameData::update (unsigned long long * src, size_t current_frame) {
  frame_number = current_frame;
  
  memset (raw, 0, framesize);
  raw = (unsigned long long *)memcpy (raw, src, framesize); 
  // reverse bits
  if (to_reverse) for (int i=0; i<framesize/size_of_long_long; ++i) raw[i] = swapBytes (raw[i]);
  
   try {
    info.setup(raw[0], getFrameNumber());
   } // try
   catch (Error er) {
    if (!is_check) {
     printf ("\n%s\nCheck your data with --check option!\n", er.getMsg());
     exit (1);
    }
    else {
     if (er.getType() & Error::SYNC) errors.addSyncMismatch();
     if (er.getType() & Error::SYNC2) errors.addSync2Mismatch();
     if (er.getType() & Error::CONTINOUS) errors.addTimeInconsistence();
     if (er.getType() & Error::FIELDS) errors.addFieldsNotTheSame();
    }
   }
   catch (...) {
    printf ("Unknown exception!\n");
    exit (1);
   }
}

// return seconds
unsigned long FrameData::getSeconds () {
 return info.getSeconds();
}

// print header info
void FrameData::getHeader () {
 info.getHeader();
}

// this function swap bytes in 64 bits word
unsigned long long FrameData::swapBytes (unsigned long long word) {
 int i;
 unsigned long long nVal = 0;
 for (i=7; i>=0; i--) { nVal |= (word & 255) << 8*i; word >>= 8; }
 return nVal;
}

// print all check info
void FrameData::getCheckInfo () {
 printf ("\n");
 printf ("Syncword mismatches: %ld\n", errors.getSyncMismatch());
 printf ("Sync2 pattern mismatches: %ld\n", errors.getSync2Mismatch());
 printf ("Fields differencies events: %ld\n", errors.getFieldsNotTheSame());
 printf ("Time continuity mismatches: %ld\n", errors.getTimeInconsistence());
}

// print all check info in simple way
void FrameData::getSimpleCheckInfo () {
 printf ("      [%4ld  %4ld  %4ld  %4ld]", errors.getSyncMismatch(), errors.getSync2Mismatch(), errors.getFieldsNotTheSame(), errors.getTimeInconsistence());
}

// get the frame number
size_t FrameData::getFrameNumber () {
 return frame_number;
}

// get number of channels
unsigned char FrameData::getNchan () {
 return info.getNchan();
}

// get number of bits
unsigned char FrameData::getNbits () {
 return info.getNbits();
}

// get id of number of quantization
unsigned char FrameData::getNbits_id () {
 return info.getNbits_id ();
}




// constructor
SaveInfo::SaveInfo () {
}

// destructor
SaveInfo::~SaveInfo () {
}

// return seconds record
unsigned long SaveInfo::getSeconds () {
 return seconds;
}

// return Nch_id record
unsigned char SaveInfo::getNch () {
 return Nch_id;
}

// return freq_id record
unsigned char SaveInfo::getFreq () {
 return freq_id;
}

// return Nbits_id record
unsigned char SaveInfo::getNbits () {
 return Nbits_id;
}

// return frame number
size_t SaveInfo::getFrameNumber () {
 return frame_number;
}

// set Nch_id
void SaveInfo::setNch (unsigned char src) {
 Nch_id = src;
}

// set freq_id
void SaveInfo::setFreq (unsigned char src) {
 freq_id = src;
}

// set Nbits_id
void SaveInfo::setNbits (unsigned char src) {
 Nbits_id = src;
}

// set seconds
void SaveInfo::setSeconds (unsigned long src) {
 seconds = src;
}

// set frame number
void SaveInfo::setFrameNumber (size_t frame) {
 frame_number = frame;
}






// constructor
FrameInfo::FrameInfo () {
 initialized = false;
}

// destructor
FrameInfo::~FrameInfo () {
}

// initialize class data from track
void FrameInfo::setup (unsigned long long fr, size_t frame) {
  frame_number = frame;
  if (frame_number == 0) initialized = false;
  syncword = (unsigned long)(fr & 0x00000000ffffffffl);
  if (syncword == sync_mask) is_syncword_ok = true; else is_syncword_ok = false;
  hd2 = (unsigned long)(fr >> 32);
  seconds = hd2 & 0x0001ffffl;
  Nch_id = (hd2 >> 17) & 0x0001;
  if (Nch_id) Nch = 4; else Nch = 1;
  freq_id = (hd2 >> 18) & 0x000f;
  freq = freq_id_table[freq_id];
  Nbits_id = (hd2 >> 22) & 0x0003;
  Nbits = 1 << Nbits_id;
  sync2 = (hd2 >> 24) & 0x00ff;
  if (sync2 == sync2_mask) is_sync2_ok = true; else is_sync2_ok = false;
  
  if (!initialized) { save.setNch (Nch_id); save.setFreq (freq_id); save.setNbits (Nbits_id); save.setSeconds (seconds); initialized = true; save.setFrameNumber(frame_number); }
   else {
    are_fields_same = (save.getNch() == Nch_id ? true : false);
    are_fields_same &= (save.getFreq() == freq_id ? true : false);
    are_fields_same &= (save.getNbits() == Nbits_id ? true : false);
    // check continuous of time
    if ( seconds - save.getSeconds() != frame_number-save.getFrameNumber() ) {
      if ( seconds + 86400 - save.getSeconds() != frame_number-save.getFrameNumber() ) is_time_continous = false; else is_time_continous = true;
    } else is_time_continous = true;

    if (is_time_continous || (!is_time_continous && is_to_search_sync)) { save.setFrameNumber(frame_number); save.setSeconds (seconds); }

    // dont't update the Nch, Nbits and Freq because we want to compare with the first frame
    // save.setNch (Nch_id); save.setFreq (freq_id); save.setNbits (Nbits_id); 

    er.clear (); // clear the error info from the previous frame
    // check the info is ok, if not exit
    if (!is_syncword_ok && is_check) { er.setType(Error::SYNC); er.setMsg("Syncword mismatch!"); }
    if (!is_sync2_ok && is_check) { er.setType(Error::SYNC2); er.setMsg("Sync2 pattern %x mismatch [should be %x]!", sync2, sync2_mask); }
    if (!are_fields_same && is_check) { er.setType(Error::FIELDS); er.setMsg("Some header fields are different for different frames (not time :))!"); }
    if (!is_time_continous) { er.setType(Error::CONTINOUS); er.setMsg("Time is not continuous!"); if (is_check || !is_ignore) throw er; }
     else { if (is_check && (!is_syncword_ok || !is_sync2_ok || !are_fields_same)) throw er; }
   }
}

// print header info
void FrameInfo::getHeader () {
  printf (" secs = %6ld  Nch: %d  freq: %4g MHz  Bits: %d  sync = %x %s  sync2 = %x %s", seconds, Nch, freq, Nbits, syncword, is_syncword_ok ? "[ok]" : "    ", sync2, is_sync2_ok ? "[ok]" : "    ");
}

// return seconds record
unsigned long FrameInfo::getSeconds () {
 return seconds; 
}

// return frame size
size_t FrameInfo::getFrameSize () {
 return ( ( ( ((unsigned long)(freq * Nbits * Nch)) * 1000000) / 8) + 8);
}





// constructor
SampleData::SampleData(size_t fs, unsigned char nch, unsigned char nb, unsigned char id, bool swap) : FrameData (fs, swap) {
  samplesize = (streamsize * 8) / ((unsigned long)nch * (unsigned long)nb); // 8 - bits in byte
  nchan = nch;
  nbits = nb;
  nbits_id = id;
  data[0] = new signed char[samplesize];
  if (nchan == 1) for (int j=1; j<4; ++j) data[j] = NULL;
   else for (int j=1; j<4; ++j) data[j] = new signed char[samplesize];
}

// destructor
SampleData::~SampleData() {
  for (int j=0; j<nchan; ++j) delete (data[j]); 
}

// update sample data
void SampleData::update (unsigned long long *src, size_t current_frame) {

  FrameData::update (src, current_frame);
  int samplecounter;
  unsigned char nbil = coding[nbits_id];
  unsigned char nchnbits = nchan * nbits;
  unsigned char jlimit = (8 * size_of_long_long) / nchnbits; // 8 - bits in byte
  unsigned long ilimit = streamsize/size_of_long_long;

   for (int b=0; b<nchan; ++b) { // cycle on channels
    samplecounter = 0;
    for (unsigned long i=0; i<ilimit; ++i) { // cycle on long-long words
     for (unsigned char j=0; j<jlimit; ++j) { // cycle on different time counts within long-long word
//      data[b][samplecounter++] = vlba[(raw[i] >> (2 * b + 8 * j)) & 3];
      data[b][samplecounter++] = 2 * (((raw[i] >> (nbits * b + nchnbits * j)) & nbil)) - nbil;
     } // j
    } // i
   } // b
}

// get coding
unsigned char SampleData::getCoding (unsigned char nbits) {
 int index = 0;
 if (nbits < 1 || nbits > 8) { perror("get coding"); exit (1); }
 while ((nbits>>=1) !=0) index++;
 if (index < 0 || index > 3) { perror("get coding"); exit (1); }
 return coding[index];
}

// return the size of the sample
size_t SampleData::getSampleSize () {
  return samplesize;
}

// return pointer to the W channel
signed char* SampleData::getW () {
  return data[0];
}

// return pointer to the X channel
signed char* SampleData::getX () {
  return data[1];
}

// return pointer to the Y channel
signed char* SampleData::getY () {
  return data[2];
}

// return pointer to the Z channel
signed char* SampleData::getZ () {
  return data[3];
}

// return pointer to the channel by its number from chan0 (W) till chan3 (Z)
signed char* SampleData::getChannel (int ch) {
  if (ch < 0 || ch > 3) {
   printf ("Bad channel number!\n");
   exit (1);
  }
  return data[ch];	
}







// constructor
Error::Error () {
 msg = new char[255];
 if (!msg) { printf("Not enough memory!\n"); exit (1); }
 memset (msg, 0, 255);
 type = 0;
};

// another constructor
Error::Error (char * src) {
 msg = new char[255];
 if (!msg) { printf("Not enough memory!\n"); exit (1); }
 memset (msg, 0, 255);
 strcpy (msg, src);
 type = 0;
};

// constructor with template and 2 arguments
Error::Error (char * mask, unsigned char arg1, unsigned char arg2) {
 msg = new char[255];
 if (!msg) { printf("Not enough memory!\n"); exit (1); }
 memset (msg, 0, 255);
 sprintf (msg, mask, arg1, arg2);
 type = 0;
}

// destructor
Error::~Error () {
 msg = 0;
}

// get message
char * Error::getMsg () {
 return msg;
}

// set message
void Error::setMsg (char * src) {
 if (!strcmp(msg, "")) strcpy (msg, src);
  else { 
   strcat (msg, " & ");
   strcat (msg, src);
  }
}

// set message by template and two arguments
void Error::setMsg (char * mask, unsigned char arg1, unsigned char arg2) {
 if (!strcmp(msg, "")) sprintf (msg, mask, arg1, arg2);
  else {
   strcat (msg, " & ");
   char *temp;
   temp = new char[255];
   if (!temp) { perror("new"); exit(1); }
   memset (temp, 0, 255);
   sprintf (temp, mask, arg1, arg2);
   strcat (msg, temp);
   delete (temp);
  }
}

// get type of the error
unsigned int Error::getType () {
 return type;
}

// set type of the error
void Error::setType (unsigned int src) {
 type |= src;
}

// clear the error
void Error::clear () {
 memset (msg, 0, 255);
 type = 0;
}

