#ifndef __FRAMEDATA_HPP
#define __FRAMEDATA_HPP

#include <stdio.h>

class FrameInfo;

/*++++++++++ class to count of errors in frames +++++++++++*/
class ErrorInfo {
	unsigned long sync_mismatch_number;   // number of sync mismathes
	unsigned long sync2_mismatch_number;  // -- // --
	unsigned long time_inconsistence_number;  // -- // --
	unsigned long fields_not_the_same_number; // -- // --
  public:
	ErrorInfo() { sync_mismatch_number = 0; sync2_mismatch_number = 0; time_inconsistence_number = 0; fields_not_the_same_number = 0; }
	~ErrorInfo() {}
	unsigned long getSyncMismatch () { return sync_mismatch_number; }
	unsigned long getSync2Mismatch () { return sync2_mismatch_number; } 
	unsigned long getTimeInconsistence () { return time_inconsistence_number; }
	unsigned long getFieldsNotTheSame () { return fields_not_the_same_number; }
	void addSyncMismatch () { sync_mismatch_number++; }
	void addSync2Mismatch () { sync2_mismatch_number++; }
	void addTimeInconsistence () { time_inconsistence_number++; }
	void addFieldsNotTheSame () { fields_not_the_same_number++; }
};

/*+++++++++ class with info that we want to keep to check the correctness of the data ++++++++++++*/
class SaveInfo {
	  unsigned char Nch_id;          // number of channels index (0 for 1 channel; 1 for 4 channels)
	  unsigned char freq_id;         // sampling frequency index
	  unsigned char Nbits_id;        // index of sampling bits number (0,1,2 or 3)
	  unsigned long seconds;         // the same as seconds but from the previous frame
	  size_t frame_number;           // frame number
  public:
          SaveInfo ();   //constructor
          virtual ~SaveInfo ();  // destructor
          unsigned long getSeconds ();  // return seconds record
	  unsigned char getNch (); // return Nch_id 
	  unsigned char getFreq (); // return freq_id
	  unsigned char getNbits (); // return Nbits_id
	  size_t getFrameNumber (); // return frame number
	  void setNch (unsigned char src); // set Nch_id
	  void setFreq (unsigned char src); // set freq_id
	  void setNbits (unsigned char src); // set Nbits_id
	  void setSeconds (unsigned long src); // set seconds
	  void setFrameNumber (size_t frame); // set frame number
};

/*+++++++++  class with header & aux info of frame  ++++++++*/
class FrameInfo {
          unsigned long syncword;        // sync word
	  static const unsigned long sync_mask = 0xffffffff; // all ones in 4 bytes

	  unsigned long hd2;             // second part of the header, HD2
	  unsigned long seconds;         // time of day (17 bits) elapsed from 00h00m00s in seconds
	  unsigned char Nch_id;          // number of channels index (0 for 1 channel; 1 for 4 channels)
	  unsigned char Nch;             // number of channels (1 or 4)
	  unsigned char freq_id;         // sampling frequency index
	  float freq;                    // sampling frequency
	  static const float freq_id_table[16];   // this table keeps correspondence between freq_id & freq
	  unsigned char Nbits_id;        // index of sampling bits number (0,1,2 or 3)
	  unsigned char Nbits;           // sampling bits (1, 2, 4 or 8)

	  unsigned char sync2;           // sync pattern2
	  static const unsigned char sync2_mask = 0x8b; // sync2 pattern mask
	  bool is_sync2_ok;              // true if sync2 is ok

          SaveInfo save;                 // to keep info about previous frame
	  
	  bool is_syncword_ok;  // true if syncword is ok (all units)
	  bool initialized;  // true when there was previous value of all the fields
	  bool is_time_continous;  // true if the time labels in sequent frames are continously increased 
	  bool are_fields_same; // true if all the fields (except for seconds) are the same as in previous frame

	  size_t frame_number;           // number of the current frame
  public:
          FrameInfo ();   //constructor
          virtual ~FrameInfo ();  // destructor
	  void setup (unsigned long long fr, size_t frame);  // initialize class data from frame 
	  void getHeader ();  // print header info
          unsigned long getSeconds ();  // return seconds record
	  size_t getFrameSize (); // return frame size
	  unsigned char getNchan () { return Nch; } // return number of channels
	  unsigned char getNbits () { return Nbits; } // return number of bits
	  unsigned char getNbits_id () { return Nbits_id; } // return index of sampling bits number
	  float getSamplingFreq () { return freq; } // return sampling frequency
};

/*++++++++++  class to work with K5 data  ++++++++*/
class FrameData {
          size_t frame_number;       // number of the current frame
	  FrameInfo info;            // all info of the frame

	  unsigned long long swapBytes (unsigned long long word); // swap bytes in 64-bits word

  protected:
          size_t framesize;                     // size of frame in bytes
	  size_t streamsize;                    // size of data
	  unsigned long long *raw;              // raw frame data for all tracks
	  int size_of_long;                     // size of long type
	  int size_of_long_long;                // size of long long type
	  bool to_reverse;     // reverse (or not) the bits in raw input data
  public:
	  FrameData (size_t fs, bool swap = false);    // constructor
	  virtual ~FrameData ();  // destructor
	  virtual void update (unsigned long long *src, size_t current_frame);  // update frame data
	  unsigned long getSeconds ();  // return seconds
	  size_t getFrameSize() { return framesize; } // return size of frame
	  void getHeader (); // print header info
	  void getCheckInfo (); // print all check info
	  void getSimpleCheckInfo (); // print all check info in simple way
	  size_t getFrameNumber (); // get the frame number
	  unsigned char getNchan (); // get number of channels
	  unsigned char getNbits (); // get number of bits
	  unsigned char getNbits_id (); // get id of number of quantization
};


/*++++++++++++++++++ class to work directly with real data ++++++++++++++++++++*/
class SampleData : public FrameData {

           ////static const signed char vlba[4];
	   // Codings:
	   // --------------------------------------------
	   // VLBA |  sign | magn       AT |  sign | magn
	   // --------------------------------------------
	   // +3       1      1         +3     0      1
	   // +1       1      0         +1     0      0
	   // -1       0      1         -1     1      0
	   // -3       0      0         -3     1      1
	   static const unsigned char coding[4];

   protected:
	   size_t samplesize;  // size of full sampled data for each sideband in bytes
	   signed char *data[4]; // data arrays of individual channels
	   // I have added these three variables because the program must work even there are
	   // some problems with header info to get full statistics of all errors
	   unsigned char nchan;  // number of channels
	   unsigned char nbits;  // number of bits
	   unsigned char nbits_id; // id of Nbits
   public:
	   SampleData(size_t fs, unsigned char nch, unsigned char nb, unsigned char id, bool swap = false);  // constructor
	   virtual ~SampleData();  // destructor
	   virtual void update (unsigned long long *src, size_t current_frame);  // update sample data
	   size_t getSampleSize ();  // return the size of the sample 
	   unsigned char getCoding (unsigned char nbits); // return coding
	   signed char* getW ();  // return pointer to the W channel
	   signed char* getX ();  // return pointer to the X channel
	   signed char* getY ();  // return pointer to the Y channel
	   signed char* getZ ();  // return pointer to the Z channel
	   signed char* getChannel (int ch);   // return pointer to the channel by its number from chan0 (W) till chan3 (Z)
};

class Error {
 public:
         enum ErrorType { CONTINOUS=1, FIELDS=2, SYNC2=4, SYNC=8 };
 private:
	 unsigned int type;
         char *msg;
 public:	 
	 Error ();
	 Error (char * src);
	 Error (char * mask, unsigned char arg1, unsigned char arg2);
         virtual ~Error ();
	 char *getMsg ();
	 void setMsg (char * src);
	 void setMsg (char * mask, unsigned char arg1, unsigned char arg2);
	 unsigned int getType ();
	 void setType (unsigned int src);
	 void clear ();
};

#endif // __FRAMEDATA_HPP
