#!/usr/bin/perl
#
# This script cuts the piece of decoded data of chosen size by pointing
# the time (in MJD) of this piece (time of the center of the window)
#
use Getopt::Long;

# if there are no parameters in command line
# # help is called and program normally exit
&help($0) && exit if $#ARGV < 0;

$mjd = -1;
$period = -1;
$basename = -1;
$framesize = 2520;  # size of the frame in bytes (for one track)
$ntracks = 64;      # number of tracks
$tracksize = $ntracks * $framesize;   # size of the frame for all frames 

$nchan = 8;         # number of channels
$nbits = 2;         # number of bits
$fanout = 4;        # fanout 1:4
$piece_size = 11;   # the size of the piece in periods

# Parse command line
GetOptions ( "mjd=f" => \$mjd,       # - MJD
             "nchan=i" => \$nchan,   # - number of channels
             "fanout=i" => \$fanout, # - fanout
             "period=f" => \$period, # - period in ms
             "o=s" => \$basename,    # - basename of the file
             "h"     => \$help);     # - to print the help

if ($help) { &help($0); exit 0; }

if ($nchan != 8 && $nchan != 16) { &error("number of channels is wrong!"); }
if ($fanout != 4 && !fanout != 2) { &error("fanout is wrong!"); }
$samp_freq = 512. / $nchan / $nbits;   # sampling frequency in MHz
$sec_step = (($framesize - 20) * 8 * $fanout) / $samp_freq / 1000000.;  # the step in seconds between frames (0.0025 sec)

# check whether input file is pointed out
&error ("no input file!") if $#ARGV < 0;

if ($mjd == -1) { &error ("MJD time is not set!"); }
$mjd_prefix = substr ($mjd, 0, 2);

$file = $ARGV[0];

if ($period == -1) { &error ("period is not set!"); }
if ($basename == -1) { 
 $outfile = `basename $file`;
 chomp $outfile;
 $outfile .= ".cut";
} else { $outfile = $basename . ".cut"; }

$file = $ARGV[0];
$syncpos = `mk5decode -q --dummy -n 1 --mjd_prefix $mjd_prefix --fanout $fanout --chan $nchan  $file   2>&1 | grep sync | awk '{print \$4}' -`;
chomp $syncpos;
$startmjd = `mk5decode -q --dummy -n 1 --mjd_prefix $mjd_prefix --fanout $fanout --chan $nchan  $file   2>&1 | grep MJD: | awk '{print \$5}' -`;
chomp $startmjd;

$left = `echo "scale=20\n((($mjd-$startmjd)*86400.)-(($piece_size/2.)*$period/1000.))" | bc -l`;  #  left boundary of the piece in seconds
chomp $left;
if ($left < 0) { $left = 0; }
$right = `echo "scale=20\n$left+($piece_size*$period)/1000." | bc -l`;  # right boundary in seconds;
chomp $right;
$lb = `echo "scale=20\n$left/$sec_step" | bc -l`;  # left boundary in integer number of frames
chomp $lb;
$lb = int ($lb);
$rb = `echo "scale=20\n$right/$sec_step" | bc -l`;  # right boundary in integer number of frames
chomp $rb;
$rb = int ($rb) + 1;

# number of bytes to skip to cut a piece
$skip=$syncpos+$lb*$tracksize;
# number of bytes in a piece
$count=($rb-$lb)*$tracksize;

print "mjd:        $mjd\n";
print "start mjd:  $startmjd\n";
print "period:     $period  ms\n";
print "#ch: $nchan   fanout 1:$fanout   #bits: $nbits   sampl freq = $samp_freq MHz   framesize = $sec_step secs\n";
print "syncpos = $syncpos bytes\n";
printf ("skip %d frames (%d bytes, %f secs)\n", $lb, $lb*$tracksize, $lb*$sec_step);
printf ("cut %d frames (%d bytes, %f secs)\n", $rb - $lb, ($rb-$lb)*$tracksize, ($rb - $lb)*$sec_step);

open (IN, "<$file") || die "can't open file \"$file\" for reading: $!\n";
seek (IN, $skip, SEEK_SET);
read (IN, $buffer, $count);
close (IN);

open (OUT, ">$outfile") || die "can't open file \"$outfile\" for writing: $!\n";
print OUT $buffer;
close (OUT);

##########################
#
#   F U N C T I O N S
#
#

# unique subroutine to output errors
# one input parameter (string): the error message
sub error {
 my ($error) = @_;
 print "Error: $error\n";
 exit 1;
}

# help
sub help {
 local ($prg) = @_;
 $prg = `basename $prg`;
 chomp $prg;
 print "Usage: $prg [options] <file>\n";
 print "        -mjd <MJD>    - time of the center of the window to cut\n";
 print "        -nchan <N>    - number of channels, either 8 or 16. Default - 8\n";
 print "        -fanout <N>   - fanout, either 4 or 2. Default - 4 (i.e. 1:4)\n";
 print "        -period <P>   - period in ms. The period is determined the size of the piece to cut\n";
 print "                        The size of the piece is 11 periods (roughly) with 5 periods before\n";
 print"                         and after the period with the event\n";
 print "        -o <basename> - basename of the file\n";
 print "        -h            - plot this help\n";
 exit 0;
}
