#ifndef __FRAMEDATA_HPP
#define __FRAMEDATA_HPP

#include <stdio.h>

class FrameInfo;

/*++++++++++  class to work with Mark5 data  ++++++++*/
class FrameData {
          int Ntracks;         // number of tracks
          size_t tracksize;    // size of frame for all tracks in bytes = 64 tracks * 2520 bytes
          static const size_t headersize = 12;  // size of the header
	  static const size_t auxsize = 8;      // size of the auxinfo field

	  // this table keep correspondences between group (5) and track (8) numbers
	  // to the real track number
	  //                   Group (Digit A10)  Track (Digit A9)
	  //                                   |  |
	  static const unsigned int TrackTable[5][8];

	  unsigned long *raw;     // raw frame data for all tracks
	  FrameInfo *info;        // array with info of all tracks
	  unsigned long *modpattern;   // pattern for modifying input data

	  void fillTracks (); // extract each track from raw data array
	  void init_mod_pattern ();  // initialize modulation pattern
  protected:
          size_t framesize;                     // size of frame in bytes
	  size_t streamsize;                    // size of data
	  int size_of_long;                     // size of long type
	  unsigned long **tracks;               // 2D-array of tracks data
  public:
	  FrameData (int Ntr, size_t fs, size_t ts);    // constructor
	  virtual ~FrameData ();  // destructor
	  virtual void update (unsigned long *src);  // update frame data
	  static unsigned int getTrackNumber (unsigned int tr, unsigned int gr); // return real track number
	  void getHeaders (); // print header info for all tracks
	  void getAuxInfos (); // print aux info for all tracks
	  void getAllInfos (); // print all info for all tracks
	  double getSeconds (unsigned int tr = 0);  // return seconds of the track tr (default - 0)
	  size_t getFrameSize() { return framesize; } // return size of frame
	  void getHeader (int track = 0); // print header info for pointed track (default - 0)
	  void getAuxInfo (int track = 0); // print aux info for pointed track (default - 0)
	  void getAllInfo (int track = 0); // print all info for pointed track (default - 0)
};

/*+++++++++  class with header & aux info of frame  ++++++++*/
class FrameInfo {
          unsigned long syncword;        // sync word
          unsigned char time[6];         // 6-byte of time record
          unsigned short crc;            // 2-byte CRC-16 record
          bool is_crc_ok;                // true if crc is OK, otherwise - false
	  unsigned short calc_crc;       // 2-byte CRC-16 calculated record 
          unsigned char station;         // 1-byte station ID cod
          unsigned char rec[3];          // 3-byte of recording & headstack parameters
          unsigned char dar[3];          // 3-byte DAR parameters
          unsigned char spare;           // zero spare BCD-digits

	  static const int mjd_prefix = 53; // two most significant digits in MJD
	  unsigned int mjd;              // three least significant mjd digits
	  double seconds;                // seconds last from mjd before observations begun
	  unsigned int station_id;       // keep station ID (combination of two digits A1 & A2)
	  unsigned int recorder;         // # of recorder  0-1 (A3)
	  unsigned int headstack;        // number of headstack 0 (A4)
	  unsigned int pos_sign;         // position sign 0:+  8:-  (A5)
	  char pos_sign_char;            // "+" or "-" in case of pos_sign  or "?" if pos_sign keeps bad value
	  unsigned int position;         // position (combination of A6, A7 & A8)
	  unsigned int track;            // track BCD-digit 0 - 7  (A9)
	  unsigned int group;            // group BCD-digit 0 - 4  (A10)
	  unsigned int trackN;           // real track number
	  unsigned int bit;              // bit BCD-digit (A11)  0: magnitude  1: sign
	  char bit_char;                 // 'm' for magnitude bit & 's' for sign bit , '?' - otherwise
	  unsigned int sideband;         // sideband BCD-digit (A12)  0: lower  1: upper
	  char sideband_char;            // 'l' for lower band,  'u' - for upper,  '?' - otherwise
          unsigned int converter;        // converter BCD-digit (A13) 0 - 7
	  unsigned int bbc;              // bbc number = converter + 1
	  unsigned int darN;             // DAR number 0 - 1 (A14)
	  bool is_spare_zero;            // true if spare 8-bit are zeros, false - otherwise
  protected:
	  unsigned short BCD16_reverse (unsigned short word);  // this function reverse bits order in 16-bit word
	  unsigned char BCD8_reverse (unsigned char word);  // this function reverse bits order in 8-bit word

	  unsigned short calcCRC (unsigned char *ti, int len);  // calculate CRC-16 on time record
  public:
          FrameInfo ();   //constructor
          virtual ~FrameInfo ();  // destructor
	  void setup (unsigned long *tr, int len);  // initialize class data from track 
	  void getHeader ();  // print header info
	  void getAuxInfo ();  // print aux info
	  void getAllInfo ();  // print all info
          double getSeconds ();  // return seconds record
};


/*++++++++++++  class that works with complete bit-streams  +++++++++++++++++++*/
class Bitstream { 
	unsigned int numoftr[4]; // track numbers for this bit-stream
//   protected:
//        unsigned long BCD32_reverse (unsigned long word);  // this function reverse bits order in 32-bit word
   public:
        unsigned long *stream;   // complete bit-stream without fan-out
	Bitstream ();  // constructor
	virtual ~Bitstream();  // destructor
	void initialize (size_t cs, int len, int shift);  // allocate memory, fill numoftr
	                                                  // should be used from BBCData constructor
        void fillStreams (unsigned long **tr, int len); // fill streams with data
};


/*++++++++++  class to work with real BBC channels rather than tracks  ++++++++*/
class BBCData : public FrameData {

  protected:
	  size_t channelsize;   // size of full bit-stream without fan-out  in bytes
	  struct BBC { Bitstream usign;  // data for upper-band sign bit-stream
		       Bitstream umagn;  // data for upper-band magn bit-stream
		       Bitstream lsign;  // data for lower-band sign bit-stream
		       Bitstream lmagn;  // data for lower-band magn bit-stream
	  };
	  BBC channels[4]; // array for all data

  public:
	  BBCData (int Ntr, size_t fs, size_t ts);   // constructor
	  virtual ~BBCData ();  // destructor
	  virtual void update (unsigned long *src);  // update bbc data
};

/*++++++++++++++++++ class to work directly with real data ++++++++++++++++++++*/
class SampleData : public BBCData {

           static const signed char vlba[4];
	   // Codings:
	   // --------------------------------------------
	   // VLBA |  sign | magn       AT |  sign | magn
	   // --------------------------------------------
	   // +3       1      1         +3     0      1
	   // +1       1      0         +1     0      0
	   // -1       0      1         -1     1      0
	   // -3       0      0         -3     1      1

   protected:
	   size_t samplesize;  // size of full sampled data for each sideband in bytes
	   struct Pol { signed char *WXYZ[4]; // it is a upper(0) & lower(1) sideband in a upper frequency and then upper (2) & lower (3) in a lower frequency
	   };
	   Pol cp[2]; // Left (0) & Right (1) circular polarization
   public:
	   SampleData(int Ntr, size_t fs, size_t ts);  // constructor
	   virtual ~SampleData();  // destructor
	   virtual void update (unsigned long *src);  // update sample data
	   size_t getSampleSize ();  // return the size of the sample 
	   signed char* getLCPW ();  // return pointer to the LCP most upper sideband (0)
	   signed char* getLCPX ();  // return pointer to the LCP lower sideband in a upper frequnecy (1)
	   signed char* getLCPY ();  // return pointer to the LCP upper sideband in a lower frequnecy (2)
	   signed char* getLCPZ ();  // return pointer to the LCP lower sideband in a lower frequnecy (3)
	   signed char* getRCPW ();  // return pointer to the RCP most upper sideband (0)
	   signed char* getRCPX ();  // return pointer to the RCP lower sideband in a upper frequnecy (1)
	   signed char* getRCPY ();  // return pointer to the RCP upper sideband in a lower frequnecy (2)
	   signed char* getRCPZ ();  // return pointer to the RCP lower sideband in a lower frequnecy (3)
	   signed char* getChannel (int ch);   // return pointer to the channel by its number from LCP W-> Z and then RCP W -> Z
};

#endif // __FRAMEDATA_HPP
