#include "framedata.hpp"
#include "crc16.hpp"
#include <string.h>
#include <bitset>

// initialize TrackTable
// 99 - mean bad track number in this program

// this table keep correspondences between group (5) and track (8) numbers
// to the real track number
//                       Group (Digit A10)  Track (Digit A9)
//                                       |  |
const unsigned int FrameData::TrackTable[5][8] = { {  3,  5,  7,  9, 11, 13, 15, 17 },
                                                   { 19, 21, 23, 25, 27, 29, 31, 33 },
	                                           {  2,  4,  6,  8, 10, 12, 14, 16 },
			                           { 18, 20, 22, 24, 26, 28, 30, 32 },
			                           {  1, 35,  0, 34, 99, 99, 99, 99 } };

// sign & magn bit correspondence
const signed char SampleData::vlba[4] = { -3, -1, 1, 3 };
// Codings:
// --------------------------------------------
// VLBA |  sign | magn       AT |  sign | magn
// --------------------------------------------
// +3       1      1         +3     0      1
// +1       1      0         +1     0      0
// -1       0      1         -1     1      0
// -3       0      0         -3     1      1

// constructor
FrameData::FrameData(int Ntr, size_t fs, size_t ts) {
  size_of_long = sizeof (unsigned long);
  Ntracks = Ntr;
  framesize = fs;
  tracksize = ts;
  streamsize = framesize - headersize - auxsize;

  raw = new unsigned long[tracksize/size_of_long];
  if (!raw) { perror("new raw"); exit(1); }
  memset (raw, 0, tracksize);

  tracks = new unsigned long*[Ntracks];
  if (!tracks) { perror("new tracks"); exit(1); }
  for (int i=0; i<Ntracks; ++i) {
   tracks[i] = new unsigned long[framesize/size_of_long];
   if (!tracks[i]) { perror("new tracks[i]"); exit(1); }
   memset (tracks[i], 0, framesize);
  }
  info = new FrameInfo[Ntracks];
  if (!info) { perror("new info"); exit(1); }

  modpattern = new unsigned long[(tracksize - Ntracks * (headersize + auxsize))/size_of_long];
  if (!modpattern) { perror("new modpattern"); exit(1); } 
  memset (modpattern, 0, tracksize - Ntracks * (headersize + auxsize));

  // initialize modulation pattern
  init_mod_pattern ();
}

// destructor
FrameData::~FrameData() {
 delete (raw);
 delete [] tracks;
 delete (info);
 delete (modpattern);
}

// this function generates modulation pattern for bits coding (see p.4 memo #230.2)
void FrameData::init_mod_pattern () {
  static unsigned ff[16] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };
  unsigned output;
  for (int k=0, j=0; k<(tracksize - Ntracks * (headersize + auxsize))/size_of_long; k+=2, j++) {
   output = ff[10] ^ ff[12] ^ ff[13] ^ ff[15];
   for (int i=15; i>0; i--) ff[i] = ff[i-1];
   ff[0] = output;
   modpattern[k] = output * 0xffffffff;
   modpattern[k + 1] = output * 0xffffffff;
   if (j % 8 == 7) {  // this check is necessary because VLBA formatter advances by one step
	   // including parity bits. Now we haven't these parity bits so we need
	   // to make extra empty step
    output = ff[10] ^ ff[12] ^ ff[13] ^ ff[15];
    for (int i=15; i>0; i--) ff[i] = ff[i-1];
    ff[0] = output;
   }
  }
}

// extract each track from raw data array
void FrameData::fillTracks () {
  std::bitset<32> output; // bitset for one long for one track
  int bitcount; // counter that indicates the filling of the output bitset
  int track_pos; // position of the current element in track

  for (int b=0; b<2; ++b) { // at first, fill first 32 tracks, and then others
  for (int t=b * (Ntracks/2), w=0; t<Ntracks/2 + b * (Ntracks/2); ++t, ++w) {
   output.reset();
   bitcount = 0;
   track_pos = 0;
   // raw data loop
   for (int d=b; d<tracksize/size_of_long; d+=2) {
    std::bitset<32> bs (raw[d]);
    output[bitcount] = bs[w];
    bitcount++;
    if (bitcount == 32) { // check are all 32 bits in output bitset set
     bitcount = 0;
     tracks[t][track_pos] = output.to_ulong();
     output.reset();
     ++track_pos;
    }
   } // raw data loop
  } // tracks loop
  } // b loop
}

// update frame with new data
void FrameData::update (unsigned long * src) {
  memset (raw, 0, tracksize);
  raw = (unsigned long *)memcpy (raw, src, tracksize); 
  for (int i=0; i<Ntracks; ++i) memset (tracks[i], 0, framesize);
  // demodulate our data
  for (int k=(headersize * Ntracks)/size_of_long, n=0; k<(tracksize - auxsize * Ntracks)/size_of_long; ++k, ++n)
   raw[k] ^= modpattern[n];
  fillTracks ();
  for (int i=0; i<Ntracks; ++i) info[i].setup(tracks[i], framesize/size_of_long);
}

// print header info for all tracks
void FrameData::getHeaders () {
 for (int i=0; i<Ntracks; ++i) { printf ("track #%2d:  ", i); info[i].getHeader(); }
}

// print aux info for all tracks
void FrameData::getAuxInfos () {
 for (int i=0; i<Ntracks; ++i) { printf ("track #%2d:  ", i); info[i].getAuxInfo(); }
}

// print all info for all tracks
void FrameData::getAllInfos () {
 for (int i=0; i<Ntracks; ++i) { printf ("track #%2d:  ", i); info[i].getAllInfo(); }
}

// return seconds of the track tr (default tr = 0)
double FrameData::getSeconds (unsigned int tr) {
 return info[tr].getSeconds();
}

// return real track number
unsigned int FrameData::getTrackNumber (unsigned int tr, unsigned int gr) {
 if (tr > 7 || gr > 4) return 99; // return bad value
 return TrackTable[gr][tr];
}

// print header info for pointed track (default - 0)
void FrameData::getHeader (int track) {
 if (track < 0 || track > Ntracks - 1) track = 0;
 printf ("track #%2d:  ", track); 
 info[track].getHeader();
}

// print aux info for pointed track (default - 0)
void FrameData::getAuxInfo (int track) {
 if (track < 0 || track > Ntracks - 1) track = 0;
 printf ("track #%2d:  ", track); 
 info[track].getAuxInfo();
}

// print all info for pointed track (default - 0)
void FrameData::getAllInfo (int track) {
 if (track < 0 || track > Ntracks - 1) track = 0;
 printf ("track #%2d:  ", track); 
 info[track].getAllInfo();
}





// constructor
FrameInfo::FrameInfo () {
}

// destructor
FrameInfo::~FrameInfo () {
}

// this function reverse bits order in 16-bit word
unsigned short FrameInfo::BCD16_reverse (unsigned short word) {
 int i;
 unsigned short nVal = 0;
 for (i=15; i>=0; i--) { nVal |= (word & 1) << i; word >>= 1; }
 return nVal;
}

// this function reverse bits order in 8-bit word
unsigned char FrameInfo::BCD8_reverse (unsigned char word) {
 int i;
 unsigned char nVal = 0;
 for (i=7; i>=0; i--) { nVal |= (word & 1) << i; word >>= 1; }
 return nVal;
}

// calculate CRC-16 on time record
unsigned short FrameInfo::calcCRC (unsigned char *ti, int len) {
 CCrc16 crcode;
 crcode.m_Polynomial = 0x8005;      // Most commonly 0x8005 or 0x1021
 crcode.m_Reflect = false;          // use TRUE to reverse bits of input bytes
 crcode.m_ReflectCrc = false;       // use TRUE to reverse bits of output crc
 crcode.m_CrcXorOutput = 0;         // use 0xFFFF to invert crc output
 crcode.m_ResetValue = 0;           // Most commonly 0 or 0xFFFF

 crcode.CreateTable();
 return crcode.ComputeCrcUsingTable (ti, len);
}

// initialize class data from track
void FrameInfo::setup (unsigned long *tr, int len) {
  syncword = tr[0];
  unsigned long charmask = 0x000000ff;
  for (int i=0; i<4; ++i) time[i] = (tr[1] >> (8*i)) & charmask; 
  for (int i=0; i<2; ++i) time[i+4] = (tr[2] >> (8*i)) & charmask;
  for (int i=0; i<6; ++i) time[i] = BCD8_reverse(time[i]);  // reverse bits
  unsigned long shortmask = 0x0000ffff;
  crc = (tr[2] >> 16) & shortmask; 
  crc = BCD16_reverse(crc);  // reverse bits
  calc_crc = calcCRC (time, 6);  // calculate CRC-16
  if (calc_crc == crc) is_crc_ok = true; else is_crc_ok = false;  // check if CRC is OK
  char mjdstring[3];
  sprintf (mjdstring, "%d%d%d", time[0]>>4, time[0] & 0x0f, time[1] >> 4);
  sscanf (mjdstring, "%d", &mjd);
  char secstring[10];
  sprintf (secstring, "%d%d%d%d%d.%d%d%d%d", time[1] & 0x0f, time[2] >> 4, time[2] & 0x0f, time[3] >> 4, time[3] & 0x0f, time[4] >> 4, time[4] & 0x0f, time[5] >> 4, time[5] & 0x0f);
  sscanf (secstring, "%lf", &seconds);
  station = tr[len - 2] & charmask;
  station = BCD8_reverse (station);  // reverse bits
  char stationstring[2];
  sprintf (stationstring, "%d%d", station >> 4, station & 0x0f);
  sscanf (stationstring, "%d", &station_id);
  for (int i=1; i<4; ++i) rec[i-1] = (tr[len - 2] >> (8*i)) & charmask;
  for (int i=0; i<3; ++i) rec[i] = BCD8_reverse (rec[i]);  // reverse bits 
  recorder = rec[0] >> 4;
  headstack = rec[0] & 0x0f;
  pos_sign = rec[1] >> 4;
  pos_sign_char = (pos_sign == 0 ? '+' : (pos_sign == 8 ? '-' : '?'));
  char positionstring[3];
  sprintf (positionstring, "%d%d%d", rec[1] & 0x0f, rec[2] >> 4, rec[2] & 0x0f);
  sscanf (positionstring, "%d", &position);
  for (int i=0; i<3; ++i) dar[i] = (tr[len - 1] >> (8*i)) & charmask;
  track = dar[0] & 0x0f;
  group = dar[0] >> 4;
  trackN = FrameData::getTrackNumber (track, group);
  bit = dar[1] & 0x0f;
  bit_char = (bit == 0 ? 'm' : (bit == 1 ? 's' : '?'));
  sideband = dar[1] >> 4;
  sideband_char = (sideband == 0 ? 'l' : (sideband == 1 ? 'u' : '?'));
  converter = dar[2] & 0x0f;
  bbc = converter + 1;
  darN = dar[2] >> 4;
  spare = (tr[len - 1] >> 24) & charmask;
  if (spare == 0) is_spare_zero = true; else is_spare_zero = false;
}

// print header info
void FrameInfo::getHeader () {
  printf ("MJD = [%d]%d  secs = %.4lf  CRC = 0x%0x [%s]\n", mjd_prefix, mjd, seconds, crc, is_crc_ok ? "OK" : "Mismatch");
}

// print aux info
void FrameInfo::getAuxInfo () {
  printf ("station: %3d rec: %d head: %2d pos: %c%d TRACK# %2d bit: \"%c\" band: \"%c\" bbc: %d dar# %d [spare: %s]\n", station_id, recorder, headstack, pos_sign_char, position, trackN, bit_char, sideband_char, bbc, darN, is_spare_zero ? "OK" : "Mismatch");
}

// print all info
void FrameInfo::getAllInfo () {
  printf ("MJD = [%d]%d  secs = %.4lf  CRC = 0x%0x [%s] station: %3d rec: %d head: %2d pos: %c%d TRACK# %2d bit: \"%c\" band: \"%c\" bbc: %d dar# %d [spare: %s]\n", mjd_prefix, mjd, seconds, crc, is_crc_ok ? "OK" : "Mismatch", station_id, recorder, headstack, pos_sign_char, position, trackN, bit_char, sideband_char, bbc, darN, is_spare_zero ? "OK" : "Mismatch");
}

// return seconds record
double FrameInfo::getSeconds () {
 return seconds; 
}



// constructor
Bitstream::Bitstream () {
}

// destructor
Bitstream::~Bitstream () {
  if (stream) delete (stream); 
}

// allocate memory, fill numoftr
// should be used from BBCData constructor
// 3rd argument "shift" is the first value (track number) 
// in the track map for this bbc
void Bitstream::initialize (size_t cs, int len, int shift) {
 // make table with info about tracks for that bit-stream
 for (int k=0; k<4; ++k) numoftr[k] = shift + 2*k;
 stream = new unsigned long[len];
 if (!stream) { perror("new stream"); exit(1); }
 memset (stream, 0, cs);
}

// fill streams with data
void Bitstream::fillStreams (unsigned long **tr, int len) {
  std::bitset<32> output;
  int stream_counter = 0;
  // this cycle on track 32-bit elements 
  // i=3 - because of header size and " -2" - because of aux info
  for (int i=3; i<len - 2; ++i) {   
    std::bitset<32> bs0(tr[numoftr[0]][i]); 
    std::bitset<32> bs1(tr[numoftr[1]][i]); 
    std::bitset<32> bs2(tr[numoftr[2]][i]); 
    std::bitset<32> bs3(tr[numoftr[3]][i]); 
    for (int o=0; o<4; ++o) { // cycle on 4 output words
     output.reset();
     for (int b=0; b<8; ++b) { // cycle on 8 iterations on filling one output word
       output[4*b]   = bs0[8*o + b]; 
       output[4*b+1] = bs1[8*o + b]; 
       output[4*b+2] = bs2[8*o + b]; 
       output[4*b+3] = bs3[8*o + b]; 
     } // b
     stream[stream_counter] = output.to_ulong();
     ++stream_counter;
    } // o
  } // i
}


// constructor
BBCData::BBCData (int Ntr, size_t fs, size_t ts) : FrameData (Ntr, fs, ts) {
 channelsize = 4 * streamsize;	
 for (int i=0; i<4; ++i) {
  // the last argument of this function is first value in the track map, i.e.
  // the track number. Other track numbers are calculated within initialize
  // function itself with step = 2 four times
  channels[i].usign.initialize (channelsize, channelsize/size_of_long, 0 + i*16);
  channels[i].umagn.initialize (channelsize, channelsize/size_of_long, 8 + i*16);
  channels[i].lsign.initialize (channelsize, channelsize/size_of_long, 1 + i*16);
  channels[i].lmagn.initialize (channelsize, channelsize/size_of_long, 9 + i*16);
 }
}

// destructor
BBCData::~BBCData () {
}

void BBCData::update (unsigned long *src) {
  FrameData::update (src);
  // then we need to initialize streams
  for (int i=0; i<4; ++i) {
   channels[i].usign.fillStreams(tracks, framesize/size_of_long);
   channels[i].umagn.fillStreams(tracks, framesize/size_of_long);
   channels[i].lsign.fillStreams(tracks, framesize/size_of_long);
   channels[i].lmagn.fillStreams(tracks, framesize/size_of_long);
  }
}


// constructor
SampleData::SampleData(int Ntr, size_t fs, size_t ts) : BBCData (Ntr, fs, ts) {
  samplesize = channelsize * 8;
  for (int i=0; i<2; ++i) {
   for (int j=0; j<4; ++j) cp[i].WXYZ[j] = new signed char[samplesize];
  }
}

// destructor
SampleData::~SampleData() {
  for (int i=0; i<2; ++i) {
   for (int j=0; j<4; ++j) delete (cp[i].WXYZ[j]); 
  }
}

// update sample data
void SampleData::update (unsigned long *src) {

  BBCData::update (src);

  std::bitset<2> output;
  for (int i=0; i<channelsize/size_of_long; ++i) { // cycle on each long-word of stream
   for (int s=0; s<2; ++s) { // cycle on poalrizations
    for (int b=0; b<4; b+=2) {
     std::bitset<32> bs0(channels[s+b].lsign.stream[i]);
     std::bitset<32> bs1(channels[s+b].lmagn.stream[i]);
     std::bitset<32> bs2(channels[s+b].usign.stream[i]);
     std::bitset<32> bs3(channels[s+b].umagn.stream[i]);
     for (int j=0; j<32; ++j) { // cycle on 32 bits of streams
      output[1] = bs0[j]; output[0] = bs1[j];   // there must be sign in the most significant bit
      cp[s].WXYZ[b][32*i + j] = vlba[output.to_ulong()];
      output[1] = bs2[j]; output[0] = bs3[j]; 
      cp[s].WXYZ[b+1][32*i + j] = vlba[output.to_ulong()];
     } // j
    } // b
   } // s
  } // i
}

// return the size of the sample
size_t SampleData::getSampleSize () {
  return samplesize;
}

// return pointer to the LCP most upper sideband (0)
signed char* SampleData::getLCPW () {
  return cp[0].WXYZ[0];
}

// return pointer to the LCP lower sideband in a upper frequency (1)
signed char* SampleData::getLCPX () {
  return cp[0].WXYZ[1];
}

// return pointer to the LCP upper sideband in a lower frequency (2)
signed char* SampleData::getLCPY () {
  return cp[0].WXYZ[2];
}

// return pointer to the LCP lower sideband in a lower frequency (3)
signed char* SampleData::getLCPZ () {
  return cp[0].WXYZ[3];
}

// return pointer to the RCP most upper sideband (0)
signed char* SampleData::getRCPW () {
  return cp[1].WXYZ[0];
}

// return pointer to the RCP lower sideband in a upper frequency (1)
signed char* SampleData::getRCPX () {
  return cp[1].WXYZ[1];
}

// return pointer to the RCP upper sideband in a lower frequency (2)
signed char* SampleData::getRCPY () {
  return cp[1].WXYZ[2];
}

// return pointer to the RCP lower sideband in a lower frequency (3)
signed char* SampleData::getRCPZ () {
  return cp[1].WXYZ[3];
}

// return pointer to the channel by its number from LCP W -> Z and then RCP W -> Z
signed char* SampleData::getChannel (int ch) {
  if (ch < 0 || ch > 7) {
   printf ("Bad channel number!\n");
   exit (1);
  }
  return cp[ch / 4].WXYZ[ch % 4];	
}
