#ifndef __FRAMEDATA_HPP
#define __FRAMEDATA_HPP

#include <stdio.h>
#include <list>
#include <map>

class FrameInfo;

/*++++++++++ class with all Error Info of the frame +++++++++++++*/
class FrameCheckInfo {
          long crc_mismatch_number;  // the number of CRC mismatches in this frame
	  long spare_mismatch_number; // the number of spare digits mismatches
	  long cont_mismatch_number; // the number of time not-continuities
	  long fields_mismatch_number; // the number of events when the fields are different for different tracks
	  long labels_mismatch_number; // the number of events when time labels are different for the same frame for different tracks
          std::map<size_t, std::list<int> > crc_mismatch_tracks;  // the map of the frame_number and corresponding list of tracks numbers where CRC mismathches occured
          std::map<size_t, std::list<int> > spare_mismatch_tracks; // the map of the frame_number and corresponding list of tracks numbers where spare digits mismatches occured
	  std::map<size_t, std::list<int> > cont_mismatch_tracks; //  -//- -//-
	  std::map<size_t, std::list<int> > fields_mismatch_tracks; //  -//- -//-
	  std::map<size_t, std::list<int> > labels_mismatch_tracks; //  -//- -//-
  public: FrameCheckInfo(); // constructor
          virtual ~FrameCheckInfo(); // destructor
	  void addSpare(size_t frame, int tr); // add spare digit mismatch
	  long getSpareN ();
	  char * getSpareList (char * str);
	  void addCRC(size_t frame, int tr); // add crc mismatch
	  long getCRCN ();
	  char * getCRCList (char * str);	  
	  void addCont(size_t frame, int tr); // add cont mismatch
	  long getContN ();
	  char * getContList (char * str);	  
	  void addFields(size_t frame, int tr); // add fields mismatch
	  long getFieldsN ();
	  char * getFieldsList (char * str);	  
	  void addLabels(size_t frame, int tr); // add labels mismatch
	  long getLabelsN ();
	  char * getLabelsList (char * str);	  
	  void clear (); // clear all the fields
};

/*++++++++++  class to work with Mark5 data  ++++++++*/
class FrameData {
          size_t frame_number;   // number of the current frame
          int Ntracks;         // number of tracks
          size_t tracksize;    // size of frame for all tracks in bytes = 64 tracks * 2520 bytes
          static const size_t headersize = 12;  // size of the header
	  static const size_t auxsize = 8;      // size of the auxinfo field

	  // this table keep correspondences between group (5) and track (8) numbers
	  // to the real track number
	  //                   Group (Digit A10)  Track (Digit A9)
	  //                                   |  |
	  static const unsigned int TrackTable[5][8];

	  unsigned long long *raw;     // raw frame data for all tracks
	  FrameInfo *info;        // array with info of all tracks
	  unsigned long long *modpattern;   // pattern for modifying input data

	  void fillTracks (); // extract each track from raw data array
	  void init_mod_pattern ();  // initialize modulation pattern
	  unsigned long long swapBytes (unsigned long long word); // swap bytes in 64-bits word

          char *output_info;   // keep the string of the output data
	  FrameCheckInfo fci;  // keep all error info in this frame
  protected:
          size_t framesize;                     // size of frame in bytes
	  size_t streamsize;                    // size of data
	  int size_of_long;                     // size of long type
	  int size_of_long_long;                // size of long long type
	  unsigned long **tracks;               // 2D-array of tracks data
	  bool to_reverse;     // reverse (or not) the bits in raw input data
  public:
	  FrameData (int Ntr, size_t fs, size_t ts, bool swap = false);    // constructor
	  virtual ~FrameData ();  // destructor
	  virtual void update (unsigned long long *src, size_t current_frame);  // update frame data
	  static unsigned int getTrackNumber (unsigned int tr, unsigned int gr); // return real track number
	  void getHeaders (); // print header info for all tracks
	  void getAuxInfos (); // print aux info for all tracks
	  void getAllInfos (); // print all info for all tracks
	  double getSeconds (unsigned int tr = 0);  // return seconds of the track tr (default - 0)
	  unsigned int getMJD (unsigned int tr = 0); // return mjd record of the track tr (default - 0)
	  size_t getFrameSize() { return framesize; } // return size of frame
	  void getHeader (int track = 0); // print header info for pointed track (default - 0)
	  void getAuxInfo (int track = 0); // print aux info for pointed track (default - 0)
	  void getAllInfo (int track = 0); // print all info for pointed track (default - 0)
	  void getTrackInfo (int track = 0); // print info about track (default - 0)
	  void getCheckInfo (); // print all check info
	  void getSimpleCheckInfo (); // print all check info in simple way
	  size_t getFrameNumber (); // get the frame number
	  unsigned int getTrackN (int track = 0); // get true number of the track (default - 0)
};

/*+++++++++ class with info that we want to keep to check the correctness of the data ++++++++++++*/
class SaveInfo {
          unsigned long aux_info[2];     // raw aux-info
	  unsigned int mjd;              // MJD date (integer number of days)
	  double seconds;                // the same as seconds but from the previous frame
	  size_t frame_number;           // frame number
  public:
          SaveInfo ();   //constructor
          virtual ~SaveInfo ();  // destructor
          double getSeconds ();  // return seconds record
	  unsigned int getMJD (); // return mjd record
	  unsigned long getAuxInfo (bool index = false); // return aux info field [0] when index = false (default) or [1] when index = true
	  size_t getFrameNumber (); // return frame number
	  void setAuxInfo (unsigned long first, unsigned long second); // set aux info
	  void setMJD (unsigned int src); // set MJD
	  void setSeconds (double src); // set seconds
	  void setFrameNumber (size_t frame); // set frame number
};

/*+++++++++  class with header & aux info of frame  ++++++++*/
class FrameInfo {
          unsigned long syncword;        // sync word
          unsigned char time[6];         // 6-byte of time record
          unsigned short crc;            // 2-byte CRC-16 record
          bool is_crc_ok;                // true if crc is OK, otherwise - false
	  unsigned short calc_crc;       // 2-byte CRC-16 calculated record 
          unsigned char station;         // 1-byte station ID cod
          unsigned char rec[3];          // 3-byte of recording & headstack parameters
          unsigned char dar[3];          // 3-byte DAR parameters
          unsigned char spare;           // zero spare BCD-digits

	  unsigned int mjd;              // MJD date (integer number of days)
	  double seconds;                // seconds last from mjd before observations begun
	  unsigned int station_id;       // keep station ID (combination of two digits A1 & A2)
	  unsigned int recorder;         // # of recorder  0-1 (A3)
	  unsigned int headstack;        // number of headstack 0 (A4)
	  unsigned int pos_sign;         // position sign 0:+  8:-  (A5)
	  char pos_sign_char;            // "+" or "-" in case of pos_sign  or "?" if pos_sign keeps bad value
	  unsigned int position;         // position (combination of A6, A7 & A8)
	  unsigned int track;            // track BCD-digit 0 - 7  (A9)
	  unsigned int group;            // group BCD-digit 0 - 4  (A10)
	  unsigned int trackN;           // real track number
	  unsigned int bit;              // bit BCD-digit (A11)  0: magnitude  1: sign
	  char bit_char;                 // 'm' for magnitude bit & 's' for sign bit , '?' - otherwise
	  unsigned int sideband;         // sideband BCD-digit (A12)  0: lower  1: upper
	  char sideband_char;            // 'l' for lower band,  'u' - for upper,  '?' - otherwise
          unsigned int converter;        // converter BCD-digit (A13) 0 - 7
	  unsigned int bbc;              // bbc number = converter + 1
	  unsigned int darN;             // DAR number 0 - 1 (A14)
	  bool is_spare_zero;            // true if spare 8-bit are zeros, false - otherwise

          SaveInfo save;                 // to keep info about previous frame
	  
	  bool initialized;  // true when there was previous value of all the fields
	  bool is_time_continous;  // true if the time labels in sequent frames are continously increased 
	  bool are_fields_same; // true if all other fields are the same

	  size_t frame_number;           // number of the current frame
  protected:
	  unsigned short BCD16_reverse (unsigned short word);  // this function reverse bits order in 16-bit word
	  unsigned char BCD8_reverse (unsigned char word);  // this function reverse bits order in 8-bit word

	  unsigned short calcCRC (unsigned char *ti, int len);  // calculate CRC-16 on time record
  public:
          FrameInfo ();   //constructor
          virtual ~FrameInfo ();  // destructor
	  void setup (unsigned long *tr, int len, size_t frame, int Ntr);  // initialize class data from track 
	  void getHeader ();  // print header info
	  void getAuxInfo ();  // print aux info
	  void getAllInfo ();  // print all info
	  void getTrackInfo (); // print info about track
          double getSeconds ();  // return seconds record
	  unsigned int getMJD (); // return mjd record
	  unsigned int getTrackN (); // get true number of the track
};


/*++++++++++++++++++ class to work directly with real data ++++++++++++++++++++*/
class SampleData : public FrameData {

           static const signed char vlba[4];
	   // Codings:
	   // --------------------------------------------
	   // VLBA |  sign | magn       AT |  sign | magn
	   // --------------------------------------------
	   // +3       1      1         +3     0      1
	   // +1       1      0         +1     0      0
	   // -1       0      1         -1     1      0
	   // -3       0      0         -3     1      1

   protected:
	   size_t samplesize;  // size of full sampled data for each sideband in bytes
	   signed char *Data[8]; // data stream, 0 - A, 1 - B, ... 7 - H
   public:
	   SampleData(int Ntr, size_t fs, size_t ts, bool swap = false);  // constructor
	   virtual ~SampleData();  // destructor
	   virtual void update (unsigned long long *src, size_t current_frame);  // update sample data
	   size_t getSampleSize ();  // return the size of the sample 
	   signed char* getChannel (int ch);   // return pointer to the channel by its number: 0 - A, 1 - B, ... 7 - H
	   void getTracksMap (); // print correspondance table between channels A, B, ..., H and info in tracks
};

class Error {
 public:
         enum ErrorType { CRC, SPARE, CONTINOUS, FIELDS, LABELS };
 private:
	 ErrorType type;
         char *msg;
 public:	 
	 Error ();
	 Error (char * src);
	 Error (char * mask, int arg1, int arg2);
         virtual ~Error ();
	 char *getMsg ();
	 void setMsg (char * src);
	 void setMsg (char * mask, int arg1, int arg2);
	 ErrorType getType ();
	 void setType (ErrorType src);
};

#endif // __FRAMEDATA_HPP
