/************************************************************************************
*  program pfs_depack
*  This programs unpacks data from the portable fast sampler.
*  Output is binary signed char point quantities.
*
*  Usage:
*  	pfs_depack -m mode
*                  [-w wires]
*                  [-d directory]
*                  [-e extension]
*                  [-b buffer_size]
*                  [-o outname]      [infile1] ... [infileN]
*
*  Options:
*
*	the -m argument specifies the data acquisition mode
*       the -w argument specifies the channel to unpack
*              Possible values:  all  - to unpack all valid channels (default)
*                                  1  - to unpack channel 1  -> RCP-Q
*                                  2  - to unpack channel 2  -> RCP-I
*                                  3  - to unpack channel 3  -> LCP-Q
*                                  4  - to unpack channel 4  -> LCP-I
*              or some combination of them,
*               for example: -w 1,3   - to unapck only channel 1 and channel 3
*              String 1,3 must not have spaces
*
*      the -d argument specifies the directory for output files
*             (default the same as for the input files)
*      the -e extension specifies the extension for names of output files
*             (default extension is ".out")
*      the -b buffer_size specifies the size of read buffer
*             (default is 1,000,000 bytes)
*      the -o outname specifies the base name (without extension) of output files
*             (default the output filename is input filename + extension)
*      the -h option prints help
*
*************************************************************************************/

/* these preprocessor' defines need to support Large Files */

# ifdef LARGEFILE
#   define _GNU_SOURCE 1
#   define _FILE_OFFSET_BITS 64
# endif


# include <math.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <asm/fcntl.h>
# include <sys/stat.h>
# include <unistd.h>

# include "unpack.h"

# define MaxString 255


void Help ();			/* print help */		
int ParseCmdLine ();		/* parse command line */
int ParseWires ();		/* parse wirestring */
char *SelectFileName ();	/* erase path part from filename */
void PrepareOutNames ();	/* prepare names of output files */

/*+++++++++++++++++++++++++++++++++++ Prompting ++++++++++++++++++++++++++++++++++++++*/

void Help (s)

char *s;	/* name of the programm */

{
  printf("\nPROGRAMM %s\n"
         "It is used to unpack data from the Portable Fast Sampler\n\n"
	 "[USAGE]: %s [options] <file(s) to process>\n\n"
	 "[OPTIONS]:\n"
	 "-m <mode>        - data acquisition mode\n"
         "                     Valid modes are 0: 2c1b (N/A)\n"
	 "                                     1: 2c2b\n"
	 "                                     2: 2c4b\n"
	 "                                     3: 2c8b\n"
	 "                                     4: 4c1b (N/A)\n"
	 "                                     5: 4c2b\n"
	 "                                     6: 4c4b\n"
	 "                                     7: 4c8b (N/A)\n"
	 "-w <wires>       - channels to unpack\n"
	 "                     Valid values of wires are:\n"
         "                         all  - to unpack all valid channels (default)\n"
         "                           1  - to unpack channel 1  -> RCP-Q\n"
         "                           2  - to unpack channel 2  -> RCP-I\n"
         "                           3  - to unpack channel 3  -> LCP-Q\n"
         "                           4  - to unpack channel 4  -> LCP-I\n"
         "                      or some combination of them, for example string -w 1,3\n"
	 "                      says programm to unapck only channel 1 and channel 3\n"
         "                      String wires (for example, -w 1,2,4) must not have spaces\n"
	 "-d <directory>   - path to the output files, default - the same as for the input ones\n"
	 "-e <extension>   - auxiliary filename extension given to the output files, default - .out\n"	
	 "                   It's possible to use one symbol %% in order to have a chance to\n"
	 "                   point number of wire in name of output file\n"
	 "-b <buffer_size> - size of read bufer, default - 1,000,000 bytes\n"
	 "-o <outname>     - name of file for output data. If -e option is given then full name\n"
	 "                   is outname + extension, default - input filename + .out\n"
	 "-h               - help\n\n", s, s);
  exit(0);
}

/*++++++++++++++++++++++ Parsing the command line +++++++++++++++++++++++++++++++++++++*/

int ParseCmdLine (argc, argv, mode, wirestring, dir, ext, bufsize, outname)

int       argc;		/* number of arguments */
char      **argv;	/* command line arguments */
int       *mode;
char      **wirestring;
char      **dir;
char      **ext;
long int  *bufsize;
char      **outname;

{
  int op;		/* option letter returned by getopt */

  extern char *optarg; 	/* if arg with option, this pts to it */
  extern int optind;	/* after call, ind into argv for next */

  /* loop over all the options in list */
  while ((op = getopt(argc, argv, "m:w:d:e:b:o:h")) != -1)
    switch(op) {
      case 'h':
        Help(argv[0]);
      break;
      
      case 'm':
        *mode = atoi(optarg);
	
	/* must specify a valid mode */
	if (*mode < 1 || *mode > 6 || *mode == 4) {
	 printf("Error: wrong mode: %d !\n", *mode);
	 exit(1);
	}
      break;
      
      case 'w':
	  if (*wirestring) { *wirestring = 0;
	                     free(*wirestring);
	            }
	  if ( !(*wirestring = (char *)calloc((int)strlen(optarg), sizeof(char *)) ) ) {
	    perror("calloc wirestring: ");
	    exit(1);
	   }
	  memcpy (*wirestring, optarg, (int)strlen(optarg));
      break;
      
      case 'b':
        *bufsize = atol(optarg);
	
	/* must specify a valid mode */
	if (*bufsize < 1 || *bufsize > 100000000) {
	 printf("Error: bufsize is out of range: %ld !\n", *bufsize);
	 exit(1);
	}
      break;
      
      case 'd':
	  if (*dir) { *dir = 0;
	              free(*dir);
	            }
	  if ( !(*dir = (char *)calloc((int)strlen(optarg), sizeof(char *)) ) ) {
	    perror("calloc dir: ");
	    exit(1);
	   }
	  memcpy (*dir, optarg, (int)strlen(optarg));
      break;
      
      case 'e':
	  if (*ext) { *ext = 0;
	              free(*ext);
	            }
	  if ( !(*ext = (char *)calloc((int)strlen(optarg), sizeof(char *)) ) ) {
	    perror("calloc ext: ");
	    exit(1);
	   }
	  memcpy (*ext, optarg, (int)strlen(optarg));
      break;
      
      case 'o':
	  if ( *outname ) { *outname = 0;
	                    free(*outname);
	                 }
	  if ( !(*outname = (char *)calloc((int)strlen(optarg), sizeof(char *)) ) ) {
	    printf("calloc outname: ");
	    exit(1);
	   }
	  memcpy (*outname, optarg, (int)strlen(optarg));
      break;
      
      case '?':
        Help(argv[0]);
      break;
    }
    
  return(optind);
}

/*+++++++++++++++++++++++++++ Parse string of wires ++++++++++++++++++++++++++++++*/

int ParseWires (wirestring, wire, mode)

char **wirestring;
int  **wire;
int  *mode;	

{
 char *p = *wirestring;
 char *cur;				/* current pointer in wirestring */
 char *dest;				/* temporary string of wire */
 int num = 0;				/* number of wires */
 int probe;				/* temporary value of current wire */
 int *work_array;			/* temporary work array of wires */
 int i, n;				/* indexes */

 /* max number of channels is 4 */	 
 work_array = (int *)calloc(4, sizeof(int));
 if (!work_array) { perror("calloc work_array: "); exit(1); }
 memset(work_array, 0, 4);

 dest = (char *)calloc(MaxString, sizeof(char));	     
 if (!dest) { perror("calloc dest: "); exit(1); }
 
 while (*p) {  cur = p; 
               n = 0;
               memset(dest, 0, MaxString);
               while(*cur && *cur != ',') { 
                if (!isdigit(*cur)) { printf("Error: symbols in option -w must be digits! - %s\n", wirestring); 
                                      exit(1);
                                     } else { cur++; n++; }
		                            
               }
               strncat(dest, p, n);
               probe = atoi(dest);
               
               /* must specify a valid mode */
               if (probe < 1 || probe > 4) { printf("Error: bad wire! - %d\n", probe); exit(1); }
               if (*mode <= 3 && probe > 2) { printf("Error: bad wire for mode=%d! - %d\n", *mode, probe); exit(1); }
               /* work_array must have different values */
               for (i=0; i<num; ++i) if (work_array[i] == probe) break;                
               if (i == num) {               
	         work_array[num] = probe;
	         num++;	       	  
	       }
	       p += ((int)strlen(dest) + 1);
             }
             
 if (num == 0) { printf("Error: you have selected 0 valid wires!\n"); exit(1); }
 
 free(dest);

 *wire = (int *)calloc(num, sizeof(int));
 if (!*wire) { perror("calloc wire: "); exit(1); }
 memset(*wire, 0, num);

 for (i=0; i<num; ++i) (*wire)[i] = work_array[i];
 free(work_array);
 
 return num;
}

/*+++++++++++++++++++ erase path part from filename +++++++++++++++++*/

char *SelectFileName (str)

char **str;	/* input string */

{
 char *s = *str;
 int len = (int)strlen(s);
 s += len;
 while (*(--s) != '/' && len--) ;
 return (++s);
}

/*+++++++++++++++++++++ Prepare names of output files ++++++++++++++++++++++++++*/

void PrepareOutNames (outfile, nwires, wire, outname, dir, ext, av)

char ***outfile;	/* array of names of output files */
int  *nwires;		/* number of wires */
int  **wire;		/* array of wire values */
char **outname;		/* base name for output files */ 
char **dir;		/* destination directory */
char **ext;		/* extension of output files */
char **av;		/* name of input file */

{ 
  int i;		/* index */
  char *base;		/* base string in sprintf function */

  base = (char *)calloc(MaxString, sizeof(char));
  if (!base) { perror("calloc base: "); exit(1); }
  
  for (i=0; i<*nwires; ++i) {
   memset((*outfile)[i], 0, MaxString);
   memset(base, 0, MaxString);
   
   if (!strcmp(*dir, "\0")) { 
    if (strcmp(*outname, "\0") && !strcmp(*ext, ".out")) sprintf((*outfile)[i], "%s", *outname);
      else { sprintf(base, "%s%s", (!strcmp(*outname, "\0") ? *av : *outname), *ext);
	     sprintf((*outfile)[i], base, (*wire)[i]);
            }
   } 
   else { 
    if (strcmp(*outname, "\0") && !strcmp(*ext, ".out")) sprintf((*outfile)[i], "%s/%s", *dir, *outname);
     else { sprintf(base, "%s/%s%s", *dir, (!strcmp(*outname, "\0") ? SelectFileName(av) : *outname), *ext);
            sprintf((*outfile)[i], base, (*wire)[i]);
           }
   }

  } /* for i*/
  
  free(base);
}

/*++++++++++++++++++++++++++++++++++ MAIN +++++++++++++++++++++++++++++++++++++*/

int main (int argc, char *argv[] )
{
  struct stat info;		/* input file status structure */
  long int bufsize = 1000000;	/* size of read buffer, default 1 MB, unless input file is smaller */
  int mode = -1;		/* data acquisition mode (default - wrong value) */
  char *wirestring;		/* value of wires from command line */
  int nwires;			/* number of wires to unpack */
  int *wire;			/* pointer to the array of numbers of wires */
  
  char *buffer;			/* buffer for packed data */
  signed char **cp;		/* cp means 'circular polarization' */
				/* massive of buffers for unpacked data */
				/* cp[0][] - wire N1 ; cp[1][] - wire N2 ; cp[2][] - wire N3 ; cp[3][] - wire N4 */
				
  int next;			/* index of first non-optioned argument in command line */
  char *outname;		/* the base name (without extension) of output files */
  char *dir;			/* path to the output files */
  char *ext;			/* the extension for names of output files */

  char smpwd;			/* N of single polar. complex samples in a 4 byte word */
  long int nsamples;		/* N of complex samples in each buffer */
  long int curbufsize;		/* current size of read buffer */
  long int curnsamples;		/* current nsamples */
  int bytesread;		/* number of bytes read from input file */
  int fdinput;			/* file descriptor of input file */
  int *fdoutput;		/* array of file descriptors of output files */
  char **outfile;		/* array of output file names */
  int i,k;			/* indexes */

  outname    = (char *)calloc(MaxString, sizeof(char));
  dir        = (char *)calloc(MaxString, sizeof(char));
  ext        = (char *)calloc(MaxString, sizeof(char));
  wirestring = (char *)calloc(MaxString, sizeof(char));
  if (!outname || !dir || !ext || !wirestring) 
   { perror("calloc outname, dir, ext, wirestring: "); exit(1); }
  memset(outname, 0, MaxString);    /* erase string */
  memset(dir, 0, MaxString);    
  memset(ext, 0, MaxString);    
  memset(wirestring, 0, MaxString);    
  sprintf(outname, "\0");	    /* make this string empty */
  sprintf(dir, "\0");	    
  sprintf(ext, ".out");	    
  sprintf(wirestring, "\0");	    

  /* parse the command line */
  next = ParseCmdLine(argc, argv, &mode, &wirestring, &dir, &ext, &bufsize, &outname);
  if (argc < next + 1) { 
   printf("Error: you have not selected input files!\n");
   exit (1);
  }
  
  switch (mode) {
    case  1: smpwd = 8;   break;
    case  2: smpwd = 4;   break;
    case  3: smpwd = 2;   break; 
    case  5: smpwd = 4;   break;
    case  6: smpwd = 2;   break;
    default: printf("Error: you must specify the mode of data, using option -m !\n");
             exit(0);
  }
  
  /* parse wirestring */
  if (!strcmp(wirestring, "\0") || !strcasecmp(wirestring, "all")) {
   if (mode == 5 || mode == 6) nwires = 4; else nwires = 2;   
   wire = (int *)calloc(nwires, sizeof(int));
   if (!wire) { perror("calloc wire: "); exit(1); }
   memset(wire, 0, nwires);
   for (i=0;i<nwires;++i) wire[i] = i + 1;
  }
  else nwires = ParseWires(&wirestring, &wire, &mode);

  nsamples = (int)rint(bufsize * smpwd / 4.0);
  curbufsize = bufsize;
  curnsamples = nsamples;

  /* allocate memory for fdoutput */
  fdoutput = (int *)calloc(nwires, sizeof(int));
  if (!fdoutput) { perror("calloc fdoutput: "); exit(1); }

  /* allocate memory for outfile */
  outfile = (char **)calloc(nwires, sizeof(char *));
  if (!outfile) { perror("calloc outfile: "); exit(1); }
  for (i=0; i<nwires; ++i) { 
   outfile[i] = (char *)calloc(MaxString, sizeof(char));
   if (!outfile[i]) { perror("calloc outfile[i]: "); exit(1); }
  }



  /* processing */
  printf("\nProcessing...\n");
  
  /* cycle of input files */
  for (k=next; k<argc; ++k) {
   printf(" [%3d] %s ... ", k-next, argv[k]); fflush(NULL);

  /* open file input */
#ifdef LARGEFILE
  if((fdinput = open(argv[k], O_RDONLY | O_LARGEFILE)) < 0 ) { perror("open input file"); exit(1); }
#else
  if((fdinput = open(argv[k], O_RDONLY)) < 0 ) { perror("open input file"); exit(1); }
#endif

  /* prepare names of output files */
  PrepareOutNames (&outfile, &nwires, &wire, &outname, &dir, &ext, &argv[k]);

  /* open file output */
#ifdef LARGEFILE
 for (i=0; i<nwires; ++i)
  if((fdoutput[i] = open(outfile[i], O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE, 0644)) < 0 ) {
   perror("open output file: ");
   exit(1);
  }
#else
 for (i=0; i<nwires; ++i)
  if((fdoutput[i] = open(outfile[i], O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0 ) {
   perror("open output file: ");
   exit(1);
  }
#endif
            
  /* check file size */
  if (stat (argv[k], &info) < 0) {
   perror("input file status: ");
   exit(1);
  }
  if (info.st_size % 4 != 0) {
   printf("Warning: file size %d is not a multiple of 4\n", info.st_size);
  }

  /* allocate storage */
  buffer = (char *)calloc(bufsize, sizeof(char));
  if (!buffer) { perror("calloc buffer: "); exit(1); }
  
  if (mode <= 3) cp = (signed char **)calloc(2, sizeof(signed char *));
   else cp = (signed char **)calloc(4, sizeof(signed char *));
  if (!cp) { perror("calloc cp: "); exit(1); }
  if (mode <= 3) for (i=0; i<2; ++i) { cp[i] = (signed char *)calloc(nsamples, sizeof(signed char));
                                       if (!cp[i]) { perror("calloc cp[i]: "); exit(1); }
                                      }
   else for (i=0; i<4; ++i) { cp[i] = (signed char *)calloc(nsamples, sizeof(signed char));           
                              if (!cp[i]) { perror("calloc cp[i]: "); exit(1); }
                             }  

  bufsize = curbufsize;
  nsamples = curnsamples;
    
  /* infinite loop */
  while (1)
    {
      /* read one buffer */
      bytesread = read(fdinput, buffer, bufsize);
      /* check for end of file */
      if (bytesread == 0) break;
      /* handle small buffers */
      if (bytesread != bufsize) 
	{
	  bufsize = bytesread;
	  nsamples = (int) rint(bufsize * smpwd / 4.0);
	}

      /* unpack */
      switch (mode) {
       case 1: unpack_pfs_2c2b(buffer, cp, bufsize); 
	       break;
       case 2: unpack_pfs_2c4b(buffer, cp, bufsize);
	       break;
       case 3: unpack_pfs_2c8b(buffer, cp, bufsize);
	       break;
       case 5: unpack_pfs_4c2b(buffer, cp, bufsize);
	       break;
       case 6: unpack_pfs_4c4b(buffer, cp, bufsize);
	       break;
      default: printf("Error: mode not implemented yet!\n"); 
               exit(1);
      }

      /* write data to output file */
      for (i=0; i<nwires; ++i) 
       if (nsamples * sizeof(signed char) != write(fdoutput[i], cp[wire[i]-1], nsamples * sizeof(signed char))) {
        perror("write: ");
        exit(1);
       }
    
     printf("."); fflush(NULL);
 	    
    } /* end of infinite loop */
    
    for (i=0; i<nwires; ++i) close(fdoutput[i]);

    free(buffer);
    if (mode <=3) for (i=0; i<2; ++i) free(cp[i]);
     else for (i=0; i<4; ++i) free(cp[i]);
    free(cp);
        
    printf(" - done\n");        
        
   } /* end of cycle of input files */
      
  free(wirestring);
  free(outname);
  free(dir);
  free(ext);
  free(wire);    
  free(fdoutput);
  for (i=0; i<nwires; ++i) free(outfile[i]);
  free(outfile);
      
  return 0;
}

