#include "unpack.h"

/******************************************************************************/
/*	unpack_pfs_2c2b							      */
/******************************************************************************/
void unpack_pfs_2c2b(char *buf, signed char **outbuf, int bufsize)
{
  /*
    unpacks 2-channel, 2-bit data from the portable fast sampler
    input array buf is of size bufsize bytes
    output array contains 2 arrays with 2*bufsize signed chars
    outbuf must have storage for at least 4*bufsize*sizeof(signed char)
  */
  
  /* order is board 1 channel A, board 1 channel B */
  /* with wiring of PFS, this corresponds to RCP-Q RCP-I */
  /* connect BBC sin and cos to I and Q, respectively, for positive freq */

  char value;
  signed char lookup[4] = {+3, +1, -1, -3}; 
  int i,k;
  
  k = 0;
  for (i = 0; i < bufsize; i += 4)
    {
      value = buf[i+1];
      outbuf[0][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[0][k+0] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+0] = lookup[value & 3];
      k += 2;

      value = buf[i+0];
      outbuf[0][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[0][k+0] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+0] = lookup[value & 3];
      k += 2;

      value = buf[i+3];
      outbuf[0][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[0][k+0] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+0] = lookup[value & 3];
      k += 2;

      value = buf[i+2];
      outbuf[0][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+1] = lookup[value & 3];
      value = value >> 2;
      outbuf[0][k+0] = lookup[value & 3];
      value = value >> 2;
      outbuf[1][k+0] = lookup[value & 3];
      k += 2;
    }

  return;
}

/******************************************************************************/
/*	unpack_pfs_2c4b   						      */
/******************************************************************************/
void unpack_pfs_2c4b(char *buf, signed char **outbuf, int bufsize)
{
  /*
    unpacks 2-channel, 4-bit data from the portable fast sampler
    input array buf is of size bufsize bytes
    output array outbuf contains 2 arrays with bufsize signed chars
    output array must have been allocated for at least 2*bufsize*sizeof(signed char)
  */

  /* order is board 1 channel A, board 1 channel B */
  /* with wiring of PFS, this corresponds to RCP-Q RCP-I */
  /* connect BBC sin and cos to I and Q, respectively, for positive freq */

  char value;
  signed char lookup[16] = {+15, +13, +11, +9, +7, +5, +3, +1, -1, -3, -5, -7, -9, -11, -13, -15}; 
  int i,k,m;
  
  k = 0; m = 0;
  for (i = 0; i < bufsize; i += 4)
    {
      value = buf[i+1];
      outbuf[0][k++] = lookup[value & 15];
      value = value >> 4;
      outbuf[1][m++] = lookup[value & 15];

      value = buf[i+0];
      outbuf[0][k++] = lookup[value & 15];
      value = value >> 4;
      outbuf[1][m++] = lookup[value & 15];

      value = buf[i+3];
      outbuf[0][k++] = lookup[value & 15];
      value = value >> 4;
      outbuf[1][m++] = lookup[value & 15];

      value = buf[i+2];
      outbuf[0][k++] = lookup[value & 15];
      value = value >> 4;
      outbuf[1][m++] = lookup[value & 15];
    }

  return;
}

/******************************************************************************/
/*	unpack_pfs_2c8b   						      */
/******************************************************************************/
void unpack_pfs_2c8b(char *buf, signed char **outbuf, int bufsize)
{
  /*
    unpacks 2-channel, 8-bit data from the portable fast sampler
    input array buf is of size bufsize bytes
    output array outbuf contains 2 arrays with 0.5*bufsize signed chars
    output array must have been allocated for at least bufsize*sizeof(signed char)
  */

  /* order is board 1 channel A, board 1 channel B */
  /* with wiring of PFS, this corresponds to RCP-Q RCP-I */
  /* connect BBC sin and cos to I and Q, respectively, for positive freq */

  int i,k,m;
  
  k = 0; m = 0;
  for (i = 0; i < bufsize; i+=4)
    {
      outbuf[0][k++] = ((unsigned char) buf[i+0]) - 128;
      outbuf[1][m++] = ((unsigned char) buf[i+1]) - 128;
      outbuf[0][k++] = ((unsigned char) buf[i+2]) - 128;
      outbuf[1][m++] = ((unsigned char) buf[i+3]) - 128;
    }

  return;
}

/******************************************************************************/
/*	unpack_pfs_4c2b							      */
/******************************************************************************/
void unpack_pfs_4c2b(char *buf, signed char **cp, int bufsize)
{
  /*
    unpacks 4-channel, 2-bit data from the portable fast sampler
    input array buf is of size bufsize bytes
    output array contains 4 arrays each with bufsize signed chars
    cp must have storage for at least 4*bufsize*sizeof(signed char)
  */
  
  /* order is board 1 channel A, board 1 channel B */
  /*          board 2 channel A, board 2 channel B */
  /* with wiring of PFS, this corresponds to RCP-Q RCP-I LCP-Q LCP-I */
  
  char value;
  signed char lookup[4] = {+3, +1, -1, -3};  
  int i,r;
  
  r = 0;
  for (i = 0; i < bufsize; i += 4) 
    {
      value = buf[i+1];
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      value = value >> 2;
      cp[2][r] = lookup[value & 3];
      value = value >> 2;
      cp[3][r] = lookup[value & 3];
      r++;

      value = buf[i+0];
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      value = value >> 2;
      cp[2][r] = lookup[value & 3];
      value = value >> 2;
      cp[3][r] = lookup[value & 3];
      r++;

      value = buf[i+3];
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      value = value >> 2;
      cp[2][r] = lookup[value & 3];
      value = value >> 2;
      cp[3][r] = lookup[value & 3];
      r++;

      value = buf[i+2];
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      value = value >> 2;
      cp[2][r] = lookup[value & 3];
      value = value >> 2;
      cp[3][r] = lookup[value & 3];
      r++;
    }

  return;
}

/******************************************************************************/
/*	unpack_pfs_4c4b							      */
/******************************************************************************/
void unpack_pfs_4c4b(char *buf, signed char **cp, int bufsize)
{
  /*
    unpacks 4-channel, 4-bit data from the portable fast sampler
    input array buf is of size bufsize bytes
    output array contains 4 arrays each with 0.5*bufsize signed chars
    cp must have storage for at least 2*bufsize*sizeof(signed char)
  */
  
  /* order is board 1 channel A, board 1 channel B */
  /*          board 2 channel A, board 2 channel B */
  /* with wiring of PFS, this corresponds to RCP-Q RCP-I LCP-Q LCP-I */
  
  char value;
  signed char lookup[16] = {+15, +13, +11, +9, +7, +5, +3, +1, -1, -3, -5, -7, -9, -11, -13, -15}; 
  int i,r;
  
  r = 0;
  for (i = 0; i < bufsize; i += 4) 
    {
      value = buf[i+1];
      cp[2][r] = lookup[value & 15];
      value = value >> 4;
      cp[3][r] = lookup[value & 15];

      value = buf[i+0];
      cp[0][r] = lookup[value & 15];
      value = value >> 4;
      cp[1][r] = lookup[value & 15];
      r++;

      value = buf[i+3];
      cp[2][r] = lookup[value & 15];
      value = value >> 4;
      cp[3][r] = lookup[value & 15];

      value = buf[i+2];
      cp[0][r] = lookup[value & 15];
      value = value >> 4;
      cp[1][r] = lookup[value & 15];
      r++;
    }

  return;
}
