
/*
   This implements a small subset of EDT library calls using Jeff Hagen's 
   open source edt driver. The purpose of this is to
   make our solaris code compatible with linux code. It is not meant
   to fully implement the edt gobblty goop. 

   Jeff's driver works differently that the EDT driver

      STARTDMA always flushes the fifo and there is no way to stop DMA
      and start it again where you left off. Therefore

      edt_flush_fifo is a null.
      edt_start_buffers does the fifo flush before starting.
      edt_reg_write presumes the second parameter is PCD_FUNCT
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#include "edt.h"
#include "edtinc.h"

 
/* open edt device */

struct EDTLIB *edt_open( name, param )
char *name;
int param;
{
  struct EDTLIB *e;
  int fd, verbose;

  if( (fd = open( "/dev/edt0", O_RDWR, 0 ))<0 ) {
    perror( "failed to open edt device");
    return(NULL);
  }

  verbose = 0;
  if( ioctl( fd, EDT_VERBOSE, &verbose ) < 0 )
    perror("edt verbose");

  /* verbose = 9; */
  verbose = 0;
  if( ioctl( fd, EDT_BYTESWAP, &verbose ) < 0 )
    perror("edt byteswap");

  e = (struct EDTLIB *)malloc( sizeof(struct EDTLIB));
  bzero( e, sizeof(struct EDTLIB ));
  e->fd = fd;
  return(e);
}


void *edt_wait_for_buffers( edt, count )
struct EDTLIB *edt;
int count;
{
  int i;
  void *ret;

  for( i=0; i<count; i++ ) {
    if( ioctl( edt->fd, EDT_NEXTBUF, &edt->next ) < 0 ) {
      edt->overflow = edt->next.total;
      perror("ioctl next buf overflow");
      break;
    }
  }
  if( edt->next.next == 0 )
    usleep(10);

  if( edt->next.fifoover & EDT_OVERFLOW )
    fprintf(stderr, "ERROR fifo overflow, data coming too fast" );

  /* printf("read buflen %d next %d total %d diff %d\n", edt->buflen, edt->next.next, edt->next.total, edt->next.diff ); */
  return( (void *)(&(edt->map[ edt->buflen * edt->next.next])) );
}

int edt_done_count( edt )
struct EDTLIB *edt;
{
  return( edt->next.total );
}

int edt_start_buffers( edt, count )
struct EDTLIB *edt;
int count;
{
  if( count != 0 )
    fprintf( stderr, "edt_start_buffers unsupported mode\n");

   edt->overflow = 0;
  if( ioctl( edt->fd, EDT_STARTDMA, 0 ) < 0 )
    perror("start dma");
}

int edt_stop_buffers( edt )
struct EDTLIB *edt;
{
  if( ioctl( edt->fd, EDT_STOPDMA, 0 ) < 0 )
    perror("start dma");
}

int edt_ring_buffer_overrun(edt) 
struct EDTLIB *edt;
{
  return( edt->overflow != 0 );
}


int edt_reset_ring_buffers(edt, val)
struct EDTLIB *edt;
int val;
{
  return(1);
}

int edt_flush_fifo( edt )
struct EDTLIB *edt;
{
  return(0);
}


/* which is ignored, but PCD_FUNCT is the value assumed */

int edt_reg_write( edt, which, value )
struct EDTLIB *edt;
unsigned int which;
unsigned int value;
{
  struct EDTFUNC func;

  func.func = value;
  func.stat = 0;

  return ioctl( edt->fd, EDT_SETFUNC, &func );
}
  

int edt_close(edt)
struct EDTLIB *edt;
{
  if( ioctl( edt->fd, EDT_STOPDMA, 0 )<0 ) {
    perror("edt stopdma");
    return(-1);
  }

  if( edt->map )
    munmap( (void *)edt->map, edt->cfg.len );

  close( edt->fd );
  bzero( edt, sizeof(struct EDTLIB ));
  edt->fd = -1;

  return(0);
}

edt_perror(s)
char *s;
{
  perror(s);
}

int edt_configure_ring_buffers( edt, len, num, rw, data )
struct EDTLIB *edt;
int len;
int num;
int rw;
void *data[];
{
  int i, j, delay;
  unsigned char low;

  edt->cfg.nbufs = num;
  edt->cfg.len = len;

  if( len % 4096 == 0 )
    edt->buflen = edt->cfg.len;
  else
    edt->buflen = 4096*(edt->cfg.len / 4096  + 1);

  if( ioctl( edt->fd, EDT_CONFIGDMA, &edt->cfg ) < 0 ) {  /* alloc kmem */
    perror("config dma");
    return(-1);
  }

  if( !( edt->map = (void *)mmap( 0, edt->cfg.nbufs*edt->buflen, 
         PROT_READ|PROT_WRITE, MAP_PRIVATE, edt->fd, 0 )) ){
    perror("mmap");
    return(-1);
  }

/* assign return value pointers */
/* and nopage all the memory in to ram */

  for( i=0; i<edt->cfg.nbufs; i++ ) { 
    if( data )
      data[i] = &edt->map[ edt->buflen *i ];
    for( j=0; j<edt->buflen; j +=4096 )
      low = edt->map[edt->buflen * i + j];
  }

  delay = 1000;  /* jiffies */
  if( ioctl( edt->fd, EDT_NEXTDELAY, &delay ) < 0 )
    perror("edt next delay");

  return(0);
}


