/************************************************************************************
*  program pfs2wire
*  This programs unpacks data from the portable fast sampler.
*  It is specialized to work only with data recorded with mode 5
*  Output is binary signed char point quantities that are 
*  interchangeable real and imaginary parts
*
*  Usage:
*  	pfs2wire [-d directory]
*                [-b buffer_size]
*                [-o outname]      [infile1] ... [infileN]
*
*  Options:
*
*      the -d argument specifies the directory for output files
*             (default the same as for the input files)
*      the -b buffer_size specifies the size of read buffer
*             (default is 1,000,000 bytes)
*      the -o outname specifies the base name (without extension) of output files
*             (default the output filename is input filename + .R/.L)
*      the -h option prints help
*
*************************************************************************************/

/* these preprocessor' defines need to support Large Files */

# ifdef LARGEFILE
#   define _GNU_SOURCE 1
#   define _FILE_OFFSET_BITS 64
# endif


# include <math.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <asm/fcntl.h>
# include <sys/stat.h>
# include <unistd.h>

# define MaxString 255


void Help ();			/* print help */		
int ParseCmdLine ();		/* parse command line */
char *SelectFileName ();	/* erase path part from filename */
void PrepareOutNames ();	/* prepare names of output files */

/*+++++++++++++++++++++++++++++++++++ Prompting ++++++++++++++++++++++++++++++++++++++*/

void Help (s)

char *s;	/* name of the programm */

{
  printf("\nPROGRAMM %s\n"
         "It is used to unpack data from the Portable Fast Sampler\n"
	 "It is specialized to work only with data recorded with mode 5\n"
         "Output is binary signed char point quantities that are\n" 
         "interchangeable real and imaginary parts\n\n"
	 "[USAGE]: %s [options] <file(s) to process>\n\n"
	 "[OPTIONS]:\n"
	 "-d <directory>   - path to the output files, default - the same as for the input ones\n"
	 "-b <buffer_size> - size of read bufer, default - 1,000,000 bytes\n"
	 "-o <outname>     - base name of file for output data\n"
	 "                   default - input filename + .R/.L otherwise <outname> + .R/.L\n"
	 "-h               - help\n\n", s, s);
  exit(0);
}

/*++++++++++++++++++++++ Parsing the command line +++++++++++++++++++++++++++++++++++++*/

int ParseCmdLine (argc, argv, dir, bufsize, outname)

int       argc;		/* number of arguments */
char      **argv;	/* command line arguments */
char      **dir;
long int  *bufsize;
char      **outname;

{
  int op;		/* option letter returned by getopt */

  extern char *optarg; 	/* if arg with option, this pts to it */
  extern int optind;	/* after call, ind into argv for next */

  /* loop over all the options in list */
  while ((op = getopt(argc, argv, "d:b:o:h")) != -1)
    switch(op) {
      case 'h':
        Help(argv[0]);
      break;
      
      case 'b':
        *bufsize = atol(optarg);
	
	/* must specify a valid mode */
	if (*bufsize < 1 || *bufsize > 100000000) {
	 printf("Error: bufsize is out of range: %ld !\n", *bufsize);
	 exit(1);
	}
      break;
      
      case 'd':
	  if (*dir) { *dir = 0;
	              free(*dir);
	            }
	  if ( !(*dir = (char *)calloc((int)strlen(optarg), sizeof(char *)) ) ) {
	    perror("calloc dir: ");
	    exit(1);
	   }
	  memcpy (*dir, optarg, (int)strlen(optarg));
      break;
      
      case 'o':
	  if ( *outname ) { *outname = 0;
	                    free(*outname);
	                 }
	  if ( !(*outname = (char *)calloc((int)strlen(optarg), sizeof(char *)) ) ) {
	    printf("calloc outname: ");
	    exit(1);
	   }
	  memcpy (*outname, optarg, (int)strlen(optarg));
      break;
      
      case '?':
        Help(argv[0]);
      break;
    }
    
  return(optind);
}

/*+++++++++++++++++++ erase path part from filename +++++++++++++++++*/

char *SelectFileName (str)

char **str;	/* input string */

{
 char *s = *str;
 int len = (int)strlen(s);
 s += len;
 while (*(--s) != '/' && len--) ;
 return (++s);
}

/*+++++++++++++++++++++ Prepare names of output files ++++++++++++++++++++++++++*/

void PrepareOutNames (outfile, nout, outname, dir, av)

char ***outfile;	/* array of names of output files */
int  *nout;             /* number of output files */
char **outname;		/* base name for output files */ 
char **dir;		/* destination directory */
char **av;		/* name of input file */

{ 
  int i;		/* index */

  for (i=0; i<*nout; ++i) {
   
   if (!strcmp(*dir, "\0")) 
     sprintf((*outfile)[i], "%s.%s", (!strcmp(*outname, "\0") ? *av : *outname), i==0 ? "R": "L"); 
   else 
     sprintf((*outfile)[i], "%s/%s.%s", *dir, (!strcmp(*outname, "\0") ? SelectFileName(av) : *outname), i==0 ? "R": "L");

  } /* for i*/
}

/******************************************************************************/
/*      unpack_pfs_4c2b                                                       */
/******************************************************************************/
void unpack_pfs_4c2b(char *buf, signed char **cp, int bufsize)
{
  /*
    unpacks 4-channel, 2-bit data from the portable fast sampler
    input array buf is of size bufsize bytes
    output array cp contains 2 arrays each with 2*bufsize signed chars
    cp must have storage for at least 2*2*bufsize*sizeof(signed char)
  */

  /* cp[0] = RCP:  RCP-I RCP-Q RCP-I RCP-Q ... */
  /* cp[1] = LCP:  LCP-I LCP-Q LCP-I LCP-Q ... */

  char value;
  signed char lookup[4] = {+3, +1, -1, -3};
  int i,r;

  r = 0;
  for (i = 0; i < bufsize; i += 4)
    {
      value = buf[i+1];
      cp[0][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      r+=2;

      value = buf[i+0];
      cp[0][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      r+=2;

      value = buf[i+3];
      cp[0][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      r+=2;

      value = buf[i+2];
      cp[0][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[0][r] = lookup[value & 3];
      value = value >> 2;
      cp[1][r+1] = lookup[value & 3];
      value = value >> 2;
      cp[1][r] = lookup[value & 3];
      r+=2;
    }

  return;
}

/*++++++++++++++++++++++++++++++++++ MAIN +++++++++++++++++++++++++++++++++++++*/

int main (int argc, char *argv[] )
{
  struct stat info;		/* input file status structure */
  long int bufsize = 1000000;	/* size of read buffer, default 1 MB, unless input file is smaller */
  
  char *buffer;			/* buffer for packed data */
  signed char **cp;		/* cp means 'circular polarization' */
				/* massive of buffers for unpacked data */
				/* cp[0][] - wire N2 + wire N1; cp[1][] - wire N4 + wire N3 */
				
  int next;			/* index of first non-optioned argument in command line */
  char *outname;		/* the base name (without extension) of output files */
  char *dir;			/* path to the output files */

  char smpwd = 4;		/* N of single polar. complex samples in a 4 byte word */
  long int nsamples;		/* N of complex samples in each buffer */
  long int curbufsize;		/* current size of read buffer */
  long int curnsamples;		/* current nsamples */
  int bytesread;		/* number of bytes read from input file */
  int fdinput;			/* file descriptor of input file */
  int nout = 2;                 /* number of output files */
  int *fdoutput;		/* array of file descriptors of output files */
  char **outfile;		/* array of output file names */
  int i,k;			/* indexes */

  outname    = (char *)calloc(MaxString, sizeof(char));
  dir        = (char *)calloc(MaxString, sizeof(char));
  if (!outname || !dir) 
   { perror("calloc outname, dir: "); exit(1); }
  memset(outname, 0, MaxString);    /* erase string */
  memset(dir, 0, MaxString);    
  sprintf(outname, "\0");	    /* make this string empty */
  sprintf(dir, "\0");	    

  /* parse the command line */
  next = ParseCmdLine(argc, argv, &dir, &bufsize, &outname);
  if (argc < next + 1) { 
   printf("Error: you have not selected input files!\n");
   exit (1);
  }
  
  nsamples = (int)rint(bufsize * smpwd / 4.0);
  curbufsize = bufsize;
  curnsamples = nsamples;

  /* allocate memory for fdoutput */
  fdoutput = (int *)calloc(nout, sizeof(int));
  if (!fdoutput) { perror("calloc fdoutput: "); exit(1); }

  /* allocate memory for outfile */
  outfile = (char **)calloc(nout, sizeof(char *));
  if (!outfile) { perror("calloc outfile: "); exit(1); }
  for (i=0; i<nout; ++i) { 
   outfile[i] = (char *)calloc(MaxString, sizeof(char));
   if (!outfile[i]) { perror("calloc outfile[i]: "); exit(1); }
  }



  /* processing */
  printf("\nProcessing...\n");
  
  /* cycle of input files */
  for (k=next; k<argc; ++k) {
   printf(" [%3d] %s ... ", k-next, argv[k]); fflush(NULL);

  /* open file input */
#ifdef LARGEFILE
  if((fdinput = open(argv[k], O_RDONLY | O_LARGEFILE)) < 0 ) { perror("open input file"); exit(1); }
#else
  if((fdinput = open(argv[k], O_RDONLY)) < 0 ) { perror("open input file"); exit(1); }
#endif

  /* prepare names of output files */
  PrepareOutNames (&outfile, &nout, &outname, &dir, &argv[k]);

  /* open file output */
#ifdef LARGEFILE
 for (i=0; i<nout; ++i)
  if((fdoutput[i] = open(outfile[i], O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE, 0644)) < 0 ) {
   perror("open output file: ");
   exit(1);
  }
#else
 for (i=0; i<nout; ++i)
  if((fdoutput[i] = open(outfile[i], O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0 ) {
   perror("open output file: ");
   exit(1);
  }
#endif
            
  /* check file size */
  if (stat (argv[k], &info) < 0) {
   perror("input file status: ");
   exit(1);
  }
  if (info.st_size % 4 != 0) {
   printf("Warning: file size %d is not a multiple of 4\n", info.st_size);
  }

  /* allocate storage */
  buffer = (char *)calloc(bufsize, sizeof(char));
  if (!buffer) { perror("calloc buffer: "); exit(1); }
  
  cp = (signed char **)calloc(2, sizeof(signed char *));
  if (!cp) { perror("calloc cp: "); exit(1); }
  for (i=0; i<2; ++i) { cp[i] = (signed char *)calloc(2*nsamples, sizeof(signed char));
                        if (!cp[i]) { perror("calloc cp[i]: "); exit(1); }
                       }

  bufsize = curbufsize;
  nsamples = curnsamples;
    
  /* infinite loop */
  while (1)
    {
      /* read one buffer */
      bytesread = read(fdinput, buffer, bufsize);
      /* check for end of file */
      if (bytesread == 0) break;
      /* handle small buffers */
      if (bytesread != bufsize) 
	{
	  bufsize = bytesread;
	  nsamples = (int) rint(bufsize * smpwd / 4.0);
	}

      /* unpack */
       unpack_pfs_4c2b(buffer, cp, bufsize);

      /* write data to output file */
      for (i=0; i<nout; ++i) 
       if (2*nsamples * sizeof(signed char) != write(fdoutput[i], cp[i], 2*nsamples * sizeof(signed char))) {
        perror("write: ");
        exit(1);
       }
    
     printf("."); fflush(NULL);
 	    
    } /* end of infinite loop */
    
    for (i=0; i<nout; ++i) close(fdoutput[i]);

    free(buffer);
    for (i=0; i<2; ++i) free(cp[i]);
    free(cp);
        
    printf(" - done\n");        
        
   } /* end of cycle of input files */
      
  free(outname);
  free(dir);
  free(fdoutput);
  for (i=0; i<nout; ++i) free(outfile[i]);
  free(outfile);
      
  return 0;
}

