/*******************************************************************************
*  program pfs_radar
*  $Id: pfs_radar.c,v 2.1 2002/05/12 16:00:54 cvs Exp $
*  This program performs data acquitisition using the Portable Fast Sampler
*  and an EDT PCI CD-x0 interface card.
*
*  usage:
*       pfs_radar 
*       [-start yyyy,mm,dd,hh,mn,sc] 
*       [-secs sec] [-step sec] [-cycles c] 
*       [-files f] [-rings r] 
*       [-bytes b] [-log l] [-code len]
*       -dir d [-dir d]... 
*
*  input:
*       the input parameters are typed in as command line arguments
*
*  output:
*       the output data is streamed to disk
*
*  Original program written by Jeff Hagen.
*  Written in the spirit of the Stewart Anderson wptape code,
*  this program reads a 16-bit wide EDT PCI CD-x0 card
*  and writes the data onto disk.  
*
*  compile with 
*  gcc -O2 -o pfs_radar -I/opt/EDTpcd pfs_radar.c /opt/EDTpcd/libedt.c -lpthread
*
*******************************************************************************/

/* 
   $Log: pfs_radar.c,v $
   Revision 2.1  2002/05/12 16:00:54  cvs
   Changed default values of i/o buffer sizes.

   Revision 2.0  2002/04/27 08:01:04  margot
   Variety of changes to accomodate POSIX threads and multiple output files.

   Revision 1.4  2000/11/01 02:17:42  margot
   Replaced -mode with -m for consistency.

   Revision 1.3  2000/10/30 22:03:40  margot
   Stop time now based on clock rather than buffer count.

   Revision 1.2  2000/10/30 04:32:32  margot
   Fixed three bugs corresponding to gsb_radar revisions 2.2 -> 2.4
   1) thr_join() did not join on previous diskwrite thread.
   Fixed bug by adding diskwrite structure wlast.
   2) Removed all instances of diskbufs and made number of disk write buffers
   equal to number of input ring buffers.  Previously there was a correspondence
   between the number of output files and the number of disk write buffers,
   with unwanted results when specifying a single output file.
   3) Fixed bug in requirement for write buffers to be multiples of
   pagesize for mlock().

   Revision 1.1  2000/07/20 21:12:07  margot
   Initial revision

*/

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/procfs.h>
#include <pthread.h>


#ifdef SOLARIS 
#include <sys/priocntl.h>
#include <sys/rtpriocntl.h>
#include <sys/tspriocntl.h>
#else
#include <stdlib.h> 
#endif

#include "fcntl.h"
#include "edtinc.h"
#include "multifile.h"

/* revision control variable */
static char const rcsid[] = 
"$Id: pfs_radar.c,v 2.1 2002/05/12 16:00:54 cvs Exp $";

void schedule_rt( int );

struct DISKWRITE { /* one of these for each diskbuffer allocated */
  struct MULTIFILE *fd;
  int tape_fd;
  int len;
  char *out;
  pthread_t proc;
};

struct RADAR { /* structure that holds the buffers and configuration */
  EdtDev *edt;
  unsigned int mode;
  int ameg;
  int secs;
  int step;
  int cycles;
  int ringbufs;
  int nfiles;
  char *dir;     /* disk directory if selected */ 
  char *istape;  /* tape device if selected */ 
  int dw_count;  /* current index into dw_multi */
  int dw_multi;  /*  ->out has size of dw_multi*ameg */
  struct DISKWRITE dw[2];
  unsigned short **rings;
  time_t start;
  time_t stop;
  time_t next;
  time_t startmone;
  char timestr[80];
  char log[80];
  FILE *logfd;
  int pack;
} radar;

#define AMEG (1000*1000)	/* default size of edt ring buffer */
#define RINGBUFS  64		/* default number of one meg edt ring buffers */
#define SECS   9000		/* default number of seconds to take */
#define AFEWSECS  3		/* interval bw key pressed and toggle EDT bit */

int ctlc_flag = 0;

int main(int argc, char *argv[])
{
  int i,cycle;
  char *p;
  unsigned char *data;
  struct DISKWRITE *w;
  struct DISKWRITE *wlast;
  int dcount;
  void *disk_write();
  struct RADAR *r;
  unsigned int off=0x00;
  unsigned int on=0x01;
  struct tm go;
  int time_set = 0;

  long long size;
  long  lcode = 7812500, lfft = 128;

#ifdef TIMER
  struct timeval   now;
  struct timeval   then;
  struct itimerval mytimer;
#endif

  /* set defaults */
  r = &radar;
  bzero( r, sizeof(struct RADAR ));
  r->ringbufs = RINGBUFS;
  r->ameg = AMEG;
  r->pack = 1;
  r->secs = SECS;
  r->step = 0;
  r->cycles = 1;
  r->nfiles = 40;
  r->dw_multi = 20;
  w = &r->dw[0];
  wlast = NULL;
  r->istape = NULL;

  r->dir = NULL;

  /* process the command line */
  for( i=1; i<argc; i++ ) {
    p = argv[i];
    if( strncasecmp( p, "-secs", strlen(p) ) == 0 ) {
      p = argv[++i];
      if( (r->secs = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -secs\n");
        pusage();
      }
    } else if( strncasecmp( p, "-step", strlen(p) ) == 0 ) {
      p = argv[++i];
      if((r->step = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -step\n");
        pusage();
      }
    } else if( strncasecmp( p, "-cycles", strlen(p) ) == 0 ) {
      p = argv[++i];
      if((r->cycles = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -cycles\n");
        pusage();
      }
    } else if( strncasecmp( p, "-files", strlen(p) ) == 0 ) {
      p = argv[++i];
      if((r->nfiles = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -files\n");
        pusage();
      }
    } else if( strncasecmp( p, "-m", strlen(p) ) == 0 ) {
      p = argv[++i];
      if((r->mode = atoi(p))<0 ) {
        fprintf(stderr, "bad value for -m\n");
        pusage();
      }
    } else if( strncasecmp( p, "-start", strlen(p) ) == 0 ) {
      p = argv[++i];
      if((sscanf(p,"%d,%d,%d,%d,%d,%d",
 		 &go.tm_year,&go.tm_mon,&go.tm_mday,
 		 &go.tm_hour,&go.tm_min,&go.tm_sec) != 6)) {
	fprintf(stderr, "bad value for -start\n");
	pusage();
      } else time_set = 1;
    } else if( strncasecmp( p, "-dir", strlen(p) ) == 0 ) {
      if( r->dir ) {
        fprintf(stderr, "one -dir switch only\n" );
        pusage();
      }
        
      r->dir = argv[++i];
      if( access( r->dir, W_OK|X_OK )) {
        fprintf(stderr, "unable to access directory %s\n", r->dir );
        pusage();
      }
    } else if( strncasecmp( p, "-rings", strlen(p) ) == 0 ) {
      p = argv[++i];
      if(( r->ringbufs = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -rings\n");
        pusage();
      }
    } else if( strncasecmp( p, "-bytes", strlen(p) ) == 0 ) {
      p = argv[++i];
      if(( r->ameg = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -bytes\n");
        pusage();
      }
    } else if( strncasecmp( p, "-log", strlen(p) ) == 0 ) {
      p = argv[++i];
      strcpy( r->log, p);
    } else if( strncasecmp( p, "-nopack", strlen(p) ) == 0 ) {
      r->pack = 0;
    } else if( strncasecmp( p, "-tape", strlen(p) ) == 0 ) {
      if( r->istape ) {
        fprintf(stderr, "one -tape switch only\n" );
        pusage();
      }
      r->istape = argv[++i];
    } else if( strncasecmp( p, "-code", strlen(p) ) == 0 ) {
      p = argv[++i];
      if(( lcode = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -code\n");
        pusage();
      }
    }  else if( strncasecmp( p, "-fft", strlen(p) ) == 0 ) {
      p = argv[++i];
      if(( lfft = atoi(p))<=0 ) {
        fprintf(stderr, "bad value for -fft\n");
        pusage();
      }
    }
  }

  /* set maximum size of individual datafiles */
  size = lcode * lfft;
  while (size < 1000000000) { size = size * 2; }
  multi_config_maxfilesize ((long long) size); 

  /* check that sampling mode is valid */
  switch (r->mode)
    {
    case 1: break;
    case 2: break;
    case 3: break;
    case 5: break;
    case 6: break;
    default: fprintf(stderr,"invalid mode\n"); pusage();
    }

  if (r->cycles > 1 && r->step < r->secs + AFEWSECS)
    {
      fprintf(stderr,"Step size must be bigger than duration of A/D\n");
      set_kb(0);
      exit(1);
    }

  if( r->dir && r->istape ) {
      fprintf(stderr,"Cant have disk and tape selected at once\n");
      set_kb(0);
      exit(1);
  }

  if( r->ringbufs <=0 )
    r->ringbufs = RINGBUFS;

  if( r->ameg <=0 )
    r->ameg = AMEG;

  if( !(r->dir || r->istape) ) {
    fprintf(stderr, "At least one -dir switch or -tape switch is required\n");
    pusage();
  }
  
  /* set scheduling priority */
  schedule_rt(2); 

  printf("Starting the Portable Fast Sampler\n");

  /* open edt device */

  if ((r->edt = edt_open("edt", 0)) == NULL)
    {
      perror("edt_open") ;
      set_kb(0);
      exit(1) ;
    }
  set_kb(1);

  /* block trigger immediately */
  edt_reg_write( r->edt, PCD_FUNCT, 0x00 | (r->mode << 1));

  /* allocate input buffers */
  allocate_ringbufs(r);

  /* allocate output buffers */
  allocate_writebufs(r);

  /* open log file */
  open_log(r);

  /* compute starting time */
  if (time_set)
    {
      printf("A/D will start on second tick at %04d %02d %02d %02d %02d %02d\n",
	     go.tm_year,go.tm_mon,go.tm_mday,
	     go.tm_hour,go.tm_min,go.tm_sec);
      /* take care of unusual definitions for month, day */
      go.tm_year = go.tm_year - 1900;
      go.tm_mon = go.tm_mon - 1;
      r->next = mktime(&go);
    }
  /* otherwise let user hit the go button */
  else
    {
      printf("Hit a key when ready to take data\n");
      printf("A/D will start on second tick following key_press event + %d seconds\n",AFEWSECS-1);
      getchar();

      /* get next start time */
      r->next = time(NULL) + AFEWSECS;
    }
  
  /* loop over desired number of cycles */
  for (cycle = 1; cycle <= r->cycles; cycle++)
    {
      /* define start times for this cycle and the next */
      r->start = r->next;
      r->stop  = r->start + r->secs;
      r->startmone = r->start - 1;
      r->next = r->start + r->step;
      set_kb(1);

      /* obtain time string and open files */
      get_tms(r->start,r->timestr); 
      open_files(r);

      fprintf(stdout  , "\nCycle %d will start at %s\n" , cycle, r->timestr );
      fprintf(r->logfd, "\nCycle %d starting at %s\n" , cycle, r->timestr );
      fflush(r->logfd);

      edt_flush_fifo( r->edt );
      edt_start_buffers( r->edt, 0 );
      
      /* wait till .5 sec before expected pulse */
      wait_till_start(r->startmone); 
      /* arm trigger */
      edt_reg_write( r->edt, PCD_FUNCT, 0x01 | (r->mode << 1));  

#ifdef TIMER
      /* set current time */
      now.tv_sec = 100000;
      now.tv_usec = 0;
      /* set next timer expiration a long time from now */
      mytimer.it_value.tv_sec  = now.tv_sec;
      mytimer.it_value.tv_usec = now.tv_usec;
      /* reload timer with incremental values  */
      mytimer.it_interval.tv_sec = now.tv_sec;
      mytimer.it_interval.tv_usec = now.tv_usec;
      /* set timer going */
      if (!setitimer(ITIMER_REAL, &mytimer, NULL))
	perror("setitimer");
#endif

      /* main loop */
      dcount = 0;
      for( i=0; ; i++ ) {
	if( ctlc_flag )
	  break;
	if (time(NULL) >= r->stop)
	  break;
 	if(!(data = edt_wait_for_buffers( r->edt, 1)))
	  printf("error \n");
	else {
	  if( i%50 == 0 ) {
	    printf("\ni = %d count = %d\n", i, edt_done_count(r->edt));
            fflush( stdout );
          }
	  fprintf(stderr,".");
	  if( edt_ring_buffer_overrun(r->edt)) {
	    printf("overrun %d\n",i );
	    fprintf(r->logfd, "overrun %d\n" , i );
	  } else {
#ifdef TIMER
	    /* read current timer value */
	    getitimer(ITIMER_REAL, &mytimer);
	    now = mytimer.it_value;
	    fprintf(r->logfd,"%6d %06ld %06ld %20ld\n",
		    i,then.tv_usec,now.tv_usec,
		    1000000*(then.tv_sec-now.tv_sec)+then.tv_usec-now.tv_usec); 
	    then.tv_sec = now.tv_sec;
	    then.tv_usec = now.tv_usec;
#endif TIMER

	    /* copy data from EDT to disk write output buffer */

	    memcpy(&w->out[r->dw_count*r->ameg], data, r->ameg);
            if( ++r->dw_count >= r->dw_multi ) {
              r->dw_count = 0;

  	      if( wlast && wlast->proc)
	        if(pthread_join( wlast->proc, NULL ))
                  perror("pthread_join");
	      if( pthread_create( &w->proc, NULL, disk_write, w ))
	        perror("pthread_create");

	      wlast = w;
              if( w == &r->dw[0] )
  	        w = &r->dw[1];
              else
  	        w = &r->dw[0];
            }
	  }
	}
      }

      if( ctlc_flag ) {
	printf("\nStopped by user, read %d buffers\n\n\n", i );
	fprintf(r->logfd, "Stopped by user, read %d buffers\n\n\n", i );
      } else {
	printf("\nFinished, read %d buffers\n\n\n", i );
	fprintf(r->logfd, "Finished, read %d buffers\n\n\n", i );
      }
      
      if( wlast && wlast->proc)
        if(pthread_join( wlast->proc, NULL ))
          perror("pthread_join");

      wlast = NULL;

      if( r->dw_count > 0 ) {

        printf("writing last buffer to disk\n");
        fflush(stdout);

        w->len = r->dw_count*r->ameg;
        if( pthread_create( &w->proc, NULL, disk_write, w ))
           perror("pthread_create");

        if(pthread_join( w->proc, NULL ))
          perror("pthread_join");

        w->len = r->dw_multi*r->ameg;
        r->dw_count = 0;
      }
      
      edt_stop_buffers( r->edt);
      /* clear trigger */
      edt_reg_write( r->edt, PCD_FUNCT, 0x00 | (r->mode << 1)); 
      edt_reset_ring_buffers(r->edt, 0);
      close_files(r);
    }
      
  edt_close(r->edt);
  fclose(r->logfd);
  set_kb(0);
}

/* disk_write thread */

void *disk_write( w )
struct DISKWRITE *w;
{
  int writ;
  int k;
  
  if( w->tape_fd >= 0 )
    tape_write( w );
  else {
    if((writ = multi_write( w->fd, w->out, w->len))!= w->len ) 
      printf(" disk write error: could only write %d bytes\n", writ );
  }
    
  return(0);
}

tape_write(w)
struct DISKWRITE *w;
{
  int writ;

  if(( writ = write( w->tape_fd, w->out, w->len )) != w->len )
      printf(" tape write error: could only write %d bytes\n", writ );
}

#ifdef SOLARIS

/* courtesy of Stuart Anderson, swiped right out of proc_ut.c */

void
schedule_rt( int rt_priority )
{
    pcparms_t pcparms;
    rtparms_t *rtparmsp;
    pcinfo_t pcinfo;
    id_t rtID;
    short maxrtpri;
    rtinfo_t *rtinfop;


    /*
     * boost process to the RT scheduling class with rt_priority.
     */

    (void) strcpy (pcinfo.pc_clname, "RT");
    if (priocntl (0L, 0L, PC_GETCID, (caddr_t) &pcinfo) == -1L) {
	perror ("PC_GETCID failed for realtime class");
        set_kb(0);
	exit (1);
    }
    
    rtID = pcinfo.pc_cid;
    rtinfop = (struct rtinfo *)pcinfo.pc_clinfo;
    maxrtpri = rtinfop->rt_maxpri;
    
    pcparms.pc_cid = rtID;
    rtparmsp = (struct rtparms *)pcparms.pc_clparms;
    rtparmsp->rt_pri = rt_priority;
    rtparmsp->rt_tqnsecs = RT_TQDEF;
    
    if (priocntl (P_PID, getpid(), PC_SETPARMS, (caddr_t) &pcparms) == -1) {
	perror ("PC_SETPARMS failed");
    }
}

#else
void
schedule_rt( int rt_priority )
{
}
#endif


/* 
  allocate output buffers using malloc and calling mlock
  r is the config structure
*/

allocate_writebufs(r)
struct RADAR *r;
{
  int pagesize, i, aout;
  struct DISKWRITE *w;

  aout = r->ameg;

#ifdef SOLARIS
  pagesize = sysconf(_SC_PAGESIZE);
  if (aout%pagesize != 0) 
    aout = (int) rint( (double) (aout / pagesize) ) * pagesize;
#endif

  for( i=0; i<2; i++ ) {
    w = &r->dw[i];
    if( !(w->out = ( char * ) malloc( aout*r->dw_multi ))) {
      fprintf(stderr, "bad malloc allocating buffer\n");
      set_kb(0);
      exit(1);
    }
    if( mlock( w->out, aout*r->dw_multi ))
      perror("failed to mlock");
    w->len = r->ameg*r->dw_multi;
  }
}

/*
  open output files 
  r is the config structure
  head is the linked list of directory names to be used in opening files
*/

open_files(r)
struct RADAR *r;
{
  int i, tape_fd;
  struct DISKWRITE *w;
  char *tms;
  struct MULTIFILE *fd;
  char name[80];

  fd = NULL;
  tape_fd = -1;

  if( r->istape ) {
    if((tape_fd = open( r->istape, O_WRONLY, 0666 ))<0 ) {
      fprintf( stderr, "cant open tape device %s\n", r->istape );
      set_kb(0);
      exit(1);
    }
    printf("opened tape device %s\n", r->istape );

  } else {

    sprintf(name, "%s/data%s", r->dir, r->timestr );
    fd = multi_open(name, O_WRONLY|O_CREAT, 0664, r->nfiles );
  }

  for( i=0; i< 2; i++ ) {
    w = &r->dw[i];
    w->fd = fd;
    w->tape_fd = tape_fd;
  }
}


/*
  close output files 
  r is the config structure
*/

close_files(r)
struct RADAR *r;
{
  struct DISKWRITE *w;

  w = &r->dw[0];
  multi_close(w->fd);
}


/* build time string */

get_tms(time_t time, char *string)
{
  struct tm *ans;
 
  ans = gmtime(&time);
  /* correct for peculiar tm_mon : months since January - [0, 11] */
  /* by adding one to get usual [1,12] interval */
  ans->tm_mon += 1;
  sprintf( string, "%04d%02d%02d%02d%02d%02d",
    ans->tm_year+1900, ans->tm_mon, ans->tm_mday, 
    ans->tm_hour, 
    ans->tm_min, 
    ans->tm_sec );
}
  
  
/* 
  allocate input buffers using malloc and calling mlock
  r is the config structure
*/

allocate_ringbufs(r)
struct RADAR *r;
{
  int i;

#ifdef SOLARIS
  if(!(r->rings = (unsigned short **)malloc( 
     sizeof(unsigned short *)*r->ringbufs))) {
      fprintf(stderr, "bad malloc allocating ring buffer\n");
      set_kb(0);
      exit(1);
  }

  if( mlock( r->rings, sizeof(unsigned short *)*r->ringbufs))
    perror("failed to mlock");

  for( i=0; i<r->ringbufs; i++ ) {
    if( !(r->rings[i] = ( unsigned short *)malloc(r->ameg)))
      printf("bad malloc\n");
    else
      if( mlock( r->rings[i], r->ameg ))
        perror("mlock data");
  }
#endif

  if( edt_configure_ring_buffers( r->edt, 
     r->ameg, r->ringbufs, EDT_READ, (void *)r->rings))
    edt_perror("configure edt card failure:");
}

open_log(r)
struct RADAR *r;
{
  if( r->log[0] == 0 ) {
    sprintf(r->log, "%s/radar.log", r->dir );
  } else {
    sprintf(r->log, "%s/%s", r->dir, r->log );
  }

  if( (r->logfd = fopen( r->log, "a+") )== NULL ) {
    fprintf( stderr, "Failed to open log file %s\n", r->log );
      set_kb(0);
    exit(1);
  }
  fprintf(r->logfd, "%s\n",rcsid);
  fprintf(r->logfd, "Input buffer size %d bytes\n", r->ameg );
  fprintf(r->logfd, "Input buffers, %d\n", r->ringbufs );
  fprintf(r->logfd, "Data taking duration %d seconds\n", r->secs );
  fprintf(r->logfd, "Write buffers, %d\n", r->ringbufs );
  fprintf(r->logfd, "Data taking mode, %d\n", r->mode );
  fflush(r->logfd);
}


/* set keyboard:
   if you call it with a 1: set_kb(1);
   that is the same as -> stty cbreak,stty -echo.
   if you call it with a 0: set_kb(0);
   that is the same as -> stty -cbreak,stty echo.
 
   it operates on the standard input (0) which is
   assumed to be associated with a terminal.
*/

void do_ctlc()
{
  ctlc_flag = 1;
  set_kb(0);
}


set_kb(makeraw)
int makeraw;
{
  struct termios t;
  struct termios tstate;
  static struct termios orig;
  static int first = 1;
  
  if( first ) {
    tcgetattr( 0, &orig );
    first = 0;
  }
  memcpy( &tstate, &orig, sizeof(struct termios) );

  if(makeraw) {
    tstate.c_lflag &= ~ICANON;
    tstate.c_lflag &= ~ECHO;
    tstate.c_cc[VMIN] = 1;
    tstate.c_cc[VTIME] = 0;
    tcsetattr( 0, TCSAFLUSH, &tstate );
    sigset( SIGINT, do_ctlc );

  } else {
    tcsetattr( 0, TCSAFLUSH, &tstate );
    sigset( SIGINT, SIG_DFL );
  }
}

/*
   Timing notes.

   Its important that we know the precise time the data was taken.
   A one second timing pulse is provided to trigger the hardware to start
   dumping data. From the OS we need to arm this trigger.  The start time 
   is encoded in the filenames. But it takes some time to open files, malloc 
   all the buffers.  

   Heres how we handle it.

   - We presume that the one second tick of the system clock is within +-0.5
     seconds of the timing pulse. This can be set and tested elsewhere. 
   - A start time in seconds is computed by adding an offset of AFEWSECS
     to the current one second time from time(NULL); 
   - Then the files are opened, memory is allocated and log file prepared.
   - wait till 0.5 seconds before the requested time using nanosleep
   - Then enable data taking.
   - Then arm the trigger.

*/

/* wait till 0.501 sec before expected pulse */

wait_till_start( ttt )
time_t ttt;
{
  double delay;
  struct timeval cur;
  struct timespec nano;

  gettimeofday( &cur, NULL );
  delay =  (double)ttt - cur.tv_sec  - cur.tv_usec*1.0e-6 + 0.499;
  nano.tv_sec = (int)delay;
  nano.tv_nsec = (int)((delay - nano.tv_sec)*1.0e9);
  
/*  fprintf(stderr,"nano %d %d\n",nano.tv_sec,nano.tv_nsec); */
  nanosleep( &nano, NULL ); 

}

pusage()
{
  fprintf( stderr, "Usage: pfs_radar -m mode -dir d [-secs sec] [-step sec] [-cycles c] [-start yyyy,mm,dd,hh,mm,ss]\n");
  fprintf( stderr, "                                             (defaults)\n");
  fprintf( stderr, "  -m mode\n\t 0: 2c1b (N/A)\n\t 1: 2c2b\n\t 2: 2c4b\n\t 3: 2c8b\n\t 4: 4c1b (N/A)\n\t 5: 4c2b\n\t 6: 4c4b\n\t 7: 4c8b (N/A)\n\n");
  fprintf( stderr, "  -dir d      directory to use\n");
  fprintf( stderr, "  -tape t     tape device to use\n");
  fprintf( stderr, "  -secs sec   number of seconds of data to take (3600)\n");
  fprintf( stderr, "  -step sec   timestep between A/D cycles (0)\n");
  fprintf( stderr, "  -cycles c   number of repeat cycles (1)\n");
  fprintf( stderr, "  -start yyyy,mm,dd,hh,mm,ss start time\n\n");
  fprintf( stderr, "  -files f    total number of files to open (1)\n");
  fprintf( stderr, "  -rings r    number of input buffers to use (8)\n");
  fprintf( stderr, "  -bytes b    size of input ring buffer (1048576 bytes)\n");
  fprintf( stderr, "  -code len   code length\n");
  fprintf( stderr, "  -log l      log file name \n");
  set_kb(0);
  exit(1);
}

