#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <sys/stat.h>

#include "polyco.hpp"

char *polycofile;     // name of the polyco-file
char *mjd;            // MJD data as a char*
bool is_mjd = false;  // true if MJD is specified in the command line
bool is_ns = true;    // print the value of period in nanoseconds (default)

/* help */
void help (char *s) {
 printf("\nPROGRAMM %s\n"
        "To calculate the current pulsar period and phase\n"
	"[USAGE]: %s [options]\n\n"
	"[OPTIONS]:\n"
	"-p, --polyco <polyco file> - polyco file to be used, default - \"polyco.dat\" in the current directory\n"
	"-m, --mjd <MJD>            - MJD date, no default value, should be specified\n"
	"-s, --seconds              - print period in seconds, default - nanoseconds\n"
	"-h, --help                 - help\n\n", s, s);
 exit(0);
}

/* parse command line */
int parse_command_line (int argc, char *argv[]) {
  int op, tm;
  struct option long_options[] = { {"help", no_argument, 0, 'h'},
                                   {"polyco", required_argument, 0, 'p'}, 
                                   {"mjd", required_argument, 0, 'm'},
				   {"seconds", no_argument, 0, 's'},
				   {0, 0, 0, 0}
				  };
  optind = 0; optarg = 0;
  while((op = getopt_long(argc, argv, "hsp:m:", long_options, 0)) != EOF)
    switch(op){
      case 'h':
        help (argv[0]);
      break;

      case 's':
        is_ns = false; 
      break;

      case 'p':
        if (polycofile) { polycofile = 0; free (polycofile); }
	if (!(polycofile = (char *)calloc((int)strlen(optarg), sizeof(char *)))) { perror("new polycofile"); exit (1); }
	memcpy (polycofile, optarg, (int)strlen(optarg));
      break;

      case 'm':
        if (mjd) { mjd = 0; free (mjd); }
	if (!(mjd = (char *)calloc((int)strlen(optarg), sizeof(char *)))) { perror("new mjd"); exit (1); }
	memcpy (mjd, optarg, (int)strlen(optarg));
        is_mjd = true;
      break;

      case '?':
        help (argv[0]);
      break;
    }
  return(optind);
}

/* checks string to be a floating-point number */
bool is_float (char *src) {
 int len = (int)strlen (src);
 for (int i=0; i<len; ++i) {
  if (isdigit(src[i]) != 0 || src[i] == '.') ;
   else return false;
 }
 return true;
}

/*--------------------------------------------------------------------------------------
 *                               M   A   I   N
 * -------------------------------------------------------------------------------------*/
int main(int argc, char *argv[]) {

 if (argc == 1) help (argv[0]);

 polycofile  = new char[255];
 if (!polycofile) { perror("new polycofile"); exit (1); }
 memset (polycofile, 0, 255);
 sprintf (polycofile, "%s", "polyco.dat");

 mjd  = new char[255];
 if (!mjd) { perror("new MJD"); exit (1); }
 memset (mjd, 0, 255);


 // parsing the command line
 parse_command_line (argc, argv);
 if (!is_mjd) { printf("You have to specify MJD date using -m (or --mjd) option!\n"); exit (1); }

 struct stat info;
 if (stat(polycofile, &info) != 0) { printf ("Can't access to polyco file \"%s\"!\n", polycofile); exit(1); }

 if (!is_float(mjd)) {
  printf ("Specified MJD value is bad. Check it!\n");
  exit (1);
 }

 PolycoFile pol (polycofile);  // initializing PolycoFile class

 bool out_of_range = pol.check_mjd_range (mjd);
 if (!out_of_range) {
  printf ("Specified MJD is out of range of used polyco file \"%s\"!\n", polycofile);
  exit (1);
 }

 double period = pol.getPeriod (mjd, is_ns);
 double phase  = pol.getPhase (mjd);

 printf ("File: %s\n", polycofile);
 printf ("MJD: %s\n", mjd);
 if (is_ns) printf ("Period: %.6lf\n", period); else printf ("Period: %.15lf\n", period);
 printf ("Phase: %.12lf\n", phase);

 delete (mjd);
 delete (polycofile);
 return 0;
}
