#include "polyco.hpp"
#include <fstream>
#include <stdio.h>
#include <stdlib.h>

// constructor
PolycoRecord::PolycoRecord (string p, string d, string u, string m, string disp, string dop, string f, string r, string freq, string o, int s, int n, string sky, string b) {
	pulsar = p;
	date = d;
	utc = u;
	mjd = m;
	dm = disp;
	doppler = dop;
	fit = f;
	refphase = r;
	f0 = freq;
	observatory = o;
	span = s;
	ncoeff = n;
	skyfreq = sky;
	binphase = b;

	// reserve memory for the vector
	coeffs.reserve(ncoeff);
}

// get time in "hh:mm:ss.ss" format
string PolycoRecord::getTime () {
 string cur = utc.substr(0, 2) + ":" + utc.substr(2, 2) + ":" + utc.substr(4, 5);
 return cur;
}

// get hours
int PolycoRecord::getHH () {
 return ((int)strtol(utc.substr(0, 2).c_str(), 0, 10));
}

// get minutes
int PolycoRecord::getMM () {
 return ((int)strtol(utc.substr(2, 2).c_str(), 0, 10));
}

// get seconds
double PolycoRecord::getSS () {
 return (strtod(utc.substr(4, 5).c_str(), 0));
}

// get time in seconds from the beginning of the day
double PolycoRecord::getSeconds () {
 double secs = (strtol(utc.substr(0, 2).c_str(), 0, 10)) * (double)3600. + (strtol(utc.substr(2, 2).c_str(), 0, 10)) * (double)60. + strtod(utc.substr(4, 5).c_str(), 0);
 return secs;
}

// get coefficient
string PolycoRecord::getCoeff (int index) {
 if (index >= 0 && index <= ncoeff - 1) return coeffs[index];
  else return "";
}

// get recursive product of value 'iter' times
// accepts and returns mpf_t data
void PolycoRecord::getProduct (mpf_t value, int iter) {
 mpf_t orig;
 mpf_init2 (orig, 256);
 mpf_set (orig, value);
 for (int i=1; i<iter; ++i) mpf_mul (value, value, orig);
 mpf_clear (orig);
}

// calculate the period for the particular MJD
// using the coefficients from the current record
// if ns = false period is in seconds, otherwise - in nanoseconds
double PolycoRecord::getPeriod (string m, bool ns) {
 mpf_t f_t, dt, temp, temp2;
 mpf_init2 (f_t, 256);   mpf_set_str (f_t, f0.c_str(), 10);
 mpf_init2 (dt, 256);
 mpf_init2 (temp, 256);
 mpf_init2 (temp2, 256);

 if (ncoeff >= 2) {
   mpf_set_str (dt, m.c_str(), 10);
   mpf_set_str (temp, mjd.c_str(), 10);
   mpf_sub (dt, dt, temp);
   mpf_set_ui (temp, 1440);
   mpf_mul (dt, dt, temp);
   mpf_set_str (temp, coeffs[1].c_str(), 10);
   mpf_set_ui (temp2, 60);
   mpf_div (temp, temp, temp2);
   mpf_add (f_t, f_t, temp);
   if (ncoeff > 2) 
    for (int h=2; h<ncoeff; ++h) { 
     mpf_set_str (temp, coeffs[h].c_str(), 10);
     mpf_set_ui (temp2, h);
     mpf_mul (temp, temp, temp2);
     mpf_set_ui (temp2, 60);
     mpf_div (temp, temp, temp2);
     mpf_set (temp2, dt);
     getProduct (temp2, h-1);
     mpf_mul (temp, temp, temp2); 
     mpf_add (f_t, f_t, temp); 
    } // for
  }
 mpf_set_ui (temp2, 1);
 mpf_div (f_t, temp2, f_t);
 if (ns) { mpf_set_str (temp2, "1000000000", 10); mpf_mul (f_t, f_t, temp2); }
 double p = mpf_get_d (f_t);
 mpf_clear (f_t);
 mpf_clear (dt);
 mpf_clear (temp);
 mpf_clear (temp2);
 return p;
}

// calculate the phase for the particular MJD
double PolycoRecord::getPhase (string m) {
 mpf_t ph_t, dt, temp, temp2;
 mpf_init2 (ph_t, 256);   mpf_set_str (ph_t, refphase.c_str(), 10);
 mpf_init2 (dt, 256);     mpf_set_str (dt, m.c_str(), 10); 
// gmp_printf ("m: %.17Ff\n", dt);
 mpf_init2 (temp2, 256);  mpf_set_str (temp2, mjd.c_str(), 10);
// gmp_printf ("mjd: %.17Ff\n", temp2);
 mpf_sub (dt, dt, temp2);
 mpf_set_ui (temp2, 1440);
 mpf_mul (dt, dt, temp2);
 mpf_init2 (temp, 256);   mpf_set_str (temp, f0.c_str(), 10);
 mpf_mul (temp, temp, dt); mpf_mul_ui (temp, temp, 60);
 mpf_add (ph_t, ph_t, temp);
// gmp_printf ("dt: %.17Ff\n", dt);
 
 if (ncoeff > 0) { 
  mpf_set_str (temp, coeffs[0].c_str(), 10);
  mpf_add (ph_t, ph_t, temp);
  if (ncoeff > 1) for (int h=1; h<ncoeff; ++h) { 
	            mpf_set_str (temp, coeffs[h].c_str(), 10); 
		    mpf_set (temp2, dt);
		    getProduct (temp2, h);
		    mpf_mul (temp, temp, temp2); 
		    mpf_add (ph_t, ph_t, temp); 
                  } // for
 }
 mpf_floor (temp, ph_t);
 mpf_sub (ph_t, ph_t, temp);
// gmp_printf ("phase: %.15Ff   %.15lf\n", ph_t, mpf_get_d(ph_t));
 double ph = mpf_get_d (ph_t);
 mpf_clear (ph_t);
 mpf_clear (dt);
 mpf_clear (temp);
 mpf_clear (temp2);
 return ph;
}

// setting the coefficient of index 'index'
void PolycoRecord::setCoeff (string value, int index) {
 if (index >=0 && index <= ncoeff - 1) coeffs[index] = value;
}

// setting the next coefficient in the vector recs
void PolycoRecord::setCoeff (string value) {
 coeffs.push_back(value);
}




// constructor
PolycoFile::PolycoFile (char* fn) {
 filename = fn;

 string p, d, u, o, cur;
 string m, disp, dop, f, r, freq, sky, b;
 int s, n;
 
 p.reserve(11);
 d.reserve(9);
 u.reserve(12);
 o.reserve(5);
 cur.reserve(100);

 ifstream in(fn);

 while (in) {
  if (getline (in, cur) == NULL) break;	
  p = cur.substr(0, 11);
  d = cur.substr(11, 9);
  u = cur.substr(20, 12);
  m = cur.substr(32, 20);
  disp = cur.substr(52, 21);
  dop = cur.substr(74, 6);
  f = cur.substr(80, 7);
  getline (in, cur);
  r = cur.substr(0, 21);
  freq = cur.substr(21, 18);
  o = cur.substr(39, 5);
  s = (int)strtol(cur.substr(44, 6).c_str(), 0, 10);
  n = (int)strtol(cur.substr(50, 5).c_str(), 0, 10);
  sky = cur.substr(55, 21).c_str();
  b = cur.substr(76, 5);

  recs.push_back(PolycoRecord (p, d, u, m, disp, dop, f, r, freq, o, s, n, sky, b));

  for (int l=0; l<(n%3 == 0 ? n/3 : n/3 + 1); ++l) {
   getline (in, cur);
   for (int m=0; m<(l == (n%3 == 0 ? n/3 : n/3 + 1) - 1 ? (n%3 == 0 ? 3 : n%3) : 3); ++m) {
    recs.back().setCoeff(cur.substr(1 + m*25, 25)); 
   } // fields in coefficient's line
  } // coefficient's lines
 }
 in.close();
}

// return pulsar name
string PolycoFile::getPulsar () {
 if (!recs.empty()) return recs.front().getPulsar();
  else return "";
}

// get pulsar period for the particular MJD
// if ns = false period is in seconds, otherwise - in nanoseconds
double PolycoFile::getPeriod (string mjd, bool ns) {
  list<PolycoRecord>::iterator it = recs.begin();
  double mjd_s = strtod(mjd.c_str(), 0), m;
  for (; it!=recs.end(); ++it) {
   m = strtod(it->getMJD().c_str(), 0);
   if (m - mjd_s < 0.) continue;
   if (m - mjd_s == 0.) return it->getPeriod (mjd, ns);
    else { --it; return it->getPeriod (mjd, ns); }
  } // it
  if (it == recs.end()) { --it; return it->getPeriod (mjd, ns); }
}

// calculate the phase for the particular MJD
double PolycoFile::getPhase (string mjd) {
  list<PolycoRecord>::iterator it = recs.begin();
  double mjd_s = strtod(mjd.c_str(), 0), m;
  for (; it!=recs.end(); ++it) {
   m = strtod(it->getMJD().c_str(), 0);
   if (m - mjd_s < 0.) continue;
   if (m - mjd_s == 0.) return it->getPhase (mjd);
    else { --it; return it->getPhase (mjd); }
  } // it
  if (it == recs.end()) { --it; return it->getPhase (mjd); }
}
	
// check the periods for particular MJD in "polyco.dat"
// using the coefficients from the previous record
void PolycoFile::checkModel () {
  printf ("pulsar: %s     date: %s\n", getPulsar().c_str(), recs.front().getDate().c_str());
  printf ("ncoeff = %d\n", recs.front().getNcoeff());

  list<PolycoRecord>::iterator it = recs.begin(), jt = recs.begin();
  ++it;
  printf ("%.12lf  %lf\n", strtod(jt->getMJD().c_str(), 0), jt->getPeriod(jt->getMJD(), true));
  double m, p, plast;
  string m_s;

  for (; it!=recs.end(); ++it, ++jt) {
   m_s = it->getMJD();
   m = strtod (m_s.c_str(), 0);
   p = it->getPeriod(m_s, true);
   plast = jt->getPeriod(m_s, true);
   printf ("%.12lf  %lf  %lf  %lf\n", m, p, plast, p - plast);
  }
}

// check if the specified MJD falls into the range in polyco-file
bool PolycoFile::check_mjd_range (string mjd) {
  list<PolycoRecord>::iterator it = recs.begin();
  list<PolycoRecord>::iterator jt = recs.end();
  --jt;
  double mjd_s = strtod(mjd.c_str(), 0);
  double m_beg = strtod(it->getMJD().c_str(), 0);
  double m_end = strtod(jt->getMJD().c_str(), 0);
  
  if (mjd_s >= m_beg && mjd_s <= m_end) return true;
   else return false;
}

