#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>
#include <math.h>

#include <inttypes.h>
#include <fftw3.h>

#include <sys/resource.h>

// dispersion constant
#define ALPHA1 2.607131e16
#define ALPHA2 (ALPHA1 / (2 * M_PI))

// type of input data
#define TYPE signed char

double Period; // pulsar period at pointed time in ns
double Shift; // shift from the beginning of the file to correspond for chosen phase in counts
double DM = 71.0398; // dispersion measure
// from Exp* log-file Fcentr = 2165 MHz  but constructing the scheme of the frequencies conversions
// with LO1 = 3665 MHz LO2 = 1750 MHz and baseband LO=300 MHz 
// Fcentr appeared to be 2140 MHz and so Flower = 2065 MHz (BW=150 MHz)
// So, I think that Fcentr=2165 MHz was just how we chose the sky radio-band but it's not really
// fcentr in how we digitize
double Fcentr = 2215.;  // the center frequency in MHz
double BW = 150.; // bandwidth in MHz
double Tres = 6.666667; // time resolution in ns
int ChannelNumber = 64;  // number of individual channels

// bandwidth of separate channel in MHz
double BWi;
// time resolution in separate channel in ns
double Tresi;
// center frequency of separate channel in MHz
double Fcentri;
double Frefi; // the reference frequency (to which we will dedisperse in separate channels)
              // we will choose it as lower edge for every channel

// band seems to be direct, rather than "inverse" (for 1937 data taken at 2005, Jan 6  with VA)
// lower sideband had to be taken, rather than "Upper" as is written in Exp* log-file
bool is_band_direct = true;


// cut or not data with period or just write whole data with spoiled points made zero
bool NoFold = false;

// the size of input data (analytic signal)
off_t filesize = 2097152;
// forward FFT size
off_t fft_fw_size = filesize / 2;
// the size of one channel 
off_t channelsize;

// size of the detected signal (in separate channel) without spoiled counts
off_t truesize;
// the maximum number of spoiled counts (in the lowest channel)
off_t spoiled_counts;
// pulsar period in counts 
off_t cPeriod;
// number of pulsar periods in detected signal
int Nperiods;
// shift in counts
off_t cShift;

// output path
char *Path;
// name of the file with band correction function
char *BandFile;

// Full Dispersion --- flag = 0
// Spolied Points at the beginning  --- flag = 1
// Spoiled Points at the end  --- flag = -1
unsigned long showdisp (signed char flag, double Fcentri, double Frefi) {

double tbad, f, f1;

 switch (flag) {
  case  0:  f  = Fcentri - BWi/2.;
            f1 = Fcentri + BWi/2.;
	    break;
  case  1:  f = Frefi;
            f1 = Fcentri + BWi/2.;
	    break;
  case -1:  f = Fcentri - BWi/2.;
            f1 = Frefi;
	    break;
  default: break;
 }

 tbad = DM * ALPHA2 * ( 1./(f * f) - 1./(f1 * f1) );
 tbad = (tbad * 0.001)/Tresi;

 return (unsigned long)ceil(fabs(tbad));
}

/******************** Calculate Phase Correction Value **********************/

double Calc (double f) {

 double z, index, freq;
 freq = Frefi;
 index = 1.e6;
 z = ( ( ALPHA1 * DM ) / f / ( freq * freq * index ) ) * ( f - freq ) * ( f - freq );

 return z;
}

/************** Compute Matrix ***************/

void ComputeMatrix (float *mat, off_t sz, double Fcentri) {

 double z = 0.;
 double T, freq, f0;
 
 float sign = (is_band_direct ? -1. : 1.);
  
 T = ( Tresi * ((double)sz) ) / ((double)1.e3);
 f0 = (is_band_direct ? Fcentri - BWi/2. : Fcentri + BWi/2.);
 for (off_t i=0, j=0; i < sz; i+=2, j++) {
   freq = f0 + (-1. * sign)*(((double)j)/T);
   z = Calc(freq);
   mat[i]   = (float)cos(z);
   mat[i+1] = (sign)*((float)sin(z));
 }
}

// get mean level
float getLevel (fftwf_complex *x, off_t sz) {
  double s = 0.;
  for (off_t i=0; i<sz; ++i) s += (x[i][0] + x[i][1]);
  s /= (2*sz);
  return (float)s;
}

// remove the mean level
void removeLevel (fftwf_complex *x, off_t sz, float level) {
 for (off_t i=0; i<sz; ++i) { x[i][0] -= level; x[i][1] -= level; }
}

// split the path and filename
// return filename
char *baseName (char * str) {
 char *s = str;
 int len = (int)strlen(s);
 s += len;
 while (*(--s) != '/' && len--) ;
 return (++s);
}

// split the path and filename
// return path
char *getPath (char * dest, char * str) {
 strcpy (dest, str);
 char *s = dest;
 int len = (int)strlen(s);
 s += len;
 while (*(--s) != '/' && len--) ;
 if (len != -1) { *s = '\0'; return dest; }
  else return ".";
}

/*===================Parsing the command line===========================*/

int ParseCmdLine (int argc, char *argv[]) {
  int op;
  struct option long_options[] = { {"direct", no_argument, 0, 10},
                                   {"inverse", no_argument, 0, 11},
				   {"dm", required_argument, 0, 12},
				   {"fcentr", required_argument, 0, 13},
				   {"bw", required_argument, 0, 14},
				   {"numch", required_argument, 0, 15},
				   {"nofold", no_argument, 0, 16},
				   {0, 0, 0, 0}
                                  };

  while((op = getopt_long(argc, argv, "p:t:s:d:b:", long_options, 0)) != EOF)
    switch(op){
      case 's':
        Shift = fabs(atof(optarg));
      break;
            
      case 'p':
        Period = fabs(atof(optarg));
      break;
      
      case 't':
        Tres = fabs(atof(optarg));
      break;
      
      case 'd':
	  if (Path) { Path = 0; delete (Path); }
	  Path = new char[255];
	  if (!Path) { perror("new"); exit(1); }
	  memcpy(Path, optarg, (int)strlen(optarg));
      break;
      
      case 'b':
	  if (BandFile) { BandFile = 0; delete (BandFile); }
	  BandFile = new char[255];
	  if (!BandFile) { perror("new"); exit(1); }
	  memcpy(BandFile, optarg, (int)strlen(optarg));
      break;
      
      case 10:
       is_band_direct = true;
      break;
      
      case 11:
       is_band_direct = false;
      break;
      
      case 12:
        DM = fabs(atof(optarg));
      break;
      
      case 13:
        Fcentr = fabs(atof(optarg));
      break;

      case 14:
        BW = fabs(atof(optarg));
      break;
      
      case 15:
        ChannelNumber = atoi(optarg);
      break;

      case 16:
       NoFold = true;
      break;
      
      case '?':
        printf ("Wrong option!\n");
	exit (1);
      break;
    }
  return(optind);
}

/*==================  m a i n  ========================*/

int main (int argc, char *argv[]) {

 if (argc == 1) { 
  printf ("Usage: %s [options] <one file from lin.[X|Y].*>\n"
	  "  [options]: -p <period>\n"
	  "             -t <timeres>\n"
	  "             -s <shift>\n"
	  "             -d <path>\n"
	  "             -b <band_correction_function_file>\n"
	  "             --dm <DM>\n"
	  "             --fcentr <Fcentr>\n"
	  "             --bw <BW>\n"
	  "             --direct (for direct band, default)\n"
	  "             --inverse (for inverse band)\n"
	  "             --nofold (don't cut with period)\n"
	  "             --numch <Nchannels>\n", baseName(argv[0])); 
  exit (0); 
 }

 /* Making this process to be able to create files larger than 2GB */
 struct rlimit r;
 r.rlim_cur = r.rlim_max = RLIM_INFINITY;
 setrlimit(RLIMIT_FSIZE,&r);


 unsigned long bytesread; // number of actually read bytes
 TYPE* data; // the array of the whole input data
 data = new TYPE[filesize];
 if (!data) { perror("new"); exit(1); }

 Path = new char[255];
 if (!Path) { perror("new"); exit(1); }
 memset (Path, 0, 255);
 sprintf (Path, "%s", "\0");

 BandFile = new char[255];
 if (!BandFile) { perror("new"); exit(1); }
 memset (BandFile, 0, 255);
 sprintf (BandFile, "%s", "\0");

 // number of command line argument with name of the file
 int next = ParseCmdLine(argc, argv);
 if (argc < next + 1) { printf("Not point the file!\n"); exit(1); }
 int in;   // input file descriptor

 if (!strcmp(Path, "\0")) { memset (Path, 0, 255); sprintf (Path, "%s", getPath(Path, argv[next])); }

 // if BandFile is pointed out
 float *band;
 if (strcmp(BandFile, "\0")) {
  band = new float[fft_fw_size]; 
  if (!band) { perror("new"); exit(1); }
  memset (band, 0, fft_fw_size * sizeof(float));
 }

 // bandwidth of separate channel in MHz
 BWi = BW / ChannelNumber;
 // time resolution in separate channel in ns
 Tresi = Tres * ChannelNumber;
 // the size of one channel 
 channelsize = fft_fw_size / ChannelNumber;

 memset (data, 0, filesize);
 in = open (argv[next], O_RDONLY | O_LARGEFILE);
 if (in == -1) { perror("open"); exit (1); }
 lseek (in, 0, SEEK_SET);
 bytesread = read (in, data, filesize);
 if (bytesread != filesize) { perror("read"); exit(1); }
 close (in);

 fftwf_complex *x, *y; // array of input data to perform the complex FFT and array with FFT results
 x = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * fft_fw_size);
 y = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * fft_fw_size);
 if (!x || !y) { perror("fftwf_malloc"); exit(1); }
 for (off_t i=0, j=0; i<fft_fw_size; ++i, j+=2) { x[i][0] = (float)data[j]; x[i][1] = (float)data[j+1]; }
 // get & remove level
 removeLevel (x, fft_fw_size, getLevel(x, fft_fw_size));
 fftwf_plan pfor; // FFTW plan to perform forward FFT
 pfor = fftwf_plan_dft_1d (fft_fw_size, x, y, FFTW_FORWARD, FFTW_ESTIMATE);
 // calculate forward FFT
 fftwf_execute (pfor);
 
 // rotate spectrum (put "-" harmonics in front)
 for (off_t i=fft_fw_size/2 + 1, j=0; i<fft_fw_size; ++i, ++j) { x[j][0] = y[i][0]; x[j][1] = y[i][1]; }
 for (off_t i=0, j=fft_fw_size/2 - 1; i<fft_fw_size/2 + 1; ++i, ++j) { x[j][0] = y[i][0]; x[j][1] = y[i][1]; }

/*
// this changes are for band with harmonics rotation
  x[2097151][0] = (x[2097150][0] + x[2097152][0])/2; x[2097151][1] = (x[2097150][1] + x[2097152][1])/2;
  x[4194303][0] = x[4194302][0]; x[4194303][1] = x[4194302][1];
  x[2095280][0] = (x[2095279][0] + x[2095281][0]) / 2; x[2095280][1] = (x[2095279][1] + x[2095281][1]) / 2;
  x[2099022][0] = (x[2099021][0] + x[2099023][0]) / 2; x[2099022][1] = (x[2099021][1] + x[2099023][1]) / 2;
  x[2096215][0] = x[2096214][0]; x[2096215][1] = x[2096214][1];
  x[2096216][0] = x[2096217][0]; x[2096216][1] = x[2096217][1];
  x[2098086][0] = x[2098085][0]; x[2098086][1] = x[2098085][1];
  x[2098087][0] = x[2098088][0]; x[2098087][1] = x[2098088][1];
  x[1048575][0] = (x[1048574][0] + x[1048576][0]) / 2; x[1048575][1] = (x[1048574][1] + x[1048576][1]) / 2;
  x[524287][0] = (x[524286][0] + x[524288][0]) / 2; x[524287][1] = (x[524286][1] + x[524288][1]) / 2;
  x[3145727][0] = (x[3145726][0] + x[3145728][0]) / 2; x[3145727][1] = (x[3145726][1] + x[3145728][1]) / 2;
  x[3670015][0] = (x[3670014][0] + x[3670016][0]) / 2; x[3670015][1] = (x[3670014][1] + x[3670016][1]) / 2;
*/

 /*
 // this changes are for band without harmonics rotation
 y[0][0] = y[1][0]; y[0][1] = y[1][1];
 y[2097152][0] = (y[2097151][0] + y[2097153][0]) / 2; y[2097152][1] = (y[2097151][1] + y[2097153][1]) / 2;
 y[4193368][0] = y[4193367][0]; y[4193368][1] = y[4193367][1];
 y[4193369][0] = y[4193370][0]; y[4193369][1] = y[4193370][1];
 y[4192433][0] = (y[4192432][0] + y[4192434][0]) / 2; y[4192433][1] = (y[4192432][1] + y[4192434][1]) / 2;
 y[935][0] = y[934][0]; y[935][1] = y[934][1];
 y[936][0] = y[937][0]; y[936][1] = y[937][1];
 y[1871][0] = (y[1870][0] + y[1872][0]) / 2; y[1871][1] = (y[1870][1] + y[1872][1]) / 2;
 y[1572864][0] = (y[1572863][0] + y[1572865][0]) / 2; y[1572864][1] = (y[1572863][1] + y[1572865][1]) / 2;
 y[2621440][0] = (y[2621439][0] + y[2621441][0]) / 2; y[2621440][1] = (y[2621439][1] + y[2621441][1]) / 2;
 y[1048576][0] = (y[1048575][0] + y[1048577][0]) / 2; y[1048576][1] = (y[1048575][1] + y[1048577][1]) / 2;
 y[3145728][0] = (y[3145727][0] + y[3145729][0]) / 2; y[3145728][1] = (y[3145727][1] + y[3145729][1]) / 2;
*/

 // read the file with band correction function
 int band_descr; // band file descriptor
 band_descr = open (BandFile, O_RDONLY | O_LARGEFILE);
 if (band_descr == -1) { perror("open"); exit (1); }
 lseek (band_descr, 0, SEEK_SET);
 read (band_descr, band, fft_fw_size * sizeof (float));
 close (band_descr);
/*
 // correct the total passband for bandpass irregularities
 for (off_t i=0; i<fft_fw_size; ++i) {
  y[i][0] = (x[i][0] - x[i][1]) * band[i];
  y[i][1] = (x[i][0] + x[i][1]) * band[i];
 }
 */

 fftwf_complex *z, *v;  // array of recovered spectrum of separate channel and array with reverse FFT results
 z = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * channelsize);
 v = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * channelsize);
 fftwf_plan prev; // FFTW plan to perform reverse FFT
 prev = fftwf_plan_dft_1d (channelsize, z, v, FFTW_BACKWARD, FFTW_ESTIMATE);

 // array of dedispersion matrix coefficients
 float *mat;
 mat = new float[2*channelsize];
 if (!mat) { perror("new"); exit(1); }

 spoiled_counts = showdisp (0, Fcentr - BW/2. + BWi/2., Fcentr - BW/2.);
 truesize = channelsize - spoiled_counts;

 // array of detected signal in separate channel
 float *det;
 det = new float[NoFold ? channelsize : truesize];
 if (!det) { perror("new"); exit(1); }

 // calculating the pulsar period in counts and number of periods and cShift
 cPeriod = (off_t)rint(Period/Tresi);
 cShift = (off_t)rint(Shift);
 Nperiods = (int)( ((double)(truesize - cShift)) / ((double)cPeriod) );
 
 // array of the average profile (folded detected signal)
 float *prof;
 prof = new float[cPeriod];
 if (!prof) { perror("new"); exit(1); }

 // output filename
 char *output;
 output = new char[255];
 if (!output) { perror("new"); exit(1); }
 int out; // output file descriptor

 printf ("dm: %lf  fcentr: %lf  bw: %lf (%s) tres: %lf\n", DM, Fcentr, BW, is_band_direct ? "direct" : "inverse", Tres);
 printf ("fftsize: %ld\n", fft_fw_size);
 printf ("number of channels: %d\n", ChannelNumber);
 printf ("channel size: %ld  truesize: %ld  spoiled: %ld\n", channelsize, truesize, spoiled_counts);
 printf ("Fold with period: %s\n", NoFold ? "no": "yes");
 printf ("period: %lf [%ld]  number: %d\n", Period, cPeriod, Nperiods);
 printf ("shift: %lf [%ld]\n", Shift, cShift);


 float sign = (is_band_direct ? 1. : -1.);
 // loop on frequency channels
 for (int ch=0; ch<ChannelNumber; ++ch) {
   // create the matrix
   // recover for the dispersion writing to another array z
   // reverse FFT
   // cut with Period and Shift
   // fold the records
   // write the profile to the file

   // calculate matrix
   memset (mat, 0, sizeof(float)*2*channelsize);
   Fcentri = Fcentr + (-1.*sign)*(BW/2.) + sign*(BWi/2.) + sign * ch * BWi;
   Frefi = Fcentr + (-1.*sign)*(BW/2.) + sign * ch * BWi;  // make the lower (upper if band inverse) frequency of every channel as reference
   ComputeMatrix (mat, channelsize, Fcentri);

   memset (z, 0, sizeof(float) * 2 * channelsize);
   memset (v, 0, sizeof(float) * 2 * channelsize);

   // dedispersing the data
   for (off_t i=ch*channelsize, j=0, k=0; i<(ch+1)*channelsize; ++i, ++j, k+=2) {
    z[j][0] = (x[i][0] * mat[k]   - mat[k+1] * x[i][1]) * band[i];
    z[j][1] = (x[i][0] * mat[k+1] + mat[k]   * x[i][1]) * band[i];
   }

   // reverse FFT
   fftwf_execute (prev);
   // getting detected signal  and normalizing
   memset (det, 0, sizeof(float)*(NoFold ? channelsize : truesize));
   if (is_band_direct) {
    for (off_t i=0, j=spoiled_counts; i<truesize; ++i, ++j) {
     det[i] = (v[j][0] / channelsize) * (v[j][0] / channelsize) + (v[j][1] / channelsize) * (v[j][1] / channelsize);
    }
   } else { // if band is inverse
    for (off_t i=0, j=0; i<truesize; ++i, ++j) {
     det[i] = (v[j][0] / channelsize) * (v[j][0] / channelsize) + (v[j][1] / channelsize) * (v[j][1] / channelsize);
    }
   }
   
   if (NoFold) {
	   
    // saving into file
    memset (output, 0, 255);
    sprintf (output, "%s/%s.det.ch%03d", Path, baseName(argv[next]), is_band_direct ? ch : ChannelNumber - 1 - ch);

    out = open (output, O_TRUNC | O_CREAT | O_WRONLY | O_LARGEFILE, 0644);
    lseek (out, 0, SEEK_SET);
    write (out, det, sizeof(float)*channelsize);
    close (out);
	   
   } else { // if NoFold = false
   
   // cutting and folding with Shift and Period 
   memset (prof, 0, sizeof(float) * cPeriod);
   for (int p=0; p<Nperiods; ++p) {
    for (off_t i=0, j=cShift+p*cPeriod; i<cPeriod; ++i, ++j) prof[i] += det[j];
   }
   for (off_t i=0; i<cPeriod; ++i) prof[i] /= Nperiods;

   // saving into file
   memset (output, 0, 255);
   sprintf (output, "%s/%s.prof.ch%03d", Path, baseName(argv[next]), is_band_direct ? ch : ChannelNumber - 1 - ch);

   out = open (output, O_TRUNC | O_CREAT | O_WRONLY | O_LARGEFILE, 0644);
   lseek (out, 0, SEEK_SET);
   write (out, prof, sizeof(float)*cPeriod);
   close (out);

   } // end of else for NoFold = false
	 
 } // end of the loop on frequency channels (ch)
 
 fftwf_free (x);
 fftwf_free (y);
 fftwf_destroy_plan (pfor);
 delete (mat);
 if (strcmp(BandFile, "\0")) delete (band);
 fftwf_free (z);
 fftwf_free (v);
 fftwf_destroy_plan (prev);
 delete (det);
 delete (prof);
 delete (output);
 delete (Path);
 delete (BandFile);
 return 0;
}


/*=================================================================================
 int out;
 out = open ("test_spectrum", O_TRUNC | O_CREAT | O_RDWR | O_LARGEFILE, 0644);
 lseek (out, 0, SEEK_SET);
 float *m;
 m = new float[fft_fw_size];
 if (!m) { perror("new"); exit(1); }
 memset (m, 0, sizeof(float)*fft_fw_size);
 for (off_t u=0; u<fft_fw_size; ++u) { m[u] = y[u][0]*y[u][0] + y[u][1]*y[u][1]; }
 write (out, m, sizeof(float)*fft_fw_size);
 close (out);
 delete (m);
 exit (0);
 =================================================================================*/
