#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>
#include <math.h>

#include <inttypes.h>
#include <fftw3.h>

#include <sys/resource.h>

// dispersion constant
#define ALPHA1 2.607131e16
#define ALPHA2 (ALPHA1 / (2 * M_PI))

// type of input data
#define TYPE signed char

double Period; // pulsar period at pointed time in ns
double Shift; // shift from the beginning of the file to correspond for chosen phase in counts
double DM = 71.0398; // dispersion measure
// from Exp* log-file Fcentr = 2165 MHz  but constructing the scheme of the frequencies conversions
// with LO1 = 3665 MHz LO2 = 1750 MHz and baseband LO=300 MHz 
// Fcentr appeared to be 2140 MHz and so Flower = 2065 MHz (BW=150 MHz)
// So, I think that Fcentr=2165 MHz was just how we chose the sky radio-band but it's not really
// fcentr in how we digitize
double Fcentr = 2215.;  // the center frequency in MHz
double BW = 150.; // bandwidth in MHz
double Tres = 6.666667; // time resolution in ns
int ChannelNumber = 64;  // number of individual channels

// bandwidth of separate channel in MHz
double BWi;
// time resolution in separate channel in ns
double Tresi;
// center frequency of separate channel in MHz
double Fcentri;
double Frefi; // the reference frequency (to which we will dedisperse in separate channels)
              // we will choose it as lower edge for every channel

// band seems to be direct, rather than "inverse" (for 1937 data taken at 2005, Jan 6  with VA)
// lower sideband had to be taken, rather than "Upper" as is written in Exp* log-file
bool is_band_direct = true;


// cut or not data with period or just write whole data with spoiled points made zero
bool NoFold = false;

// the size of input data (analytic signal)
off_t filesize = 8388608;
// forward FFT size
off_t fft_fw_size = filesize / 2;
// the size of one channel 
off_t channelsize;

// size of the detected signal (in separate channel) without spoiled counts
off_t truesize;
// the maximum number of spoiled counts (in the lowest channel)
off_t spoiled_counts;
// pulsar period in counts 
off_t cPeriod;
// number of pulsar periods in detected signal
int Nperiods;
// shift in counts
off_t cShift;

// output path
char *Path;

// Full Dispersion --- flag = 0
// Spolied Points at the beginning  --- flag = 1
// Spoiled Points at the end  --- flag = -1
unsigned long showdisp (signed char flag, double Fcentri, double Frefi) {

double tbad, f, f1;

 switch (flag) {
  case  0:  f  = Fcentri - BWi/2.;
            f1 = Fcentri + BWi/2.;
	    break;
  case  1:  f = Frefi;
            f1 = Fcentri + BWi/2.;
	    break;
  case -1:  f = Fcentri - BWi/2.;
            f1 = Frefi;
	    break;
  default: break;
 }

 tbad = DM * ALPHA2 * ( 1./(f * f) - 1./(f1 * f1) );
 tbad = (tbad * 0.001)/Tresi;

 return (unsigned long)ceil(fabs(tbad));
}

/******************** Calculate Phase Correction Value **********************/

double Calc (double f) {

 double z, index, freq;
 freq = Frefi;
 index = 1.e6;
 z = ( ( ALPHA1 * DM ) / f / ( freq * freq * index ) ) * ( f - freq ) * ( f - freq );

 return z;
}

/************** Compute Matrix ***************/

void ComputeMatrix (float *mat, off_t sz, double Fcentri) {

 double z = 0.;
 double T, freq, f0;
 
 float sign = (is_band_direct ? -1. : 1.);
  
 T = ( Tresi * ((double)sz) ) / ((double)1.e3);
 f0 = (is_band_direct ? Fcentri - BWi/2. : Fcentri + BWi/2.);
 for (off_t i=0, j=0; i < sz; i+=2, j++) {
   freq = f0 + (-1. * sign)*(((double)j)/T);
   z = Calc(freq);
   mat[i]   = (float)cos(z);
   mat[i+1] = (sign)*((float)sin(z));
 }
}

// get mean level
float getLevel (fftwf_complex *x, off_t sz) {
  double s = 0.;
  for (off_t i=0; i<sz; ++i) s += (x[i][0] + x[i][1]);
  s /= (2*sz);
  return (float)s;
}

// remove the mean level
void removeLevel (fftwf_complex *x, off_t sz, float level) {
 for (off_t i=0; i<sz; ++i) { x[i][0] -= level; x[i][1] -= level; }
}

// split the path and filename
// return filename
char *baseName (char * str) {
 char *s = str;
 int len = (int)strlen(s);
 s += len;
 while (*(--s) != '/' && len--) ;
 return (++s);
}

// split the path and filename
// return path
char *getPath (char * dest, char * str) {
 strcpy (dest, str);
 char *s = dest;
 int len = (int)strlen(s);
 s += len;
 while (*(--s) != '/' && len--) ;
 if (len != -1) { *s = '\0'; return dest; }
  else return ".";
}

/*===================Parsing the command line===========================*/

int ParseCmdLine (int argc, char *argv[]) {
  int op;
  struct option long_options[] = { {"direct", no_argument, 0, 10},
                                   {"inverse", no_argument, 0, 11},
				   {"dm", required_argument, 0, 12},
				   {"fcentr", required_argument, 0, 13},
				   {"bw", required_argument, 0, 14},
				   {"numch", required_argument, 0, 15},
				   {"nofold", no_argument, 0, 16},
				   {0, 0, 0, 0}
                                  };

  while((op = getopt_long(argc, argv, "p:t:s:d:", long_options, 0)) != EOF)
    switch(op){
      case 's':
        Shift = fabs(atof(optarg));
      break;
            
      case 'p':
        Period = fabs(atof(optarg));
      break;
      
      case 't':
        Tres = fabs(atof(optarg));
      break;
      
      case 'd':
	  if (Path) { Path = 0; delete (Path); }
	  Path = new char[255];
	  if (!Path) { perror("new"); exit(1); }
	  memcpy(Path, optarg, (int)strlen(optarg));
      break;
      
      case 10:
       is_band_direct = true;
      break;
      
      case 11:
       is_band_direct = false;
      break;
      
      case 12:
        DM = fabs(atof(optarg));
      break;
      
      case 13:
        Fcentr = fabs(atof(optarg));
      break;

      case 14:
        BW = fabs(atof(optarg));
      break;
      
      case 15:
        ChannelNumber = atoi(optarg);
      break;

      case 16:
       NoFold = true;
      break;
      
      case '?':
        printf ("Wrong option!\n");
	exit (1);
      break;
    }
  return(optind);
}

/*==================  m a i n  ========================*/

int main (int argc, char *argv[]) {

 if (argc == 1) { 
  printf ("Usage: %s [options] <one file from lin.[X|Y].*>\n"
	  "  [options]: -p <period>\n"
	  "             -t <timeres>\n"
	  "             -s <shift>\n"
	  "             -d <path>\n"
	  "             --dm <DM>\n"
	  "             --fcentr <Fcentr>\n"
	  "             --bw <BW>\n"
	  "             --direct (for direct band, default)\n"
	  "             --inverse (for inverse band)\n"
	  "             --nofold (don't cut with period)\n"
	  "             --numch <Nchannels>\n", baseName(argv[0])); 
  exit (0); 
 }

 /* Making this process to be able to create files larger than 2GB */
 struct rlimit r;
 r.rlim_cur = r.rlim_max = RLIM_INFINITY;
 setrlimit(RLIMIT_FSIZE,&r);


 unsigned long bytesread; // number of actually read bytes
 TYPE* data; // the array of the whole input data
 data = new TYPE[filesize];
 if (!data) { perror("new"); exit(1); }

 Path = new char[255];
 if (!Path) { perror("new"); exit(1); }
 memset (Path, 0, 255);
 sprintf (Path, "%s", "\0");

 // number of command line argument with name of the file
 int next = ParseCmdLine(argc, argv);
 if (argc < next + 1) { printf("Not point the file!\n"); exit(1); }
 int in;   // input file descriptor

 if (!strcmp(Path, "\0")) { memset (Path, 0, 255); sprintf (Path, "%s", getPath(Path, argv[next])); }

 // bandwidth of separate channel in MHz
 BWi = BW / ChannelNumber;
 // time resolution in separate channel in ns
 Tresi = Tres * ChannelNumber;
 // the size of one channel 
 channelsize = fft_fw_size / ChannelNumber;

 memset (data, 0, filesize);
 in = open (argv[next], O_RDONLY | O_LARGEFILE);
 if (in == -1) { perror("open"); exit (1); }
 lseek (in, 0, SEEK_SET);
 bytesread = read (in, data, filesize);
 if (bytesread != filesize) { perror("read"); exit(1); }
 close (in);

 fftwf_complex *x, *y; // array of input data to perform the complex FFT and array with FFT results
 x = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * fft_fw_size);
 y = (fftwf_complex *) fftwf_malloc (sizeof(fftwf_complex) * fft_fw_size);
 if (!x || !y) { perror("fftwf_malloc"); exit(1); }
 for (off_t i=0, j=0; i<fft_fw_size; ++i, j+=2) { x[i][0] = (float)data[j]; x[i][1] = (float)data[j+1]; }
 // get & remove level
 removeLevel (x, fft_fw_size, getLevel(x, fft_fw_size));
 fftwf_plan pfor; // FFTW plan to perform forward FFT
 pfor = fftwf_plan_dft_1d (fft_fw_size, x, y, FFTW_FORWARD, FFTW_ESTIMATE);
 // calculate forward FFT
 fftwf_execute (pfor);
 // rotate spectrum (put "-" harmonics in front)
 for (off_t i=fft_fw_size/2 + 1, j=0; i<fft_fw_size; ++i, ++j) { x[j][0] = y[i][0]; x[j][1] = y[i][1]; }
 for (off_t i=0, j=fft_fw_size/2 - 1; i<fft_fw_size/2 + 1; ++i, ++j) { x[j][0] = y[i][0]; x[j][1] = y[i][1]; }

 float *sp;
 sp = new float[fft_fw_size];
 if (!sp) { perror("new"); exit(1); }
 memset (sp, 0, sizeof(float) * fft_fw_size);
 for (off_t i=0; i<fft_fw_size; ++i) sp[i] = x[i][0] * x[i][0] + x[i][1] * x[i][1];
// for (off_t i=0; i<fft_fw_size; ++i) sp[i] = y[i][0] * y[i][0] + y[i][1] * y[i][1];

 // output filename
 char *output;
 output = new char[255];
 if (!output) { perror("new"); exit(1); }
 int out; // output file descriptor

    // saving into file
    memset (output, 0, 255);
    sprintf (output, "%s/%s.sp2", Path, baseName(argv[next]));

    out = open (output, O_TRUNC | O_CREAT | O_WRONLY | O_LARGEFILE, 0644);
    lseek (out, 0, SEEK_SET);
    write (out, sp, sizeof(float)*fft_fw_size);
    close (out);
	   
 fftwf_free (x);
 fftwf_free (y);
 fftwf_destroy_plan (pfor);
 delete (sp);
 delete (output);
 delete (Path);
 return 0;
}


/*=================================================================================
 int out;
 out = open ("test_spectrum", O_TRUNC | O_CREAT | O_RDWR | O_LARGEFILE, 0644);
 lseek (out, 0, SEEK_SET);
 float *m;
 m = new float[fft_fw_size];
 if (!m) { perror("new"); exit(1); }
 memset (m, 0, sizeof(float)*fft_fw_size);
 for (off_t u=0; u<fft_fw_size; ++u) { m[u] = y[u][0]*y[u][0] + y[u][1]*y[u][1]; }
 write (out, m, sizeof(float)*fft_fw_size);
 close (out);
 delete (m);
 exit (0);
 =================================================================================*/
