#!/usr/local/bin/perl
#
# read info from OBS_LOGS.txt file and make a DEC - RA plot
# Note: the last file in every session is not counted, because it could be smaller than 80s
# Note: if "Files" = 999 then this record is not counted
#
use PGPLOT;
use Getopt::Long;
use Math::Trig;

$pgplot_dev = "/xwin";
$filedur = 80.02732032;       # duration of every file (in sec)
$beam = 35.;                  # beam-width at 350 MHz (in arcmin)
$gbt_longitude = -79.8398333; # GBT longitude (in degrees);  GBT longitude = 79 deg 50' 23.40" W
$gbt_latitude = 38.43311944;  # GBT latitude (in degrees);   GBT latitude =  38 deg 25' 59.23" N

$pi = 3.14159265359;
$rad = $pi/180.;

$sphere = 41252.96125; # square degrees in sphere
$covered_area;
$covered_area_percent;
$survey_area;          # total possible survey area
$survey_area_percent;  # total possible survey area (in percent of the total sky)
$overlapped_area;      
$overlapped_area_percent;
$net_covered_area;
$net_covered_area_percent;

$is_coords = 0;

# for conversion galactic coordinates to equatorial
$dec_NGP = 27.1282511;  # declination of the North Galactic Pole (in degrees), J2000
$ra_NGP = 192.8594813;  # right assension of the North Galactic Pole (in degrees), J2000
$lasc = 33.;            # longitude of the ascending node of the Galactic Plane

# Parse command line
GetOptions ( "h" => \$is_help,      # to print the help
             "help" => \$is_help,   # to print the help
             "p=s" => \$pgplot_dev, # pgplot device
	     "who" => \$is_who,     # mark obs session of different observer with different color
             "az" => \$is_az,       # mark sessions of different azimuth with different color
             "overlap" => \$is_overlap, # mark overlap regions
             "test" => \$is_test,   # for testing purposes (checking new elevations)
             "update" => \$is_update ); # to update OBS_LOGS.txt file

if ($is_help) {  &help($0); exit 0; }

if ($is_test) {
 $file = "/home/users/vkondrat/obs/drift/OBS_LOGS_PLAN.txt";
} else {
 $file = "/home/users/vkondrat/obs/drift/OBS_LOGS.txt";
}
# first upfate OBS_LOGS.txt file
if ($is_update) {
 system ("scp spigot2:/users/sransom/driftscan/OBS_LOGS.txt $file");
}

open (OBS, "<$file") || die "can't open file \"$file\": $!\n";
while (defined ($line = <OBS>)) {
 chomp $line;
 if ($line =~ /^#/) { next; }
 @temp = split /\s+/, $line;
 push @{ $list{"mjd"} }, $temp[0];
 push @{ $list{"who"} }, $temp[1];
 push @{ $list{"files"} }, $temp[2];
 push @{ $list{"az"} }, $temp[3];
 push @{ $list{"el"} }, $temp[4];
 push @{ $list{"bits"} }, $temp[5];
 push @{ $list{"basename"} }, $temp[6];
 push @{ $list{"disks"} }, $temp[7];
}
close (OBS);


# number of lines on OBS_LOGS.txt file
$numlines = $#{@{$list{"mjd"}}};

#
# graphics part
#
($dec_mino, $dec_maxo) = &dec_range (); # get the declination range taking into account the used azimuths and elevations
&total_area ($dec_mino, $dec_maxo); # calculate the total sky square to cover
$dec_min = &rounddown ($dec_mino);   # round the value down to the closest integer divisible by 5
$dec_max = &roundup ($dec_maxo);     # round the value up to the closest integer divisible by 5 
$dec_min *= 3600.; # converting to arc seconds
$dec_max *= 3600.;
$ra_min = 0;
$ra_max = 3600*24;

$ydataoffset = ($dec_max - $dec_min) * 0.015;
$xdataoffset = ($ra_max - $ra_min) * 0.015;
$xlabel = "Right Assension";
$ylabel = "Declination";

$char_size = 1.;
$psfile = "decra" . ".ps/cps"; # possible ps-file for writing

$id0 = pgopen ($pgplot_dev);
pgsch ($char_size);

&mainwindow ($pgplot_dev);

if (($pgplot_dev =~ /ps/) || ($pgplot_dev =~ /PS/)) {
 exit;
}

# waiting for quit
&winhelp();
while (1) {
 # look for cursor inpra
 pgcurs ($x_cursor, $y_cursor, $ch_cursor);

 # to quit
 if ($ch_cursor =~ /[xXqQ]/) {
   last;
 }
 # to help
 if ($ch_cursor =~ /[hH]/) { &winhelp(); next; }

 # get value of closest candidate
 if ($ch_cursor =~ /[aA]/) {
  &print_coords ($x_cursor, $y_cursor);
  next;
 }

 # save to postscript-file
 if ($ch_cursor =~ /[pP]/) {
  $id1 = pgopen ($psfile);
  pgslct ($id1);
  pgsch ($char_size);
  &mainwindow($psfile);
  pgclos ();
  pgslct ($id0);
  next;
 }

} # while


#######################################################################################
#
#
#                            R O U T I N E S
#
#

# help
sub help {
 local ($prg) = @_;
 $prg = `basename $prg`; chomp $prg;
 print "Usage: $prg [options]\n";
 print "       -p <dev>  - pgplot device\n";
 print "       -who      - mark observers\n";
 print "       -az       - mark azimuths\n";
 print "       -overlap  - mark overlap regions\n";
 print "       -test     - mark only test records in \"OBS_LOGS.txt\" file\n";
 print "       -update   - to download file \"OBS_LOGS.txt\" from spigot2\n";
 print "       -h, -help - print this help\n"; 
}

# get major tic
sub gettics {
 local ($min, $max, @mult) = @_;
 local $beg = int(($max - $min)/5.);
 local $end = int(($max - $min)/4.);
 for ($j=0; $j<=$#mult+1; $j++) {
  for ($i=$beg; $i<=$end; $i++) {
   if ( (int($i/$mult[$j]))*$mult[$j] == $i ) { return $i; }
  }
 }
 return $end;
}

# window help
sub winhelp {
 print "\n";
 print "Help:\n";
 print "x, X, q, Q, Right Button  ->  quit\n";
 print "a, A, Left Button  ->  print cursor coordinates\n";
 print "p, P  ->  save plot to postscript file\n";
 print "h, H  ->  this help\n";
 print "\n";
}

# print coordinates of the cursor
sub print_coords {
 local ($xc, $yc) = @_;
 printf ("%s: %s   %s: %s\n", "DEC", &get_decra($yc), "RA", &get_decra($xc));
}

# plot points in main window
sub mainwindow {

local ($dev) = @_;

pgsvp (0.07, 0.99, 0.17, 0.95);
pgswin ($ra_min - $xdataoffset, $ra_max + $xdataoffset, $dec_min - $ydataoffset, $dec_max + $ydataoffset);

pgslw (1);
pgsfs (1);

# set gray scale map
&grayscale ($dev);

pgsci (18);
pgbox ("G", 1*3600, 0, "G", 5*60*60, 0);

pgsci (1);
pgtbox ("BCNTSZYHOI", 0.0, 4, "BCSNTZYDOI", 0.0, 5);
pgmtxt ("B", 2.7, 0.5, 0.5, $xlabel);
pgmtxt ("L", 2.7, 0.5, 0.5, $ylabel);
pgmtxt ("T", 1., 0.5, 0.5, "350-MHz Pulsar Drift-Scan Survey");
pgiden ();
# draw declination range
pgsls (4);
pgmove ($ra_min - $xdataoffset, $dec_mino * 3600.);
pgdraw ($ra_max + $xdataoffset, $dec_mino * 3600.);
pgmove ($ra_min - $xdataoffset, $dec_maxo * 3600.);
pgdraw ($ra_max + $xdataoffset, $dec_maxo * 3600.);
pgsls (1);
#pgsfs (2);
pgsfs (1);
pgsch (0.5*$char_size);
$tt = sprintf ("%.3f", $dec_mino);
pgmtxt ("B", -2.0, 0.01, 0.0, $tt);
$tt = sprintf ("%.3f", $dec_maxo);
pgmtxt ("T", -1.5, 0.01, 0.0, $tt);
pgsch ($char_size);

pgsci (25);
$overlapped_area = 0;
$covered_area = 0;
# draw a rectangular for every record
for ($i=0; $i<=$numlines; $i++) {
 $t1 = ${@{$list{"mjd"}}}[$i];
 $lst1 = &siderial ($t1);
 $lst1 *= 3600.;

 $t2 = ${@{$list{"files"}}}[$i];
 if ($t2 == 999) { next; }
 $t2 = $t1 + (($t2 - 1) * $filedur) / 86400.;
 $lst2 = &siderial ($t2);
 $lst2 *= 3600.;

 $el = ${@{$list{"el"}}}[$i];
 $az = ${@{$list{"az"}}}[$i];

 &get_delta_extremum ($el, $az);
 $dec1 = $delta_min * 3600.;
 $dec2 = $delta_max * 3600.;
 &get_ha_extremum ($el, $az);

 $ra1 = $lst1 - $ha_max * 3600.;
 $ra2 = $lst2 - $ha_min * 3600.;

 if ($ra1 < 0) { $ra1 += 86400; }
 if ($ra2 < 0) { $ra2 += 86400; }
 if ($ra1 >= 86400) { $ra1 -= 86400; }
 if ($ra2 < $ra1) { $ra2 += 86400; }

 # making the list of unique overlap regions
 &overlap ($i, $ra1, $ra2, $dec1, $dec2);
 $overlapped_area += &overlap_area ($i, $ra1, $ra2, $dec1, $dec2);
 $covered_area += &covered ($ra1, $ra2, $dec1, $dec2);

 if ($is_coords == 0) {
  # add the range of DEC into the list
  push @{ $coord{"declo"} }, $dec1;
  push @{ $coord{"dechi"} }, $dec2;
  push @{ $coord{"ralo"} }, $ra1;
  push @{ $coord{"rahi"} }, $ra2;
 }

 if ($is_who) {
  pgsci (25);
  $who = ${@{$list{"who"}}}[$i];
  if ($who eq "SMR") { pgsci (3); }
  if ($who eq "VIK") { pgsci (4); }
  if ($who eq "JVL") { pgsci (6); }
  if ($who eq "MAM") { pgsci (7); }
  if ($who eq "DRL") { pgsci (8); }
  if ($who eq "SDSS") { pgsci (9); }
  if ($who eq "JWTH") { pgsci (10); }
  if ($who eq "JRB") { pgsci (11); }
  if ($who eq "CRE") { pgsci (12); }
  if ($who eq "DAL") { pgsci (13); }
  if ($who eq "MSR") { pgsci (14); }
  if ($who eq "TEST") { pgsci (1); }
 }

 if ($is_test) {
  pgsci (25);
  $who = ${@{$list{"who"}}}[$i];
  if ($who eq "TEST") { pgsci (1); }
 }
 
 if ($is_az) {
  pgsci (25);
  $azim = ${@{$list{"az"}}}[$i];
  if ($azim > 200) { pgsci (10); }
  if ($azim < 200 && $azim > 190) { pgsci (11); }
  if ($azim < 190) { pgsci (12); }
 } # if ($is_az)
 
 if ($ra2 <= 86400) {
   pgrect ($ra1, $ra2, $dec1, $dec2);
 } else {
   pgrect ($ra1, 86400, $dec1, $dec2);
   pgrect (0, $ra2 - 86400, $dec1, $dec2);
 }
  pgsci (25);
#  sleep (2);
}
 # print overlap regions
 if ($is_overlap) {
  &print_overlap();
 }
 $net_covered_area = $covered_area - $overlapped_area;
 $net_covered_area_percent = sprintf ("%.2f", ($net_covered_area/$survey_area) * 100.);
 $net_covered_area = sprintf ("%.2f", $net_covered_area);
 $covered_area_percent = sprintf ("%.2f", ($covered_area/$survey_area) * 100.);
 $covered_area = sprintf ("%.2f", $covered_area);
 $overlapped_area_percent = sprintf ("%.2f", ($overlapped_area/$survey_area) * 100.);
 $overlapped_area = sprintf ("%.2f", $overlapped_area);

 pgsci (1);
 pgsch (0.6 * $char_size);
 pgmtxt ("B", 7, 0.0, 0.0, "Total survey area = $survey_area sq. deg ($survey_area_percent%)");
 pgmtxt ("B", 7, 0.35, 0.0, "Covered area = $covered_area sq. deg ($covered_area_percent%)");
 pgmtxt ("B", 7, 0.7, 0.0, "Overlapped area = $overlapped_area sq. deg ($overlapped_area_percent%)");
 pgmtxt ("B", 8.5, 0.0, 0.0, "Net covered area = $net_covered_area sq. deg ($net_covered_area_percent%)");
 pgsch ($char_size);

 print "\n";
 print "----------------------------------------------\n";
 print "Total survey area = $survey_area sq. deg ($survey_area_percent%)\n";
 print "Covered area = $covered_area sq. deg ($covered_area_percent%)\n";
 print "Overlapped area = $overlapped_area sq. deg ($overlapped_area_percent%)\n";
 print "Net covered area = $net_covered_area sq. deg ($net_covered_area_percent%)\n";
 print "----------------------------------------------\n";
 print "\n";
 $is_coords = 1;
}

# set gray scale colors
sub grayscale {
 local ($dev) = @_;
 $step = 1./15.;
 if ($dev =~ /\.ps/) {
  for ($i=0; $i<16; $i++) {
   $val = (15 - $i) * $step;
   pgscr (16+$i, $val, $val, $val);
  }
 } else {
    for ($i=0; $i<16; $i++) {
     $val = $i * $step;
     pgscr (16+$i, $val, $val, $val);
    }
   }
}

# print DEC or RA in user-friendly format
sub get_decra {
 my ($sec) = @_;
 if ($sec > 0) { $sign = ""; } else { $sign = "-"; }
 $sec = abs($sec);
 $dh = int ($sec / 3600.);
 $min = int (($sec - $dh * 3600) / 60.);
 $sec = sprintf ("%.2f", $sec - $dh * 3600. - $min * 60.);
 if ($dh < 10) { $dh = "0$dh"; }
 if ($min < 10) { $min = "0$min"; }
 if ($sec < 10) { $sec = "0$sec"; }
 return "$sign$dh:$min:$sec";
}

# calculate local mean siderial time  (in hours)
# see    http://aa.usno.navy.mil/faq/docs/GAST.html
sub siderial {
 my ($mjd) = @_;
 $jd = $mjd + 2400000.5;
 $part = $jd - int ($jd);
 if ($part >= 0.5) { $jd0 = int($jd) + 0.5; $H = ($part - 0.5) * 24.; }
  else { $jd0 = int($jd) - 0.5; $H = ($part + 0.5) * 24.; }

 $D  = $jd  - 2451545.0;
 $D0 = $jd0 - 2451545.0;
 $T = $D / 36525;

 $gmst = 6.697374558 + 0.06570982441908 * $D0 + 1.00273790935 * $H + 0.000026 * $T * $T + $gbt_longitude/15.;

 return $gmst - (int($gmst/24.)) * 24.;
}

# calculate delta from EL and AZ
sub get_delta {
 my ($el, $az) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az += 180.;
 while ($az >= 360.) { $az -= 360.; }

 $delta = asin ( sin ($el * $rad) * sin ($gbt_latitude * $rad) - cos ($el * $rad) * cos ($gbt_latitude * $rad) * cos ($az * $rad) );
 $delta /= $rad;
 return $delta;
}

# calculate hour angle from EL and AZ
sub get_hour_angle {
 my ($el, $az) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az += 180.;
 while ($az >= 360.) { $az -= 360.; }

 $hour_angle = atan2 ( cos ($el * $rad) * sin ($az * $rad), sin ($el * $rad) * cos ($gbt_latitude * $rad) + cos ($el * $rad) * sin ($gbt_latitude * $rad) * cos ($az * $rad) );
 $hour_angle /= $rad;
 $hour_angle /= 15.;
 return $hour_angle;
}

# calculate maximum & minimum achieved hour angle taking into account the beam size
sub get_ha_extremum {
 my ($el0, $az0) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az0 += 180.;
 while ($az0 >= 360.) { $az0 -= 360.; }

 $hour_angle = atan2 ( cos ($el0 * $rad) * sin ($az0 * $rad), sin ($el0 * $rad) * cos ($gbt_latitude * $rad) + cos ($el0 * $rad) * sin ($gbt_latitude * $rad) * cos ($az0 * $rad) );
 $hour_angle /= $rad;
 $hour_angle /= 15.;

 $ha_max = $hour_angle + ((($beam/2) * 4) / 3600.);
 $ha_min = $hour_angle - ((($beam/2) * 4) / 3600.);
}

# calculate maximum & minimum achieved declination taking into account the beam size
sub get_delta_extremum {
 my ($el0, $az0) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az0 += 180.;
 while ($az0 >= 360.) { $az0 -= 360.; }

 $delta = asin ( sin ($el0 * $rad) * sin ($gbt_latitude * $rad) - cos ($el0 * $rad) * cos ($gbt_latitude * $rad) * cos ($az0 * $rad) );
 $delta /= $rad;
 
 $delta_max = $delta + (($beam/2) / 60.);
 $delta_min = $delta - (($beam/2) / 60.);
}

# get the declination range taking into account the used azimuths and elevations
sub dec_range {
 @els = sort { $a <=> $b } @{ $list{"el"} };
# $abs_min = -(90.-$gbt_latitude-$els[0] + ($beam/2)/60.);
 $hmin = $els[0] - ($beam/2)/60.;

 @azs = sort { $a <=> $b } @{ $list{"az"} };
 $azmin = $azs[0] + 180.;
 while ($azmin >= 360.) { $azmin -= 360.; }

 $abs_min = asin ( sin ($hmin * $rad) * sin ($gbt_latitude * $rad) -  cos ($hmin * $rad) * cos ($gbt_latitude * $rad) * cos ($azmin * $rad) );
 $abs_min /= $rad;
 
 $abs_max = 90.;
 $flag = 0;
 foreach $a (@{ $list{"az"} }) {
  if ($a > 90 && $a < 270) { $flag = $flag | 1; }
  if (($a >= 0 && $a <=90) || ($a >= 270 && $a <=360)) { $flag = $flag | 2; }
 }
 if ($flag == 1) { @range = ($abs_min, $gbt_latitude); }
 if ($flag == 2) { @range = ($gbt_latitude, $abs_max); }
 if ($flag == 3) { @range = ($abs_min, $abs_max); }
 return @range;
}

# round the value down to the closest integer divisible by 5 
sub rounddown {
 my ($input) = @_;
 $input = int($input);
 $input -= 1; 
 while ( 5*(int($input/5)) != $input ) { $input--;  }
 return $input;
}

# round the value up to the closest integer divisible by 5 
sub roundup {
 my ($input) = @_;
 $input = int($input);
 $input += 1; 
 while ( 5*(int($input/5)) != $input ) { $input++;  }
 return $input;
}

# convert galactic coordinates to equatorial
# returned ra and dec are in hours and degrees
# input l and b are in degrees
sub gal2eq {
 my ($l, $b) = @_;
 local $decl, $ra;
 $b *= $rad;
 $l = ($l - $lasc) * $rad;
 $decl = asin (cos($b) * cos($dec_NGP * $rad) * sin($l) + sin($b) * sin($dec_NGP * $rad));
 $ra = atan2 (cos($b) * cos($l), sin($b) * cos($dec_NGP * $rad) - cos($b) * sin($l) * sin($dec_NGP * $rad));
 $decl /= $rad; 
 $ra /= $rad;
 $ra /= 15.;
 return ($ra, $decl);
}

# calculates area of overlapped regions
sub overlap_area {
 my ($index, $r1, $r2, $d1, $d2) = @_;
 my $w;
 my $offset = 0;
 my $res1 = 0;
 my $rr1, $rr2, $dd1, $dd2;
 
 for ($w=0; $w<$index; $w++) {
  if ($d1 >= ${@{$coord{"declo"}}}[$w] && $d1 <  ${@{$coord{"dechi"}}}[$w]-$offset) { 
   if ($r1 >= ${@{$coord{"ralo"}}}[$w] && $r1 < ${@{$coord{"rahi"}}}[$w]-$offset) {
     $dd1 = $d1 / 3600.; 
     $dd2 = ($d2 > ${@{$coord{"dechi"}}}[$w] ? ${@{$coord{"dechi"}}}[$w] : $d2) / 3600.;
     $rr1 = $r1 / 3600.; $rr1 *= 15.;
     $rr2 = ($r2 > ${@{$coord{"rahi"}}}[$w] ? ${@{$coord{"rahi"}}}[$w] : $r2) / 3600.;
     $rr2 *= 15.;
     $res1 = ($rr2 - $rr1) * ($dd2 - $dd1);
   }
   if ($r2 <= ${@{$coord{"rahi"}}}[$w] && $r2 > ${@{$coord{"ralo"}}}[$w]+$offset) {
     $dd1 = $d1 / 3600.; 
     $dd2 = ($d2 > ${@{$coord{"dechi"}}}[$w] ? ${@{$coord{"dechi"}}}[$w] : $d2) / 3600.;
     $rr1 = ($r1 > ${@{$coord{"ralo"}}}[$w] ? $r1 :  ${@{$coord{"ralo"}}}[$w]) / 3600.;
     $rr1 *= 15.;
     $rr2 = $r2 / 3600.; $rr2 *= 15.;
     $res1 = ($rr2 - $rr1) * ($dd2 - $dd1);
   }
  } #d1
  if ($d2 <= ${@{$coord{"dechi"}}}[$w] && $d2 >  ${@{$coord{"declo"}}}[$w]+$offset) { 
   if ($r1 >= ${@{$coord{"ralo"}}}[$w] && $r1 < ${@{$coord{"rahi"}}}[$w]-$offset) {
     $dd1 = ($d1 > ${@{$coord{"declo"}}}[$w] ? $d1 : ${@{$coord{"declo"}}}[$w]) / 3600.;
     $dd2 = $d2 / 3600.;
     $rr1 = $r1 / 3600.; $rr1 *= 15.;
     $rr2 = ($r2 > ${@{$coord{"rahi"}}}[$w] ? ${@{$coord{"rahi"}}}[$w] : $r2) / 3600.;
     $rr2 *= 15.;
     $res1 = ($rr2 - $rr1) * ($dd2 - $dd1);
   }
   if ($r2 <= ${@{$coord{"rahi"}}}[$w] && $r2 > ${@{$coord{"ralo"}}}[$w]+$offset) {
     $dd1 = ($d1 > ${@{$coord{"declo"}}}[$w] ? $d1 : ${@{$coord{"declo"}}}[$w]) / 3600.;
     $dd2 = $d2 / 3600.;
     $rr1 = ($r1 > ${@{$coord{"ralo"}}}[$w] ? $r1 :  ${@{$coord{"ralo"}}}[$w]) / 3600.;
     $rr1 *= 15.;
     $rr2 = $r2 / 3600.; $rr2 *= 15.;
     $res1 = ($rr2 - $rr1) * ($dd2 - $dd1);
   }
  } #d2
 } # for
 return $res1;
}

# calculates overlapping regions
sub overlap {
 my ($index, $r1, $r2, $d1, $d2) = @_;
 my $w;
 my $offset = 525;
 
 for ($w=0; $w<$index; $w++) {
  if ($d1 >= ${@{$coord{"declo"}}}[$w] && $d1 <  ${@{$coord{"dechi"}}}[$w]-$offset) { 
   if ($r1 >= ${@{$coord{"ralo"}}}[$w] && $r1 < ${@{$coord{"rahi"}}}[$w]-$offset) {
     push @{ $ol{"declo"} }, $d1;
     push @{ $ol{"dechi"} }, ($d2 > ${@{$coord{"dechi"}}}[$w] ? ${@{$coord{"dechi"}}}[$w] : $d2);
     push @{ $ol{"ralo"} }, $r1;
     push @{ $ol{"rahi"} }, ($r2 > ${@{$coord{"rahi"}}}[$w] ? ${@{$coord{"rahi"}}}[$w] : $r2);
   }
   if ($r2 <= ${@{$coord{"rahi"}}}[$w] && $r2 > ${@{$coord{"ralo"}}}[$w]+$offset) {
     push @{ $ol{"declo"} }, $d1;
     push @{ $ol{"dechi"} }, ($d2 > ${@{$coord{"dechi"}}}[$w] ? ${@{$coord{"dechi"}}}[$w] : $d2);
     push @{ $ol{"ralo"} }, ($r1 > ${@{$coord{"ralo"}}}[$w] ? $r1 :  ${@{$coord{"ralo"}}}[$w]);
     push @{ $ol{"rahi"} }, $r2;
   }
  } #d1
  if ($d2 <= ${@{$coord{"dechi"}}}[$w] && $d2 >  ${@{$coord{"declo"}}}[$w]+$offset) { 
   if ($r1 >= ${@{$coord{"ralo"}}}[$w] && $r1 < ${@{$coord{"rahi"}}}[$w]-$offset) {
     push @{ $ol{"declo"} }, ($d1 > ${@{$coord{"declo"}}}[$w] ? $d1 : ${@{$coord{"declo"}}}[$w]);
     push @{ $ol{"dechi"} }, $d2;
     push @{ $ol{"ralo"} }, $r1;
     push @{ $ol{"rahi"} }, ($r2 > ${@{$coord{"rahi"}}}[$w] ? ${@{$coord{"rahi"}}}[$w] : $r2);
   }
   if ($r2 <= ${@{$coord{"rahi"}}}[$w] && $r2 > ${@{$coord{"ralo"}}}[$w]+$offset) {
     push @{ $ol{"declo"} }, ($d1 > ${@{$coord{"declo"}}}[$w] ? $d1 : ${@{$coord{"declo"}}}[$w]);
     push @{ $ol{"dechi"} }, $d2;
     push @{ $ol{"ralo"} }, ($r1 > ${@{$coord{"ralo"}}}[$w] ? $r1 :  ${@{$coord{"ralo"}}}[$w]);
     push @{ $ol{"rahi"} }, $r2;
   }
  } #d2
 } # for
}

# print overlap regions
sub print_overlap {
 # overlap entries
 $ollines = $#{@{$ol{"declo"}}};

 pgsci (8);
 pgsfs (1);

 for ($e=0; $e<=$ollines; $e++) {
  if (${@{$ol{"rahi"}}}[$e] <= 86400) {
   pgrect (${@{$ol{"ralo"}}}[$e], ${@{$ol{"rahi"}}}[$e], ${@{$ol{"declo"}}}[$e], ${@{$ol{"dechi"}}}[$e]);
 } else {
   pgrect (${@{$ol{"ralo"}}}[$e], 86400, ${@{$ol{"declo"}}}[$e], ${@{$ol{"dechi"}}}[$e]);
   pgrect (0, ${@{$ol{"rahi"}}}[$e] - 86400, ${@{$ol{"declo"}}}[$e], ${@{$ol{"dechi"}}}[$e]);
 }
 } #for
}

# calculate the total sky area to cover
sub total_area {
 my ($dmi, $dma) = @_;
 $survey_area = ($dma - $dmi) * 360.;
 $survey_area_percent = sprintf ("%.2f", ($survey_area/$sphere) * 100.);
 $survey_area = sprintf ("%.2f", $survey_area); 
}

# calculate total covered area 
# not taking into account the overlap regions
sub covered {
 my ($r1, $r2, $d1, $d2) = @_;

 $r1 /= 3600.; $r1 *= 15.;
 $r2 /= 3600.; $r2 *= 15.;
 $d1 /= 3600.; $d2 /= 3600.;
 $res = ($r2 - $r1) * ($d2 - $d1); 
 return $res;
}
