#!/usr/local/bin/perl
#
# read info from OBS_LOGS.txt file and make a EL - UT plot
# Note: the last file in every session is not counted, because it could be smaller than 80s
# Note: if "Files" = 999 then this record is not counted
#
use PGPLOT;
use Getopt::Long;

$pgplot_dev = "/xwin";
$filedur = 80.02732032; # duration of every file (in sec)
$beam = 35.;            # beam-width at 350 MHz (in arcmin)

# Parse command line
GetOptions ( "h" => \$is_help,      # to print the help
             "help" => \$is_help,   # to print the help
             "p=s" => \$pgplot_dev, # pgplot device
	     "who" => \$is_who,     # mark obs session of different observer with different color
             "update" => \$is_update ); # to update OBS_LOGS.txt file

if ($is_help) {  &help($0); exit 0; }

$file = "/home/users/vkondrat/obs/drift/OBS_LOGS.txt";
# first upfate OBS_LOGS.txt file
if ($is_update) {
 system ("scp spigot2:/users/sransom/driftscan/OBS_LOGS.txt $file");
}

open (OBS, "<$file") || die "can't open file \"$file\": $!\n";
while (defined ($line = <OBS>)) {
 chomp $line;
 if ($line =~ /^#/) { next; }
 @temp = split /\s+/, $line;
 push @{ $list{"mjd"} }, $temp[0];
 push @{ $list{"who"} }, $temp[1];
 push @{ $list{"files"} }, $temp[2];
 push @{ $list{"az"} }, $temp[3];
 push @{ $list{"el"} }, $temp[4];
 push @{ $list{"bits"} }, $temp[5];
 push @{ $list{"basename"} }, $temp[6];
 push @{ $list{"disks"} }, $temp[7];
}
close (OBS);


# number of lines on OBS_LOGS.txt file
$numlines = $#{@{$list{"mjd"}}};

#
# graphics part
#
$el_min = 0;
$el_max = 90*60*60;
$ut_min = 0;
$ut_max = 3600*24;

$ydataoffset = ($el_max - $el_min) * 0.015;
$xdataoffset = ($ut_max - $ut_min) * 0.015;
$xlabel = "UT";
$ylabel = "Elevation";

$char_size = 1.;
$psfile = "elut" . ".ps/cps"; # possible ps-file for writing

$id0 = pgopen ($pgplot_dev);
pgsch ($char_size);

&mainwindow ($pgplot_dev);

if (($pgplot_dev =~ /ps/) || ($pgplot_dev =~ /PS/)) {
 exit;
}

# waiting for quit
&winhelp();
while (1) {
 # look for cursor input
 pgcurs ($x_cursor, $y_cursor, $ch_cursor);

 # to quit
 if ($ch_cursor =~ /[xXqQ]/) {
   last;
 }
 # to help
 if ($ch_cursor =~ /[hH]/) { &winhelp(); next; }

 # get value of closest candidate
 if ($ch_cursor =~ /[aA]/) {
  &print_coords ($x_cursor, $y_cursor);
  next;
 }

 # save to postscript-file
 if ($ch_cursor =~ /[pP]/) {
  $id1 = pgopen ($psfile);
  pgslct ($id1);
  pgsch ($char_size);
  &mainwindow($psfile);
  pgclos ();
  pgslct ($id0);
  next;
 }

} # while


#######################################################################################
#
#
#                            R O U T I N E S
#
#

# help
sub help {
 local ($prg) = @_;
 $prg = `basename $prg`; chomp $prg;
 print "Usage: $prg [options]\n";
 print "       -p <dev>  - pgplot device\n";
 print "       -who      - mark observers\n";
 print "       -update   - to download file \"OBS_LOGS.txt\" from spigot2\n";
 print "       -h, -help - print this help\n"; 
}

# get major tic
sub gettics {
 local ($min, $max, @mult) = @_;
 local $beg = int(($max - $min)/5.);
 local $end = int(($max - $min)/4.);
 for ($j=0; $j<=$#mult+1; $j++) {
  for ($i=$beg; $i<=$end; $i++) {
   if ( (int($i/$mult[$j]))*$mult[$j] == $i ) { return $i; }
  }
 }
 return $end;
}

# window help
sub winhelp {
 print "\n";
 print "Help:\n";
 print "x, X, q, Q, Right Button  ->  quit\n";
 print "a, A, Left Button  ->  print cursor coordinates\n";
 print "p, P  ->  save plot to postscript file\n";
 print "h, H  ->  this help\n";
 print "\n";
}

# print coordinates of the cursor
sub print_coords {
 local ($xc, $yc) = @_;
 printf ("%s: %s   %s: %s\n", "EL", &get_elut($yc), "UT", &get_elut($xc));
}

# plot points in main window
sub mainwindow {

local ($dev) = @_;

pgsvp (0.07, 0.99, 0.12, 0.95);
pgswin ($ut_min - $xdataoffset, $ut_max + $xdataoffset, $el_min - $ydataoffset, $el_max + $ydataoffset);

pgslw (1);
pgsfs (1);

# set gray scale map
&grayscale ($dev);

pgsci (18);
pgbox ("G", 1*3600, 0, "G", 5*60*60, 0);

pgsci (1);
pgtbox ("BCNTSZYHOI", 0.0, 4, "BCSNTZYDOI", 0.0, 4);
pgmtxt ("B", 2.7, 0.5, 0.5, $xlabel);
pgmtxt ("L", 2.7, 0.5, 0.5, $ylabel);
pgmtxt ("T", 1., 0.5, 0.5, "350-MHz Pulsar Drift-Scan Survey");
pgiden ();

pgsci (25);
# draw a rectangular for every record
for ($i=0; $i<=$numlines; $i++) {
 $t1 = ${@{$list{"mjd"}}}[$i];
 $t1 =~ s/^\d+(\.\d+)$/0$1/;
 $t1 *= 86400;
 $t2 = ${@{$list{"files"}}}[$i];
 if ($t2 == 999) { next; }
 $t2 = $t1 + ($t2 - 1) * $filedur;
 $el = ${@{$list{"el"}}}[$i];
 $el1 = ($el * 60. - ($beam/2.)) * 60.;
 $el2 = ($el * 60. + ($beam/2.)) * 60.;

 if ($is_who) {
  pgsci (25);
  $who = ${@{$list{"who"}}}[$i];
  if ($who eq "SMR") { pgsci (3); }
  if ($who eq "VIK") { pgsci (4); }
  if ($who eq "JVL") { pgsci (6); }
  if ($who eq "MAM") { pgsci (7); }
  if ($who eq "DRL") { pgsci (8); }
  if ($who eq "SDSS") { pgsci (9); }
  if ($who eq "JWTH") { pgsci (10); }
  if ($who eq "JRB") { pgsci (11); }
  if ($who eq "CRE") { pgsci (12); }
  if ($who eq "DAL") { pgsci (13); }
  if ($who eq "MSR") { pgsci (14); }
 }
 
 if ($t2 <= 86400) {
  pgrect ($t1, $t2, $el1, $el2);
 } else {
  pgrect ($t1, 86400, $el1, $el2);
  pgrect (0, $t2 - 86400, $el1, $el2);
 }
}

}

# set gray scale colors
sub grayscale {
 local ($dev) = @_;
 $step = 1./15.;
 if ($dev =~ /\.ps/) {
  for ($i=0; $i<16; $i++) {
   $val = (15 - $i) * $step;
   pgscr (16+$i, $val, $val, $val);
  }
 } else {
    for ($i=0; $i<16; $i++) {
     $val = $i * $step;
     pgscr (16+$i, $val, $val, $val);
    }
   }
}

# print elevation or UT in user-friendly format
sub get_elut {
 my ($sec) = @_;
 $dh = int ($sec / 3600.);
 $min = int (($sec - $dh * 3600) / 60.);
 $sec = sprintf ("%.2f", $sec - $dh * 3600. - $min * 60.);
 if ($dh < 10) { $dh = "0$dh"; }
 if ($min < 10) { $min = "0$min"; }
 if ($sec < 10) { $sec = "0$sec"; }
 return "$dh:$min:$sec";
}
