#!/usr/local/bin/perl
#
# read info from OBS_LOGS.txt file and make a DEC - RA plot
# Note: the last file in every session is not counted, because it could be smaller than 80s
# Note: if "Files" = 999 then this record is not counted
#
use PGPLOT;
use Getopt::Long;
use Math::Trig;

$pgplot_dev = "/xwin";
$filedur = 80.02732032;       # duration of every file (in sec)
$beam = 35.;                  # beam-width at 350 MHz (in arcmin)
$gbt_longitude = -79.8398333; # GBT longitude (in degrees);  GBT longitude = 79 deg 50' 23.40" W
$gbt_latitude = 38.43311944;  # GBT latitude (in degrees);   GBT latitude =  38 deg 25' 59.23" N

$pi = 3.14159265359;
$rad = $pi/180.;
$mult = 1;

# Parse command line
GetOptions ( "h" => \$is_help,      # to print the help
             "help" => \$is_help,   # to print the help
             "p=s" => \$pgplot_dev, # pgplot device
	     "who" => \$is_who,     # mark obs session of different observer with different color
             "update" => \$is_update ); # to update OBS_LOGS.txt file

if ($is_help) {  &help($0); exit 0; }

$file = "/home/users/vkondrat/obs/drift/OBS_LOGS.txt";
# first upfate OBS_LOGS.txt file
if ($is_update) {
 system ("scp spigot2:/users/sransom/driftscan/OBS_LOGS.txt $file");
}

open (OBS, "<$file") || die "can't open file \"$file\": $!\n";
while (defined ($line = <OBS>)) {
 chomp $line;
 if ($line =~ /^#/) { next; }
 @temp = split /\s+/, $line;
 push @{ $list{"mjd"} }, $temp[0];
 push @{ $list{"who"} }, $temp[1];
 push @{ $list{"files"} }, $temp[2];
 push @{ $list{"az"} }, $temp[3];
 push @{ $list{"el"} }, $temp[4];
 push @{ $list{"bits"} }, $temp[5];
 push @{ $list{"basename"} }, $temp[6];
 push @{ $list{"disks"} }, $temp[7];
}
close (OBS);


# number of lines on OBS_LOGS.txt file
$numlines = $#{@{$list{"mjd"}}};

#
# graphics part
#
$b_min = -90.*60*60;
$b_max = 90.*60*60;
$l_min = -180. * 3600.;
$l_max = 180. * 3600.;

$ydataoffset = ($b_max - $b_min) * 0.015;
$xdataoffset = ($l_max - $l_min) * 0.015;
$xlabel = "Galactic Longitude";
$ylabel = "Galactic Latitude";

$char_size = 1.;
$psfile = "galact" . ".ps/cps"; # possible ps-file for writing

$id0 = pgopen ($pgplot_dev);
pgsch ($char_size);

&mainwindow ($pgplot_dev);

if (($pgplot_dev =~ /ps/) || ($pgplot_dev =~ /PS/)) {
 exit;
}

# waiting for quit
&winhelp();
while (1) {
 # look for cursor inpra
 pgcurs ($x_cursor, $y_cursor, $ch_cursor);

 # to quit
 if ($ch_cursor =~ /[xXqQ]/) {
   last;
 }
 # to help
 if ($ch_cursor =~ /[hH]/) { &winhelp(); next; }

 # get value of closest candidate
 if ($ch_cursor =~ /[aA]/) {
  &print_coords ($x_cursor, $y_cursor);
  next;
 }

 # save to postscript-file
 if ($ch_cursor =~ /[pP]/) {
  $id1 = pgopen ($psfile);
  pgslct ($id1);
  pgsch ($char_size);
  &mainwindow($psfile);
  pgclos ();
  pgslct ($id0);
  &winhelp();
  next;
 }

} # while


#######################################################################################
#
#
#                            R O U T I N E S
#
#

# help
sub help {
 local ($prg) = @_;
 $prg = `basename $prg`; chomp $prg;
 print "Usage: $prg [options]\n";
 print "       -p <dev>  - pgplot device\n";
 print "       -who      - mark observers\n";
 print "       -update   - to download file \"OBS_LOGS.txt\" from spigot2\n";
 print "       -h, -help - print this help\n"; 
}

# get major tic
sub gettics {
 local ($min, $max, @mult) = @_;
 local $beg = int(($max - $min)/5.);
 local $end = int(($max - $min)/4.);
 for ($j=0; $j<=$#mult+1; $j++) {
  for ($i=$beg; $i<=$end; $i++) {
   if ( (int($i/$mult[$j]))*$mult[$j] == $i ) { return $i; }
  }
 }
 return $end;
}

# window help
sub winhelp {
 print "\n";
 print "Help:\n";
 print "x, X, q, Q, Right Button  ->  quit\n";
 print "a, A, Left Button  ->  print cursor coordinates\n";
 print "p, P  ->  save plot to postscript file\n";
 print "h, H  ->  this help\n";
 print "\n";
}

# print coordinates of the cursor
sub print_coords {
 local ($xc, $yc) = @_;
 printf ("%s: %s   %s: %s\n", "B", $yc/3600., "L", $xc/3600.);
}

# plot points in main window
sub mainwindow {

local ($dev) = @_;

pgsvp (0.07, 0.99, 0.12, 0.95);
pgswin ($l_min - $xdataoffset, $l_max + $xdataoffset, $b_min - $ydataoffset, $b_max + $ydataoffset);

pgslw (1);
pgsfs (1);

# set gray scale map
&grayscale ($dev);

pgsci (18);
pgbox ("G", 20*3600, 0, "G", 10*60*60, 0);

pgsci (1);
pgtbox ("BCNTSZYDOI", 0.0, 4, "BCSNTZYDOI", 0.0, 4);
pgmtxt ("B", 2.7, 0.5, 0.5, $xlabel);
pgmtxt ("L", 2.7, 0.5, 0.5, $ylabel);
pgmtxt ("T", 1., 0.5, 0.5, "350-MHz Pulsar Drift-Scan Survey");
pgiden ();

pgsci (25);
# draw a rectangular for every record
printf ("# total: %d observing sessions\n", $numlines+1);
for ($i=0; $i<=$numlines; $i++) {
 $t1 = ${@{$list{"mjd"}}}[$i];
 $el = ${@{$list{"el"}}}[$i];
 $az = ${@{$list{"az"}}}[$i];
 $who = ${@{$list{"who"}}}[$i];

 $num = ${@{$list{"files"}}}[$i];
 if ($num == 999) { next; }
 print ("#$i   $t1   $who   $num   $az   $el\n");

 &get_delta_extremum ($el, $az);
 &get_ha_extremum ($el, $az);

 for ($k=0; $k<=$mult*($num-1); $k++) { 

  $t = $t1 + ($k * ($filedur/$mult)) / 86400.;
  $lst = &siderial ($t);

  $ra_min = $lst - $ha_min;
  $ra_max = $lst - $ha_max;

 if ($ra_min < 0) { $ra_min += 24.; }
 if ($ra_max < 0) { $ra_max += 24.; }

 &get_galact ($ra_min, $delta_min);
 $l1 = $galL * 3600.;
 $b1 = $galB * 3600.;
 &get_galact ($ra_max, $delta_max);
 $l2 = $galL * 3600.;
 $b2 = $galB * 3600.;
 
 if ($is_who) {
  pgsci (25);
  $who = ${@{$list{"who"}}}[$i];
  if ($who eq "SMR") { pgsci (3); }
  if ($who eq "VIK") { pgsci (4); }
  if ($who eq "JVL") { pgsci (6); }
  if ($who eq "MAM") { pgsci (7); }
  if ($who eq "DRL") { pgsci (8); }
  if ($who eq "SDSS") { pgsci (9); }
  if ($who eq "JWTH") { pgsci (10); }
  if ($who eq "JRB") { pgsci (11); }
  if ($who eq "CRE") { pgsci (12); }
  if ($who eq "DAL") { pgsci (13); }
  if ($who eq "MSR") { pgsci (14); }
 }
  
  &get_circ ($l1, $l2, $b1, $b2);
  pgcirc ($lc, $bc, $rc); 

 } # for $k
}

}

# set gray scale colors
sub grayscale {
 local ($dev) = @_;
 $step = 1./15.;
 if ($dev =~ /\.ps/) {
  for ($i=0; $i<16; $i++) {
   $val = (15 - $i) * $step;
   pgscr (16+$i, $val, $val, $val);
  }
 } else {
    for ($i=0; $i<16; $i++) {
     $val = $i * $step;
     pgscr (16+$i, $val, $val, $val);
    }
   }
}

# print DEC or RA in user-friendly format
sub get_decra {
 my ($sec) = @_;
 $dh = int ($sec / 3600.);
 $min = int (($sec - $dh * 3600) / 60.);
 $sec = sprintf ("%.2f", $sec - $dh * 3600. - $min * 60.);
 if ($dh < 10) { $dh = "0$dh"; }
 if ($min < 10) { $min = "0$min"; }
 if ($sec < 10) { $sec = "0$sec"; }
 return "$dh:$min:$sec";
}

# calculate local mean siderial time  (in hours)
# see    http://aa.usno.navy.mil/faq/docs/GAST.html
sub siderial {
 my ($mjd) = @_;
 $jd = $mjd + 2400000.5;
 $part = $jd - int ($jd);
 if ($part >= 0.5) { $jd0 = int($jd) + 0.5; $H = ($part - 0.5) * 24.; }
  else { $jd0 = int($jd) - 0.5; $H = ($part + 0.5) * 24.; }

 $D  = $jd  - 2451545.0;
 $D0 = $jd0 - 2451545.0;
 $T = $D / 36525;

 $gmst = 6.697374558 + 0.06570982441908 * $D0 + 1.00273790935 * $H + 0.000026 * $T * $T + $gbt_longitude/15.;

 return $gmst - (int($gmst/24.)) * 24.;
}

# calculate delta from EL and AZ
sub get_delta {
 my ($el, $az) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az += 180.;
 while ($az >= 360.) { $az -= 360.; }

 $delta = asin ( sin ($el * $rad) * sin ($gbt_latitude * $rad) - cos ($el * $rad) * cos ($gbt_latitude * $rad) * cos ($az * $rad) );
 $delta /= $rad;
 return $delta;
}

# calculate hour angle from EL and AZ
sub get_hour_angle {
 my ($el, $az) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az += 180.;
 while ($az >= 360.) { $az -= 360.; }

 $hour_angle = atan2 ( cos ($el * $rad) * sin ($az * $rad), sin ($el * $rad) * cos ($gbt_latitude * $rad) + cos ($el * $rad) * sin ($gbt_latitude * $rad) * cos ($az * $rad) );
 $hour_angle /= $rad;
 $hour_angle /= 15.;
 return $hour_angle;
}

# calculate maximum & minimum achieved hour angle taking into account the beam size
sub get_ha_extremum {
 my ($el0, $az0) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az0 += 180.;
 while ($az0 >= 360.) { $az0 -= 360.; }

 $hour_angle = atan2 ( cos ($el0 * $rad) * sin ($az0 * $rad), sin ($el0 * $rad) * cos ($gbt_latitude * $rad) + cos ($el0 * $rad) * sin ($gbt_latitude * $rad) * cos ($az0 * $rad) );
 $hour_angle /= $rad;
 $hour_angle /= 15.;

 $ha_max = $hour_angle;
 $ha_min = $hour_angle;

 for ($j=0; $j<360; $j+=1) {
  $el = $el0 + (($beam/2.)/60.) * sin ($j * $rad);
  $az = $az0 + (($beam/2.)/60.) * cos ($j * $rad);
  $hour_angle = atan2 ( cos ($el * $rad) * sin ($az * $rad), sin ($el * $rad) * cos ($gbt_latitude * $rad) + cos ($el * $rad) * sin ($gbt_latitude * $rad) * cos ($az * $rad) );
  $hour_angle /= $rad;
  $hour_angle /= 15.;
  if ($hour_angle > $ha_max) { $ha_max = $hour_angle; }
  if ($hour_angle < $ha_min) { $ha_min = $hour_angle; }
 }
}

# calculate maximum & minimum achieved declination taking into account the beam size
sub get_delta_extremum {
 my ($el0, $az0) = @_;

 # converting the azimuth
 # because azimuths used here are geodesy rather than astronomical
 # in OBS_LOGS.txt file azimuth is counting clockwise from N
 # in formulas here we count azimuth clockwise from S 
 $az0 += 180.;
 while ($az0 >= 360.) { $az0 -= 360.; }

 $delta = asin ( sin ($el0 * $rad) * sin ($gbt_latitude * $rad) - cos ($el0 * $rad) * cos ($gbt_latitude * $rad) * cos ($az0 * $rad) );
 $delta /= $rad;
 
 $delta_max = $delta;
 $delta_min = $delta;

 for ($j=0; $j<360; $j+=1) {
  $el = $el0 + (($beam/2.)/60.) * sin ($j * $rad);
  $az = $az0 + (($beam/2.)/60.) * cos ($j * $rad);
  $delta = asin ( sin ($el * $rad) * sin ($gbt_latitude * $rad) - cos ($el * $rad) * cos ($gbt_latitude * $rad) * cos ($az * $rad) );
  $delta /= $rad;
  if ($delta > $delta_max) { $delta_max = $delta; }
  if ($delta < $delta_min) { $delta_min = $delta; }
 }
}

# convert RA-DEC to galactic coordinates
sub get_galact {
 my ($ra, $dec) = @_;
 
 $g00 = -0.06699; $g01 = -0.87276; $g02 = -0.48354;
 $g10 = 0.49273;  $g11 = -0.45035; $g12 = 0.74458;
 $g20 = -0.86760; $g21 = -0.18837; $g22 = 0.46020;

 $cdec = cos ($dec * $rad);
 $e0 = $cdec * cos ($ra * 15. * $rad);
 $e1 = $cdec * sin ($ra * 15. * $rad);
 $e2 = sin ($dec * $rad);

 $z = $g20 * $e0 + $g21 * $e1 + $g22 * $e2;
 $y = $g10 * $e0 + $g11 * $e1 + $g12 * $e2;
 $x = $g00 * $e0 + $g01 * $e1 + $g02 * $e2;

 $galB = asin ($z);
 $galB /= $rad;
 $galL = atan2 ($y, $x);
 $galL /= $rad;
}

# determin the center of circle and radius to plot in B-L figure
sub get_circ {
 my ($l1, $l2, $b1, $b2) = @_;
 $l1 /= 3600.; $l2 /= 3600.; $b1 /= 3600.; $b2 /= 3600.;
 if ($l1 > 90. && $l2 < 0.) { $l2 += 360.; }
 if ($l1 < -90. && $l2 > 0.) { $l1 += 360.; }
 if ($l2 > $l1) { $ld = $l2 - $l1; $lc = $l1 + $ld/2.; } else { $ld = $l1 - $l2; $lc = $l2 + $ld/2.; }
 if ($b2 > $b1) { $bd = $b2 - $b1; $bc = $b1 + $bd/2.; } else { $bd = $b1 - $b2; $bc = $b2 + $bd/2.; }
 if ($ld > $bd) { $rc = $bd/2.; } else { $rc = $ld/2.; }
 if ($lc > 180.) { $lc -= 360.; }
 if ($ls < - 180.) { $lc += 360.; }
 if ($rc > 0.5) { $rc = 0.5; }
 $rc *= 3600.;
 $lc *= 3600.;
 $bc *= 3600.;
}
