#!/usr/bin/perl
#
# 
# in formulae azimuth is counted from S clockwise to W
# but we are accepting the azimuth as for GBT, counted from N to E clockwise
# and then converting it.
#
use Math::Trig;

$pi = 3.14159265359;
$rad = $pi/180.;

# GBT latitude: 38 deg 25' 59.23"
$phi = 38.4331194444;

# pulsar catalog list
$catalog = "/users/vkondrat/obs/drift/get_psr/catlist2";

#offset from pointing position 35'
$eloff = 0.583333 / 2.; # in deg
$azoff = 0.58333 / 2.; # in deg
$off = 35 / 2.;  # in arcmin

print "\n";
print "Azimuth (from N to E clockwise, 0-360 deg): ";
while (not defined ($azdeg = <>)) {}
print "Elevation (deg): ";
while (not defined ($eldeg = <>)) {}

chomp $azdeg; chomp $eldeg;

if ($azdeg =~ /[^\d\.\-\+]/) { print "Wrong azimuth value!\n"; exit 1; }
if ($eldeg =~ /[^\d\.\-\+]/) { print "Wrong elevation value!\n"; exit 1; }

#converting the azimuth
$azdeg += 180.;
while ($azdeg >= 360.) { $azdeg -= 360.; }

# delta is in deg, ha is in hours
($delta, $ha)  = &delta ($azdeg, $eldeg);
$delta_min = $delta - $off/60.;
$delta_max =  $delta + $off/60.;

$ha_min = $ha - ($off * 4) / 3600.;
$ha_max = $ha + ($off * 4) / 3600.;

printf ("Dec range (deg): %.5f - %.5f\n", $delta_min, $delta_max);
printf ("Hour angle range (h): %.5f - %.5f\n\n", $ha_min, $ha_max);

open (CAT, "$catalog") || die "can't open catalog: $catalog ($!)\n";
while (defined ($line = <CAT>)) {
 chomp $line;
 if ($line =~ /^##/) { print "$line   offset\n#\n"; next; }
 if ($line =~ /^#/) { print "$line\n"; next; }
 if ($line =~ /^\s+/) { next; }
 @list = split /\s+/, $line;
 $decg = $list[5]; $decg =~ s/^([+-]\d\d)\:.*$/$1/;
 $decm = $list[5]; $decm =~ s/^[+-]\d\d\:(\d\d).*$/$1/;
 $decs = $list[5];
 if ($decs =~ /^[+-]\d\d\:\d\d\:\d\d\.?\d?$/) {
  $decs =~ s/^[+-]\d\d\:\d\d\:(\d\d\.?\d?)$/$1/;
 } else { $decs = 0.; }

 if ($decg =~ /^-/) { $psrdelta = $decg - ($decs/60. + $decm)/60.; }
  else { $psrdelta = ($decs/60. + $decm)/60. + $decg; }

 if ($psrdelta >= $delta_min && $psrdelta <= $delta_max) {
  # elevation of the pulsar
  $psrel = &elevation ($ha_min, $ha_max, $psrdelta);
  # azimuth of the pulsar
  $psraz = &azimuth ($ha_min, $ha_max, $psrel, $psrdelta);
  # offset in arcmin
  $offset_el = $psrel - $eldeg;
  $offset_az = $psraz - $azdeg;
  $offset = acos ( cos ($offset_el*$rad) * cos ($offset_az*$rad) );
  $offset = (abs($offset) / $pi)*180.*60.;  # convert to arcmin
  $offset = sprintf ("%.2f", $offset);
  if ($offset <= $off) { print "$line   $offset\n"; }
  if ($offset > $off && $offset <= $off + 5) { print "#$line ($offset)\n"; }
 }

}
close (CAT);




# calculate delta of the source
sub delta {
 my ($az, $el) = @_;

 $az *= $rad;
 $el *= $rad;
 
 $delta =  asin ( sin($el)*sin($phi*$rad) - cos($el)*cos($phi*$rad)*cos($az) );
 $delta /= $rad;

 # hour angle
 $hour_angle = atan2 ( cos ($el) * sin ($az), sin ($el) * cos ($phi * $rad) + cos ($el) * sin ($phi * $rad) * cos ($az) );
 $hour_angle /= $rad;
 $hour_angle /= 15.;

 push @dec, $delta;
 push @ha, $hour_angle;
 
 return ($delta, $hour_angle);
}

sub elevation {
 my ($hmin, $hmax, $delta) = @_;

 $delta *= $rad;
 $hmin *= (15 * $rad);
 $hmax *= (15 * $rad);
 $hstep = ($hmax-$hmin) / 100.;
 $eldiff = 1000;

 for ($h=$hmin; $h<=$hmax; $h=$h+$hstep) {
  $el = acos ( sin($delta)*sin($phi*$rad) + cos($delta)*cos($phi*$rad)*cos($h) );
  $el = ($el/$pi)*180.;
  $el = 90. - $el;
  if (abs($el - $eldeg) < $eldiff) { $eldiff = abs($el - $eldeg); $elmin = $el; }
 }
 return $elmin;
}

sub azimuth {
 my ($hmin, $hmax, $el, $delta) = @_;

 $el *= $rad;
 $delta *= $rad;
 $hmin *= (15 * $rad);
 $hmax *= (15 * $rad);
 $hstep = ($hmax-$hmin) / 100.;
 $azdiff = 1000;

 for ($h=$hmin; $h<=$hmax; $h=$h+$hstep) {
  $az1 = acos ( (-sin($delta)*cos($phi*$rad) + cos($delta)*sin($phi*$rad)*cos($h)) / cos ($el) );
  $az1 = ($az1/$pi)*180.;
  $az2 = asin ( cos($delta)*sin($h) / cos($el) );
  $az2 = ($az2/$pi)*180.;
  if ($az1 > 0 && $az2 < 0) { $az = 360. - $az1;}
  if ($az1 > 0 && $az2 > 0) { $az = $az1; }
  if ($az1 < 0 && $az2 < 0) { $az = 360. - $az1; }
  if ($az1 < 0 && $az2 > 0) { $az = $az1; }
  if (abs($az - $azdeg) < $azdiff) { $azdiff = abs($az - $azdeg); $azmin = $az; }
 }
  return $azmin;
}

# calc min
sub min {
 local (@arr) = sort { $a <=> $b } @_;
 return $arr[0];
}

# calc max
sub max {
 local (@arr) = sort { $b <=> $a } @_;
 return $arr[0];
}
