#!/usr/bin/perl
#
# makes a list of pulsars from ATNF catalog that are visible at selected
# elevations. Last 2 columns are elevation of the pulsar at selected azimuth
# and its hour angle
# 
# in formulae azimuth is counted from S clockwise to W
# but we are accepting the azimuth as for GBT, counted from N to E clockwise
# and then converting it.
#
use Math::Trig;

$pi = 3.14159265359;
$rad = $pi/180.;

# GBT latitude: 38 deg 25' 59.23"
$phi = 38.4331194444;

# pulsar catalog list
$catalog = "/users/vkondrat/obs/drift/get_psr/catlist2";

#offset from pointing position 35'
$eloff = 0.583333 / 2.; # in deg
$azoff = 0.58333 / 2.; # in deg
$off = 35 / 2.;  # in arcmin

print "\n";
print "Azimuth (from N to E clockwise, 0-360 deg): ";
while (not defined ($azdeg = <>)) {}
print "Elevation (deg) min: ";
while (not defined ($eldeg_min = <>)) {}
print "                max: ";
while (not defined ($eldeg_max = <>)) {}

chomp $azdeg; chomp $eldeg_min; chomp $eldeg_max;

if ($azdeg =~ /[^\d\.\-\+]/) { print "Wrong azimuth value!\n"; exit 1; }
if ($eldeg_min =~ /[^\d\.\-\+]/) { print "Wrong elevation (min) value!\n"; exit 1; }
if ($eldeg_max =~ /[^\d\.\-\+]/) { print "Wrong elevation (max) value!\n"; exit 1; }

#converting the azimuth
$azdeg += 180.;
while ($azdeg >= 360.) { $azdeg -= 360.; }

# delta is in deg, ha is in hours
($delta_min, $ha1)  = &delta ($azdeg, $eldeg_min);
($delta_max, $ha2)  = &delta ($azdeg, $eldeg_max);
$delta_min = $delta_min - $off/60.;
$delta_max =  $delta_max + $off/60.;

$ha_min = $ha2 - ($off * 4) / 3600.;
$ha_max = $ha1 + ($off * 4) / 3600.;

printf ("Dec range (deg): %.5f - %.5f\n", $delta_min, $delta_max);
printf ("Hour angle range (h): %.5f - %.5f\n\n", $ha_min, $ha_max);

open (CAT, "$catalog") || die "can't open catalog: $catalog ($!)\n";
while (defined ($line = <CAT>)) {
 chomp $line;
 if ($line =~ /^##/) { print "$line   EL      HA\n#\n"; next; }
 if ($line =~ /^#/) { print "$line\n"; next; }
 if ($line =~ /^\s+/) { next; }
 if ($line == "") { next; } 
 @list = split /\s+/, $line;
 $decg = $list[5]; $decg =~ s/^([+-]\d\d)\:.*$/$1/;
 $decm = $list[5]; $decm =~ s/^[+-]\d\d\:(\d\d).*$/$1/;
 $decs = $list[5];
 if ($decs =~ /^[+-]\d\d\:\d\d\:\d\d\.?\d?$/) {
  $decs =~ s/^[+-]\d\d\:\d\d\:(\d\d\.?\d?)$/$1/;
 } else { $decs = 0.; }

 if ($decg =~ /^-/) { $psrdelta = $decg - ($decs/60. + $decm)/60.; }
  else { $psrdelta = ($decs/60. + $decm)/60. + $decg; }

 if ($psrdelta >= $delta_min && $psrdelta <= $delta_max) {
  # elevation  & azimuth of the pulsar
  $ha_psr1 = $ha_min;
  $ha_psr2 = $ha_max;
  for ($i=0; $i<5; $i++) { # iterations
   ($psrel1, $psrel2) = &elevation ($ha_psr1, $ha_psr2, $psrdelta);
   ($delta_psr1, $ha_psr1)  = &delta ($azdeg, $psrel1);
   ($delta_psr2, $ha_psr2)  = &delta ($azdeg, $psrel2);
  }
  $psrel = ($psrel1 + $psrel2) / 2.;
  $psrel = sprintf ("%.3f", $psrel);
  $ha_psr = ($ha_psr1 + $ha_psr2) / 2.;
  $ha_psr = sprintf ("%.3f", $ha_psr);

  print "$line  $psrel  $ha_psr\n";
 }

}
close (CAT);




# calculate delta of the source
sub delta {
 my ($az, $el) = @_;

 $az *= $rad;
 $el *= $rad;
 
 $delta =  asin ( sin($el)*sin($phi*$rad) - cos($el)*cos($phi*$rad)*cos($az) );
 $delta /= $rad;

 # hour angle
 $hour_angle = atan2 ( cos ($el) * sin ($az), sin ($el) * cos ($phi * $rad) + cos ($el) * sin ($phi * $rad) * cos ($az) );
 $hour_angle /= $rad;
 $hour_angle /= 15.;

 push @dec, $delta;
 push @ha, $hour_angle;
 
 return ($delta, $hour_angle);
}


sub elevation {
 my ($hmin, $hmax, $delta) = @_;
 
 $delta *= $rad;
 $hmin *= (15 * $rad);
 $hmax *= (15 * $rad);
 
  $el1 = acos ( sin($delta)*sin($phi*$rad) + cos($delta)*cos($phi*$rad)*cos($hmin) );
  $el2 = acos ( sin($delta)*sin($phi*$rad) + cos($delta)*cos($phi*$rad)*cos($hmax) );
  $el1 = ($el1/$pi)*180.;
  $el1 = 90. - $el1;
  $el2 = ($el2/$pi)*180.;
  $el2 = 90. - $el2;

 return ($el1, $el2);
}




sub azimuth {
 my ($hmin, $hmax, $el, $delta) = @_;

 $el *= $rad;
 $delta *= $rad;
 $hmin *= (15 * $rad);
 $hmax *= (15 * $rad);
 $hstep = ($hmax-$hmin) / 1000.;
 $azdiff = 10000;

 for ($h=$hmin; $h<=$hmax; $h=$h+$hstep) {
  $az1 = acos ( (-sin($delta)*cos($phi*$rad) + cos($delta)*sin($phi*$rad)*cos($h)) / cos ($el) );
  $az1 = ($az1/$pi)*180.;
  $az2 = asin ( cos($delta)*sin($h) / cos($el) );
  $az2 = ($az2/$pi)*180.;
  if ($az1 > 0 && $az2 < 0) { $az = 360. - $az1;}
  if ($az1 > 0 && $az2 > 0) { $az = $az1; }
  if ($az1 < 0 && $az2 < 0) { $az = 360. - $az1; }
  if ($az1 < 0 && $az2 > 0) { $az = $az1; }
  if (abs($az - $azdeg) < $azdiff) { $azdiff = abs($az - $azdeg); $azmin = $az; }
 }
  return $azmin;
}

# calc min
sub min {
 local (@arr) = sort { $a <=> $b } @_;
 return $arr[0];
}

# calc max
sub max {
 local (@arr) = sort { $b <=> $a } @_;
 return $arr[0];
}
