#!/usr/bin/perl
#
# This scripts reads txt-file with observational parameters
# like Tsys, G, ..., freq, bandwidth, number of channels, ...
# and calculates expected DM, scattering time
# and minimum detectable flux

use Getopt::Long;

$period = 1000;
$wintr = 100;
$l = 90;
$b = 30;
$dist = 3;
$spind = -1.8;

$Tcmb = 2.7;  # temperature of cosmic microwave background
$Tspill = 0;  # temperature from the ground, etc. (spill-over temperature)
              # consider it to be zero now

$prg_tsky = "~dlorimer/bin/tsky";    # Dunc's program to calculate the sky temperature
$prg_ne2001 = "~mmclaugh/bin/getdm"; # Maura's program to get DM from NE2001 model
$prg_disp = "~vkondrat/psips/bin/dide";  # My script to calculate dispersion smearing
$prg_scat = "~vkondrat/psips/bin/tscat"; # My script to calculate scattering time

# if there are no parameters in command line
# # help is called and program normally exit
&help($0) && exit if $#ARGV < 0;

# Parse command line
GetOptions ( "p=f" => \$period, # - period
             "w=f" => \$wintr,  # - intrinsic width
	     "l=f" => \$l,      # - galactic longitude 
	     "b=f" => \$b,      # - galactic latitude
	     "d=f" => \$dist,   # - distance
             "s=f" => \$spind,  # - spectral index
             "h" => \$is_help); # - to print the help

if ($is_help) { &help($0); exit 0; }
&error ("No files with the observational setup defined!") if $#ARGV < 0;

foreach $file (@ARGV) {
 open (SETUP, "<$file") || &error("Can't open file \"$file\": $!");
 while (defined ($line = <SETUP>)) {
  chomp $line;
  if ($line =~ /^#/) { next; }
   @temp = split /\s+/, $line;
   if ($temp[0] =~ /Frequency/)   { push @{ $list{"freq"} }, $temp[2]; }
   if ($temp[0] =~ /Tsys/)        { push @{ $list{"tsys"} }, $temp[2]; }
   if ($temp[0] =~ /Gain/)        { push @{ $list{"gain"} }, $temp[2]; }
   if ($temp[0] =~ /Degradation/) { push @{ $list{"degr"} }, $temp[2]; }
   if ($temp[0] =~ /Bandwidth/)   { push @{ $list{"bw"} }, $temp[2]; }
   if ($temp[0] =~ /Nchannels/)   { push @{ $list{"nch"} }, $temp[2]; }
   if ($temp[0] =~ /Npol/)        { push @{ $list{"npol"} }, $temp[2]; }
   if ($temp[0] =~ /SNR/)         { push @{ $list{"snr"} }, $temp[2]; }
   if ($temp[0] =~ /Integration/) { push @{ $list{"time"} }, $temp[2]; }
   if ($temp[0] =~ /Sampling/)    { push @{ $list{"sample"} }, $temp[2]; }
 }
 close (SETUP);
}

# number of setups
$setups = $#{@{$list{"freq"}}};


print "\n";
print "# Period = $period ms   Intrinsic width = $wintr ms   Coordinates = ($l degr, $b degr, $dist kpc)  Sp.index = $spind\n";
print "#\n";
print "#     Freq     BW      Nch      Tsky     Trcv     Tsys     Gain     DM       tDM      tScat    Width    MinFlux  Flux1400\n";
print "#    (MHz)    (MHz)             (K)      (K)      (K)            (pc cm^-3)  (mcs)    (msc)    (ms)      (mJy)    (mJy)  \n";
print "#-------------------------------------------------------------------------------------------------------------------------\n";
for ($i=0; $i<=$setups; $i++) {
 $sky = &tsky($l, $b, ${@{$list{"freq"}}}[$i]);
 $temp = &tsys (${@{$list{"tsys"}}}[$i], $sky, $Tspill, $Tcmb);
 $DM = &get_dm ($l, $b, $dist);
 $tdm = &disp ($DM, ${@{$list{"bw"}}}[$i], ${@{$list{"nch"}}}[$i], ${@{$list{"freq"}}}[$i]);
 $tscat = &scat ($DM, ${@{$list{"freq"}}}[$i]);
 $width = &get_width ($wintr, $tdm, $tscat, ${@{$list{"sample"}}}[$i]);
 $minflux = &smin (${@{$list{"snr"}}}[$i], $temp, ${@{$list{"gain"}}}[$i], ${@{$list{"degr"}}}[$i], ${@{$list{"npol"}}}[$i], ${@{$list{"bw"}}}[$i], ${@{$list{"time"}}}[$i], $width, $period);
 $flux1400 = &scale_flux ($minflux, ${@{$list{"freq"}}}[$i], $spind);
 printf ("%-4d %-8g %-8.3g %4d     %-8.3g %-8.3g %-8.3g %-8.3g %-8.3g %-8.3g %-8.3g %-8.3g %-8.3g %-8.3g\n", $i+1, ${@{$list{"freq"}}}[$i], ${@{$list{"bw"}}}[$i], ${@{$list{"nch"}}}[$i], $sky, ${@{$list{"tsys"}}}[$i], $temp, ${@{$list{"gain"}}}[$i], $DM, $tdm, $tscat, $width, $minflux, $flux1400);
}
print "\n";

##################################################################
#
#                    R O U T I N E S
#
#


# help
sub help {
 local ($prg) = @_;
 $prg = `basename $prg`;
 chomp $prg;
 print "Usage: $prg [options] <file1 .. fileN>\n";
 print "  [options]:\n";
 print "   -p - pulsar period (in ms), default - 1000 ms\n";
 print "   -w - pulse intrinsic width (in ms), default - 100 ms\n";
 print "   -l - galactic longitude (in degrees), default - 90\n";
 print "   -b - galactic latitude (in degrees), default - 30\n";
 print "   -d - distance to the pulsar (in kpc), default - 3 kpc\n";
 print "   -s - spectral index, default - -1.8\n"; 
 print "   -h - print this help\n";
 print " file1 .. fileN - files with the observational setup\n";
}

# unique subroutine to output errors
# one input parameter (string): the error message
sub error {
 my ($error) = @_;
 print "Error: $error\n";
 exit 1;
}

# calculate the sky temperature
sub tsky {
 my ($l, $b, $f) = @_;
 my $t;
 $t = `$prg_tsky $l $b $f | awk \'\{print \$11\}\' -`;
 chomp $t;
 $t = sprintf ("%.3f", $t);
 return $t;
}

# calculate the Tsys temperature
sub tsys {
 my ($feed, $sky, $spill, $cmb) = @_;
 my $t;
 $t = $feed + $sky + $spill + $cmb;
 $t = sprintf ("%.3f", $t);
 return $t;
}

# calculate DM
sub get_dm {
 my ($l, $b, $d) = @_;
 my $dm;
 $dm = `$prg_ne2001 $l $b $d | awk \'\{print \$5\}\' -`;
 chomp $dm;
 $dm = sprintf ("%.3f", $dm);
 return $dm;
}

# calculate dispersion smearing
# return value in mcs
sub disp {
 my ($dm, $band, $n, $freq) = @_;
 my $ch, $t;
 $ch = $band / $n;
 $t = `$prg_disp $dm $ch $freq | awk \'\{print \$3\}\' -`;
 chomp $t;
 $t = sprintf ("%.3f", $t*1000.);
 return $t;
}

# calculate scattering time
# return value in mcs
sub scat {
 my ($dm, $freq) = @_;
 $freq /= 1000.;
 $t = `$prg_scat $dm $freq | awk \'\{print \$4\}\' -`;
 chomp $t;
 return $t;
}

# calculate observed width of the pulsar
# return value in ms
sub get_width {
 my ($i, $d, $s, $sm) = @_;
 my $w;
 $w = sqrt ($i*$i + ($d/1000.)*($d/1000.) + ($s/1000.)*($s/1000.) + ($sm/1000.)*($sm/1000.));
 $w = sprintf ("%.3f", $w);
 return $w;
}

# calculate minimum detectable flux
# return value in mJy
sub smin {
 my ($snr, $t, $g, $dgr, $np, $bw, $tint, $w, $p) = @_;
 my $fl, $factor;
 $factor = $w / ($p - $w); 
 if ($factor < 0) { $factor = 1.; }
 $fl = ($dgr * $snr * ($t/$g) * sqrt ($factor)) / (sqrt ($np * $bw * 1000000. * $tint * 60.));
 $fl *= 1000.;
 $fl = sprintf ("%.3f", $fl);
 return $fl;
}

# calculate flux @ 1400 MHz
sub scale_flux {
 my ($mf, $f, $sp) = @_;
 my $flux;
 $flux = $mf * (&power ((1400./$f), $sp));
 $flux = sprintf ("%.3f", $flux);
 return $flux;
}

# Subroutine to calculate x^y
sub power {
 my ($x, $y) = @_;
 return exp($y * log($x));
}
